=== modified file 'NEWS'
--- NEWS	2008-09-24 06:52:03 +0000
+++ NEWS	2008-09-27 22:58:34 +0000
@@ -9,6 +9,15 @@
 
   CHANGES:
 
+    * ``bzr log file`` has been changed. It now uses a different method
+      for determining which revisions to show as merging the changes to
+      the file. It now only shows revisions which merged the change
+      towards your mainline. This simplifies the output, makes it faster,
+      and reduces memory consumption.  (John Arbash Meinel)
+
+    * ``bzr+http//`` will now optionally load plugins and write logs on the
+      server. (Marius Kruger)
+
   FEATURES
 
   IMPROVEMENTS:
@@ -26,6 +35,9 @@
       push.  It is also faster if there are no tags in the local branch.
       (Andrew Bennetts)
 
+    * File changes during a commit will update the tree stat cache.
+      (Robert Collins)
+
     * Location aliases can now accept a trailing path.  (Micheal Hudson)
 
     * New hooks ``Lock.hooks`` when LockDirs are acquired and released. 
@@ -51,6 +63,9 @@
       repository now preserves the repository format.
       (Andrew Bennetts, #269214)
 
+    * ``bzr log`` now accepts a ``--change`` option.
+      (Vincent Ladeuil, #248427)
+
     * ``bzr missing`` now accepts an ``--include-merges`` option.
       (Vincent Ladeuil, #233817)
 
@@ -72,12 +87,33 @@
     * Make the first line of the manpage preamble a comment again.
       (David Futcher, #242106)
 
+    * The autopacking logic will now always create a single new pack from
+      all of the content which it deems is worth moving. This avoids the
+      'repack a single pack' bug and should result in better packing
+      overall.  (John Arbash Meinel, #242510, #172644)
+
+    * Trivial documentation fix.
+      (John Arbash Meinel, #270471)
+
   DOCUMENTATION:
 
     * Explain revision/range identifiers. (Daniel Clemente)
 
   API CHANGES:
 
+    * ``CommitBuilder.record_entry_contents`` returns one more element in
+      its result tuple - an optional file system hash for the hash cache
+      to use. (Robert Collins)
+
+    * ``dirstate.DirState.update_entry`` will now only calculate the sha1
+      of a file if it is likely to be needed in determining the output
+      of iter_changes. (Robert Collins)
+
+    * The PackRepository, RepositoryPackCollection, NewPack classes have a
+      slightly changed interface to support different index types; as a
+      result other users of these classes need to supply the index types
+      they want. (Robert Collins)
+
   TESTING:
 
     * ``bzrlib.tests.repository_implementations`` has been renamed to
@@ -103,13 +139,35 @@
       tree walking, and modular directory listing code to aid future
       performance optimisations and refactoring. (Robert Collins)
 
+    * ``bzrlib.trace.debug_memory`` can be used to get a quick memory dump
+      in the middle of processing. It only reports memory if
+      ``/proc/PID/status`` is available. (John Arbash Meinel)
+
     * New method ``RevisionSpec.as_tree`` for representing the revision
       specifier as a revision tree object. (Lukáš Lalinský)
 
+    * New race-free method on MutableTree ``get_file_with_stat`` for use
+      when generating stat cache results. (Robert Collins)
+
     * New win32utils.get_local_appdata_location() provides access to a local
       directory for storing data.  (Mark Hammond)
 
 
+bzr 1.7.1rc1 2008-09-24
+-----------------------
+
+This release just includes an update to how the merge algorithm handles
+file paths when we encounter complex history.
+
+  FEATURES:
+
+    * If we encounter a criss-cross in history, use information from
+      direct Least Common Ancestors to resolve inventory shape (locations
+      of files, adds, deletes, etc). This is similar in concept to using
+      ``--lca`` for merging file texts, only applied to paths.
+      (John Arbash Meinel)
+
+
 bzr 1.7 2008-09-23
 ------------------
 

=== modified file 'bzrlib/_dirstate_helpers_c.pyx'
--- bzrlib/_dirstate_helpers_c.pyx	2008-09-24 04:15:03 +0000
+++ bzrlib/_dirstate_helpers_c.pyx	2008-09-26 05:47:03 +0000
@@ -19,13 +19,29 @@
 This is the python implementation for DirState functions.
 """
 
-from bzrlib import errors
+import binascii
+import bisect
+import errno
+import os
+import stat
+
+from bzrlib import cache_utf8, errors, osutils
 from bzrlib.dirstate import DirState
-
-
-#python2.4 support
+from bzrlib.osutils import pathjoin, splitpath
+
+
+# This is the Windows equivalent of ENOTDIR
+# It is defined in pywin32.winerror, but we don't want a strong dependency for
+# just an error code.
+# XXX: Perhaps we could get it from a windows header ?
+cdef int ERROR_PATH_NOT_FOUND
+ERROR_PATH_NOT_FOUND = 3
+cdef int ERROR_DIRECTORY
+ERROR_DIRECTORY = 267
+
+#python2.4 support, and other platform-dependent includes
 cdef extern from "python-compat.h":
-    pass
+    unsigned long htonl(unsigned long)
 
 # Give Pyrex some function definitions for it to understand.
 # All of these are just hints to Pyrex, so that it can try to convert python
@@ -41,10 +57,18 @@
     ctypedef int intptr_t
 
 
+
 cdef extern from "stdlib.h":
     unsigned long int strtoul(char *nptr, char **endptr, int base)
 
 
+cdef extern from 'sys/stat.h':
+    int S_ISDIR(int mode)
+    int S_ISREG(int mode)
+    # On win32, this actually comes from "python-compat.h"
+    int S_ISLNK(int mode)
+    int S_IXUSR
+
 # These functions allow us access to a bit of the 'bare metal' of python
 # objects, rather than going through the object abstraction. (For example,
 # PyList_Append, rather than getting the 'append' attribute of the object, and
@@ -54,21 +78,36 @@
 # will automatically Py_INCREF and Py_DECREF when appropriate. But for some
 # inner loops, we don't need to do that at all, as the reference only lasts for
 # a very short time.
+# Note that the C API GetItem calls borrow references, so pyrex does the wrong
+# thing if you declare e.g. object PyList_GetItem(object lst, int index) - you
+# need to manually Py_INCREF yourself.
 cdef extern from "Python.h":
     ctypedef int Py_ssize_t
+    ctypedef struct PyObject:
+        pass
     int PyList_Append(object lst, object item) except -1
     void *PyList_GetItem_object_void "PyList_GET_ITEM" (object lst, int index)
+    void *PyList_GetItem_void_void "PyList_GET_ITEM" (void * lst, int index)
+    object PyList_GET_ITEM(object lst, Py_ssize_t index)
     int PyList_CheckExact(object)
+    Py_ssize_t PyList_GET_SIZE (object p)
 
     void *PyTuple_GetItem_void_void "PyTuple_GET_ITEM" (void* tpl, int index)
+    object PyTuple_GetItem_void_object "PyTuple_GET_ITEM" (void* tpl, int index)
+    object PyTuple_GET_ITEM(object tpl, Py_ssize_t index)
+
 
     char *PyString_AsString(object p)
+    char *PyString_AsString_obj "PyString_AsString" (PyObject *string)
     char *PyString_AS_STRING_void "PyString_AS_STRING" (void *p)
+    int PyString_AsStringAndSize(object str, char **buffer, Py_ssize_t *length) except -1
     object PyString_FromString(char *)
     object PyString_FromStringAndSize(char *, Py_ssize_t)
     int PyString_Size(object p)
     int PyString_GET_SIZE_void "PyString_GET_SIZE" (void *p)
     int PyString_CheckExact(object p)
+    void Py_INCREF(object o)
+    void Py_DECREF(object o)
 
 
 cdef extern from "string.h":
@@ -118,6 +157,7 @@
 
 cdef object safe_string_from_size(char *s, Py_ssize_t size):
     if size < 0:
+        # XXX: On 64-bit machines the <int> cast causes a C compiler warning.
         raise AssertionError(
             'tried to create a string with an invalid size: %d @0x%x'
             % (size, <int>s))
@@ -723,3 +763,1011 @@
 
     reader._parse_dirblocks()
     state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED
+
+
+cdef int minikind_from_mode(int mode):
+    # in order of frequency:
+    if S_ISREG(mode):
+        return c"f"
+    if S_ISDIR(mode):
+        return c"d"
+    if S_ISLNK(mode):
+        return c"l"
+    return 0
+
+
+_encode = binascii.b2a_base64
+
+
+from struct import pack
+cdef _pack_stat(stat_value):
+    """return a string representing the stat value's key fields.
+
+    :param stat_value: A stat oject with st_size, st_mtime, st_ctime, st_dev,
+        st_ino and st_mode fields.
+    """
+    cdef char result[6*4] # 6 long ints
+    cdef int *aliased
+    aliased = <int *>result
+    aliased[0] = htonl(stat_value.st_size)
+    aliased[1] = htonl(int(stat_value.st_mtime))
+    aliased[2] = htonl(int(stat_value.st_ctime))
+    aliased[3] = htonl(stat_value.st_dev)
+    aliased[4] = htonl(stat_value.st_ino & 0xFFFFFFFF)
+    aliased[5] = htonl(stat_value.st_mode)
+    packed = PyString_FromStringAndSize(result, 6*4)
+    return _encode(packed)[:-1]
+
+
+def update_entry(self, entry, abspath, stat_value):
+    """Update the entry based on what is actually on disk.
+
+    This function only calculates the sha if it needs to - if the entry is
+    uncachable, or clearly different to the first parent's entry, no sha
+    is calculated, and None is returned.
+
+    :param entry: This is the dirblock entry for the file in question.
+    :param abspath: The path on disk for this file.
+    :param stat_value: (optional) if we already have done a stat on the
+        file, re-use it.
+    :return: None, or The sha1 hexdigest of the file (40 bytes) or link
+        target of a symlink.
+    """
+    return _update_entry(self, entry, abspath, stat_value)
+
+
+cdef _update_entry(self, entry, abspath, stat_value):
+    """Update the entry based on what is actually on disk.
+
+    This function only calculates the sha if it needs to - if the entry is
+    uncachable, or clearly different to the first parent's entry, no sha
+    is calculated, and None is returned.
+
+    :param self: The dirstate object this is operating on.
+    :param entry: This is the dirblock entry for the file in question.
+    :param abspath: The path on disk for this file.
+    :param stat_value: The stat value done on the path.
+    :return: None, or The sha1 hexdigest of the file (40 bytes) or link
+        target of a symlink.
+    """
+    # TODO - require pyrex 0.9.8, then use a pyd file to define access to the
+    # _st mode of the compiled stat objects.
+    cdef int minikind, saved_minikind
+    cdef void * details
+    minikind = minikind_from_mode(stat_value.st_mode)
+    if 0 == minikind:
+        return None
+    packed_stat = _pack_stat(stat_value)
+    details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
+    saved_minikind = PyString_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
+    saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
+    saved_file_size = PyTuple_GetItem_void_object(details, 2)
+    saved_executable = PyTuple_GetItem_void_object(details, 3)
+    saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
+    # Deal with pyrex decrefing the objects
+    Py_INCREF(saved_link_or_sha1)
+    Py_INCREF(saved_file_size)
+    Py_INCREF(saved_executable)
+    Py_INCREF(saved_packed_stat)
+    #(saved_minikind, saved_link_or_sha1, saved_file_size,
+    # saved_executable, saved_packed_stat) = entry[1][0]
+
+    if (minikind == saved_minikind
+        and packed_stat == saved_packed_stat):
+        # The stat hasn't changed since we saved, so we can re-use the
+        # saved sha hash.
+        if minikind == c'd':
+            return None
+
+        # size should also be in packed_stat
+        if saved_file_size == stat_value.st_size:
+            return saved_link_or_sha1
+
+    # If we have gotten this far, that means that we need to actually
+    # process this entry.
+    link_or_sha1 = None
+    if minikind == c'f':
+        executable = self._is_executable(stat_value.st_mode,
+                                         saved_executable)
+        if self._cutoff_time is None:
+            self._sha_cutoff_time()
+        if (stat_value.st_mtime < self._cutoff_time
+            and stat_value.st_ctime < self._cutoff_time
+            and len(entry[1]) > 1
+            and entry[1][1][0] != 'a'):
+                # Could check for size changes for further optimised
+                # avoidance of sha1's. However the most prominent case of
+                # over-shaing is during initial add, which this catches.
+            link_or_sha1 = self._sha1_file(abspath)
+            entry[1][0] = ('f', link_or_sha1, stat_value.st_size,
+                           executable, packed_stat)
+        else:
+            entry[1][0] = ('f', '', stat_value.st_size,
+                           executable, DirState.NULLSTAT)
+    elif minikind == c'd':
+        link_or_sha1 = None
+        entry[1][0] = ('d', '', 0, False, packed_stat)
+        if saved_minikind != c'd':
+            # This changed from something into a directory. Make sure we
+            # have a directory block for it. This doesn't happen very
+            # often, so this doesn't have to be super fast.
+            block_index, entry_index, dir_present, file_present = \
+                self._get_block_entry_index(entry[0][0], entry[0][1], 0)
+            self._ensure_block(block_index, entry_index,
+                               pathjoin(entry[0][0], entry[0][1]))
+    elif minikind == c'l':
+        link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
+        if self._cutoff_time is None:
+            self._sha_cutoff_time()
+        if (stat_value.st_mtime < self._cutoff_time
+            and stat_value.st_ctime < self._cutoff_time):
+            entry[1][0] = ('l', link_or_sha1, stat_value.st_size,
+                           False, packed_stat)
+        else:
+            entry[1][0] = ('l', '', stat_value.st_size,
+                           False, DirState.NULLSTAT)
+    self._dirblock_state = DirState.IN_MEMORY_MODIFIED
+    return link_or_sha1
+
+
+cdef char _minikind_from_string(object string):
+    """Convert a python string to a char."""
+    return PyString_AsString(string)[0]
+
+
+cdef object _kind_absent
+cdef object _kind_file
+cdef object _kind_directory
+cdef object _kind_symlink
+cdef object _kind_relocated
+cdef object _kind_tree_reference
+_kind_absent = "absent"
+_kind_file = "file"
+_kind_directory = "directory"
+_kind_symlink = "symlink"
+_kind_relocated = "relocated"
+_kind_tree_reference = "tree-reference"
+
+
+cdef object _minikind_to_kind(char minikind):
+    """Create a string kind for minikind."""
+    cdef char _minikind[1]
+    if minikind == c'f':
+        return _kind_file
+    elif minikind == c'd':
+        return _kind_directory
+    elif minikind == c'a':
+        return _kind_absent
+    elif minikind == c'r':
+        return _kind_relocated
+    elif minikind == c'l':
+        return _kind_symlink
+    elif minikind == c't':
+        return _kind_tree_reference
+    _minikind[0] = minikind
+    raise KeyError(PyString_FromStringAndSize(_minikind, 1))
+
+
+cdef int _versioned_minikind(char minikind):
+    """Return non-zero if minikind is in fltd"""
+    return (minikind == c'f' or
+            minikind == c'd' or
+            minikind == c'l' or
+            minikind == c't')
+
+
+cdef class ProcessEntryC:
+
+    cdef object old_dirname_to_file_id # dict
+    cdef object new_dirname_to_file_id # dict
+    cdef readonly object uninteresting
+    cdef object last_source_parent
+    cdef object last_target_parent
+    cdef object include_unchanged
+    cdef object use_filesystem_for_exec
+    cdef object utf8_decode
+    cdef readonly object searched_specific_files
+    cdef object search_specific_files
+    cdef object state
+    # Current iteration variables:
+    cdef object current_root
+    cdef object current_root_unicode
+    cdef object root_entries
+    cdef int root_entries_pos, root_entries_len
+    cdef object root_abspath
+    cdef int source_index, target_index
+    cdef int want_unversioned
+    cdef object tree
+    cdef object dir_iterator
+    cdef int block_index
+    cdef object current_block
+    cdef int current_block_pos
+    cdef object current_block_list
+    cdef object current_dir_info
+    cdef object current_dir_list
+    cdef int path_index
+    cdef object root_dir_info
+    cdef object bisect_left
+    cdef object pathjoin
+    cdef object fstat
+    cdef object sha_file
+
+    def __init__(self, include_unchanged, use_filesystem_for_exec,
+        search_specific_files, state, source_index, target_index,
+        want_unversioned, tree):
+        self.old_dirname_to_file_id = {}
+        self.new_dirname_to_file_id = {}
+        # Just a sentry, so that _process_entry can say that this
+        # record is handled, but isn't interesting to process (unchanged)
+        self.uninteresting = object()
+        # Using a list so that we can access the values and change them in
+        # nested scope. Each one is [path, file_id, entry]
+        self.last_source_parent = [None, None]
+        self.last_target_parent = [None, None]
+        self.include_unchanged = include_unchanged
+        self.use_filesystem_for_exec = use_filesystem_for_exec
+        self.utf8_decode = cache_utf8._utf8_decode
+        # for all search_indexs in each path at or under each element of
+        # search_specific_files, if the detail is relocated: add the id, and add the
+        # relocated path as one to search if its not searched already. If the
+        # detail is not relocated, add the id.
+        self.searched_specific_files = set()
+        self.search_specific_files = search_specific_files
+        self.state = state
+        self.current_root = None
+        self.current_root_unicode = None
+        self.root_entries = None
+        self.root_entries_pos = 0
+        self.root_entries_len = 0
+        self.root_abspath = None
+        if source_index is None:
+            self.source_index = -1
+        else:
+            self.source_index = source_index
+        self.target_index = target_index
+        self.want_unversioned = want_unversioned
+        self.tree = tree
+        self.dir_iterator = None
+        self.block_index = -1
+        self.current_block = None
+        self.current_block_list = None
+        self.current_block_pos = -1
+        self.current_dir_info = None
+        self.current_dir_list = None
+        self.path_index = 0
+        self.root_dir_info = None
+        self.bisect_left = bisect.bisect_left
+        self.pathjoin = osutils.pathjoin
+        self.fstat = os.fstat
+        self.sha_file = osutils.sha_file
+
+    cdef _process_entry(self, entry, path_info):
+        """Compare an entry and real disk to generate delta information.
+
+        :param path_info: top_relpath, basename, kind, lstat, abspath for
+            the path of entry. If None, then the path is considered absent.
+            (Perhaps we should pass in a concrete entry for this ?)
+            Basename is returned as a utf8 string because we expect this
+            tuple will be ignored, and don't want to take the time to
+            decode.
+        :return: None if the these don't match
+                 A tuple of information about the change, or
+                 the object 'uninteresting' if these match, but are
+                 basically identical.
+        """
+        cdef char target_minikind
+        cdef char source_minikind
+        cdef object file_id
+        cdef int content_change
+        cdef object details_list
+        file_id = None
+        details_list = entry[1]
+        if -1 == self.source_index:
+            source_details = DirState.NULL_PARENT_DETAILS
+        else:
+            source_details = details_list[self.source_index]
+        target_details = details_list[self.target_index]
+        target_minikind = _minikind_from_string(target_details[0])
+        if path_info is not None and _versioned_minikind(target_minikind):
+            if self.target_index != 0:
+                raise AssertionError("Unsupported target index %d" % target_index)
+            link_or_sha1 = _update_entry(self.state, entry, path_info[4], path_info[3])
+            # The entry may have been modified by update_entry
+            target_details = details_list[self.target_index]
+            target_minikind = _minikind_from_string(target_details[0])
+        else:
+            link_or_sha1 = None
+        # the rest of this function is 0.3 seconds on 50K paths, or
+        # 0.000006 seconds per call.
+        source_minikind = _minikind_from_string(source_details[0])
+        if ((_versioned_minikind(source_minikind) or source_minikind == c'r')
+            and _versioned_minikind(target_minikind)):
+            # claimed content in both: diff
+            #   r    | fdlt   |      | add source to search, add id path move and perform
+            #        |        |      | diff check on source-target
+            #   r    | fdlt   |  a   | dangling file that was present in the basis.
+            #        |        |      | ???
+            if source_minikind != c'r':
+                old_dirname = entry[0][0]
+                old_basename = entry[0][1]
+                old_path = path = None
+            else:
+                # add the source to the search path to find any children it
+                # has.  TODO ? : only add if it is a container ?
+                if not osutils.is_inside_any(self.searched_specific_files,
+                                             source_details[1]):
+                    self.search_specific_files.add(source_details[1])
+                # generate the old path; this is needed for stating later
+                # as well.
+                old_path = source_details[1]
+                old_dirname, old_basename = os.path.split(old_path)
+                path = self.pathjoin(entry[0][0], entry[0][1])
+                old_entry = self.state._get_entry(self.source_index,
+                                             path_utf8=old_path)
+                # update the source details variable to be the real
+                # location.
+                if old_entry == (None, None):
+                    raise errors.CorruptDirstate(self.state._filename,
+                        "entry '%s/%s' is considered renamed from %r"
+                        " but source does not exist\n"
+                        "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
+                source_details = old_entry[1][self.source_index]
+                source_minikind = _minikind_from_string(source_details[0])
+            if path_info is None:
+                # the file is missing on disk, show as removed.
+                content_change = 1
+                target_kind = None
+                target_exec = False
+            else:
+                # source and target are both versioned and disk file is present.
+                target_kind = path_info[2]
+                if target_kind == 'directory':
+                    if path is None:
+                        old_path = path = self.pathjoin(old_dirname, old_basename)
+                    file_id = entry[0][2]
+                    self.new_dirname_to_file_id[path] = file_id
+                    if source_minikind != c'd':
+                        content_change = 1
+                    else:
+                        # directories have no fingerprint
+                        content_change = 0
+                    target_exec = False
+                elif target_kind == 'file':
+                    if source_minikind != c'f':
+                        content_change = 1
+                    else:
+                        # If the size is the same, check the sha:
+                        if target_details[2] == source_details[2]:
+                            if link_or_sha1 is None:
+                                # Stat cache miss:
+                                file_obj = file(path_info[4], 'rb')
+                                try:
+                                    # XXX: TODO: Use lower level file IO rather
+                                    # than python objects for sha-misses.
+                                    statvalue = self.fstat(file_obj.fileno())
+                                    link_or_sha1 = self.sha_file(file_obj)
+                                finally:
+                                    file_obj.close()
+                                self.state._observed_sha1(entry, link_or_sha1,
+                                    statvalue)
+                            content_change = (link_or_sha1 != source_details[1])
+                        else:
+                            # Size changed, so must be different
+                            content_change = 1
+                    # Target details is updated at update_entry time
+                    if self.use_filesystem_for_exec:
+                        # We don't need S_ISREG here, because we are sure
+                        # we are dealing with a file.
+                        target_exec = bool(S_IXUSR & path_info[3].st_mode)
+                    else:
+                        target_exec = target_details[3]
+                elif target_kind == 'symlink':
+                    if source_minikind != c'l':
+                        content_change = 1
+                    else:
+                        content_change = (link_or_sha1 != source_details[1])
+                    target_exec = False
+                elif target_kind == 'tree-reference':
+                    if source_minikind != c't':
+                        content_change = 1
+                    else:
+                        content_change = 0
+                    target_exec = False
+                else:
+                    raise Exception, "unknown kind %s" % path_info[2]
+            if source_minikind == c'd':
+                if path is None:
+                    old_path = path = self.pathjoin(old_dirname, old_basename)
+                if file_id is None:
+                    file_id = entry[0][2]
+                self.old_dirname_to_file_id[old_path] = file_id
+            # parent id is the entry for the path in the target tree
+            if old_dirname == self.last_source_parent[0]:
+                source_parent_id = self.last_source_parent[1]
+            else:
+                try:
+                    source_parent_id = self.old_dirname_to_file_id[old_dirname]
+                except KeyError:
+                    source_parent_entry = self.state._get_entry(self.source_index,
+                                                           path_utf8=old_dirname)
+                    source_parent_id = source_parent_entry[0][2]
+                if source_parent_id == entry[0][2]:
+                    # This is the root, so the parent is None
+                    source_parent_id = None
+                else:
+                    self.last_source_parent[0] = old_dirname
+                    self.last_source_parent[1] = source_parent_id
+            new_dirname = entry[0][0]
+            if new_dirname == self.last_target_parent[0]:
+                target_parent_id = self.last_target_parent[1]
+            else:
+                try:
+                    target_parent_id = self.new_dirname_to_file_id[new_dirname]
+                except KeyError:
+                    # TODO: We don't always need to do the lookup, because the
+                    #       parent entry will be the same as the source entry.
+                    target_parent_entry = self.state._get_entry(self.target_index,
+                                                           path_utf8=new_dirname)
+                    if target_parent_entry == (None, None):
+                        raise AssertionError(
+                            "Could not find target parent in wt: %s\nparent of: %s"
+                            % (new_dirname, entry))
+                    target_parent_id = target_parent_entry[0][2]
+                if target_parent_id == entry[0][2]:
+                    # This is the root, so the parent is None
+                    target_parent_id = None
+                else:
+                    self.last_target_parent[0] = new_dirname
+                    self.last_target_parent[1] = target_parent_id
+
+            source_exec = source_details[3]
+            if (self.include_unchanged
+                or content_change
+                or source_parent_id != target_parent_id
+                or old_basename != entry[0][1]
+                or source_exec != target_exec
+                ):
+                if old_path is None:
+                    path = self.pathjoin(old_dirname, old_basename)
+                    old_path = path
+                    old_path_u = self.utf8_decode(old_path)[0]
+                    path_u = old_path_u
+                else:
+                    old_path_u = self.utf8_decode(old_path)[0]
+                    if old_path == path:
+                        path_u = old_path_u
+                    else:
+                        path_u = self.utf8_decode(path)[0]
+                source_kind = _minikind_to_kind(source_minikind)
+                return (entry[0][2],
+                       (old_path_u, path_u),
+                       content_change,
+                       (True, True),
+                       (source_parent_id, target_parent_id),
+                       (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),
+                       (source_kind, target_kind),
+                       (source_exec, target_exec))
+            else:
+                return self.uninteresting
+        elif source_minikind == c'a' and _versioned_minikind(target_minikind):
+            # looks like a new file
+            path = self.pathjoin(entry[0][0], entry[0][1])
+            # parent id is the entry for the path in the target tree
+            # TODO: these are the same for an entire directory: cache em.
+            parent_id = self.state._get_entry(self.target_index,
+                                         path_utf8=entry[0][0])[0][2]
+            if parent_id == entry[0][2]:
+                parent_id = None
+            if path_info is not None:
+                # Present on disk:
+                if self.use_filesystem_for_exec:
+                    # We need S_ISREG here, because we aren't sure if this
+                    # is a file or not.
+                    target_exec = bool(
+                        S_ISREG(path_info[3].st_mode)
+                        and S_IXUSR & path_info[3].st_mode)
+                else:
+                    target_exec = target_details[3]
+                return (entry[0][2],
+                       (None, self.utf8_decode(path)[0]),
+                       True,
+                       (False, True),
+                       (None, parent_id),
+                       (None, self.utf8_decode(entry[0][1])[0]),
+                       (None, path_info[2]),
+                       (None, target_exec))
+            else:
+                # Its a missing file, report it as such.
+                return (entry[0][2],
+                       (None, self.utf8_decode(path)[0]),
+                       False,
+                       (False, True),
+                       (None, parent_id),
+                       (None, self.utf8_decode(entry[0][1])[0]),
+                       (None, None),
+                       (None, False))
+        elif _versioned_minikind(source_minikind) and target_minikind == c'a':
+            # unversioned, possibly, or possibly not deleted: we dont care.
+            # if its still on disk, *and* theres no other entry at this
+            # path [we dont know this in this routine at the moment -
+            # perhaps we should change this - then it would be an unknown.
+            old_path = self.pathjoin(entry[0][0], entry[0][1])
+            # parent id is the entry for the path in the target tree
+            parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
+            if parent_id == entry[0][2]:
+                parent_id = None
+            return (entry[0][2],
+                   (self.utf8_decode(old_path)[0], None),
+                   True,
+                   (True, False),
+                   (parent_id, None),
+                   (self.utf8_decode(entry[0][1])[0], None),
+                   (_minikind_to_kind(source_minikind), None),
+                   (source_details[3], None))
+        elif _versioned_minikind(source_minikind) and target_minikind == c'r':
+            # a rename; could be a true rename, or a rename inherited from
+            # a renamed parent. TODO: handle this efficiently. Its not
+            # common case to rename dirs though, so a correct but slow
+            # implementation will do.
+            if not osutils.is_inside_any(self.searched_specific_files, target_details[1]):
+                self.search_specific_files.add(target_details[1])
+        elif ((source_minikind == c'r' or source_minikind == c'a') and
+              (target_minikind == c'r' or target_minikind == c'a')):
+            # neither of the selected trees contain this path,
+            # so skip over it. This is not currently directly tested, but
+            # is indirectly via test_too_much.TestCommands.test_conflicts.
+            pass
+        else:
+            raise AssertionError("don't know how to compare "
+                "source_minikind=%r, target_minikind=%r"
+                % (source_minikind, target_minikind))
+            ## import pdb;pdb.set_trace()
+        return None
+
+    def __iter__(self):
+        return self
+
+    def iter_changes(self):
+        return self
+
+    cdef void _update_current_block(self):
+        if (self.block_index < len(self.state._dirblocks) and
+            osutils.is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
+            self.current_block = self.state._dirblocks[self.block_index]
+            self.current_block_list = self.current_block[1]
+            self.current_block_pos = 0
+        else:
+            self.current_block = None
+            self.current_block_list = None
+
+    def __next__(self):
+        # Simple thunk to allow tail recursion without pyrex confusion
+        return self._iter_next()
+
+    cdef _iter_next(self):
+        """Iterate over the changes."""
+        # This function single steps through an iterator. As such while loops
+        # are often exited by 'return' - the code is structured so that the
+        # next call into the function will return to the same while loop. Note
+        # that all flow control needed to re-reach that step is reexecuted,
+        # which can be a performance problem. It has not yet been tuned to
+        # minimise this; a state machine is probably the simplest restructuring
+        # to both minimise this overhead and make the code considerably more
+        # understandable.
+
+        # sketch: 
+        # compare source_index and target_index at or under each element of search_specific_files.
+        # follow the following comparison table. Note that we only want to do diff operations when
+        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
+        # for the target.
+        # cases:
+        # 
+        # Source | Target | disk | action
+        #   r    | fdlt   |      | add source to search, add id path move and perform
+        #        |        |      | diff check on source-target
+        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
+        #        |        |      | ???
+        #   r    |  a     |      | add source to search
+        #   r    |  a     |  a   | 
+        #   r    |  r     |      | this path is present in a non-examined tree, skip.
+        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
+        #   a    | fdlt   |      | add new id
+        #   a    | fdlt   |  a   | dangling locally added file, skip
+        #   a    |  a     |      | not present in either tree, skip
+        #   a    |  a     |  a   | not present in any tree, skip
+        #   a    |  r     |      | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #   a    |  r     |  a   | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #  fdlt  | fdlt   |      | content in both: diff them
+        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
+        #  fdlt  |  a     |      | unversioned: output deleted id for now
+        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
+        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+
+        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
+        #       keeping a cache of directories that we have seen.
+        cdef object current_dirname, current_blockname
+        cdef char * current_dirname_c, * current_blockname_c
+        cdef int advance_entry, advance_path
+        cdef int path_handled
+        uninteresting = self.uninteresting
+        searched_specific_files = self.searched_specific_files
+        # Are we walking a root?
+        while self.root_entries_pos < self.root_entries_len:
+            entry = self.root_entries[self.root_entries_pos]
+            self.root_entries_pos = self.root_entries_pos + 1
+            result = self._process_entry(entry, self.root_dir_info)
+            if result is not None and result is not self.uninteresting:
+                return result
+        # Have we finished the prior root, or never started one ?
+        if self.current_root is None:
+            # TODO: the pending list should be lexically sorted?  the
+            # interface doesn't require it.
+            try:
+                self.current_root = self.search_specific_files.pop()
+            except KeyError:
+                raise StopIteration()
+            self.current_root_unicode = self.current_root.decode('utf8')
+            self.searched_specific_files.add(self.current_root)
+            # process the entries for this containing directory: the rest will be
+            # found by their parents recursively.
+            self.root_entries = self.state._entries_for_path(self.current_root)
+            self.root_entries_len = len(self.root_entries)
+            self.root_abspath = self.tree.abspath(self.current_root_unicode)
+            try:
+                root_stat = os.lstat(self.root_abspath)
+            except OSError, e:
+                if e.errno == errno.ENOENT:
+                    # the path does not exist: let _process_entry know that.
+                    self.root_dir_info = None
+                else:
+                    # some other random error: hand it up.
+                    raise
+            else:
+                self.root_dir_info = ('', self.current_root,
+                    osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
+                    self.root_abspath)
+                if self.root_dir_info[2] == 'directory':
+                    if self.tree._directory_is_tree_reference(
+                        self.current_root_unicode):
+                        self.root_dir_info = self.root_dir_info[:2] + \
+                            ('tree-reference',) + self.root_dir_info[3:]
+            if not self.root_entries and not self.root_dir_info:
+                # this specified path is not present at all, skip it.
+                # (tail recursion, can do a loop once the full structure is
+                # known).
+                return self._iter_next()
+            path_handled = 0
+            self.root_entries_pos = 0
+            # XXX Clarity: This loop is duplicated a out the self.current_root
+            # is None guard above: if we return from it, it completes there
+            # (and the following if block cannot trigger because
+            # path_handled must be true, so the if block is not # duplicated.
+            while self.root_entries_pos < self.root_entries_len:
+                entry = self.root_entries[self.root_entries_pos]
+                self.root_entries_pos = self.root_entries_pos + 1
+                result = self._process_entry(entry, self.root_dir_info)
+                if result is not None:
+                    path_handled = -1
+                    if result is not self.uninteresting:
+                        return result
+            # handle unversioned specified paths:
+            if self.want_unversioned and not path_handled and self.root_dir_info:
+                new_executable = bool(
+                    stat.S_ISREG(self.root_dir_info[3].st_mode)
+                    and stat.S_IEXEC & self.root_dir_info[3].st_mode)
+                return (None,
+                       (None, self.current_root_unicode),
+                       True,
+                       (False, False),
+                       (None, None),
+                       (None, splitpath(self.current_root_unicode)[-1]),
+                       (None, self.root_dir_info[2]),
+                       (None, new_executable)
+                      )
+            # If we reach here, the outer flow continues, which enters into the
+            # per-root setup logic.
+        if self.current_dir_info is None and self.current_block is None:
+            # setup iteration of this root:
+            self.current_dir_list = None
+            if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
+                self.current_dir_info = None
+            else:
+                self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
+                    prefix=self.current_root)
+                self.path_index = 0
+                try:
+                    self.current_dir_info = self.dir_iterator.next()
+                    self.current_dir_list = self.current_dir_info[1]
+                except OSError, e:
+                    # there may be directories in the inventory even though
+                    # this path is not a file on disk: so mark it as end of
+                    # iterator
+                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
+                        self.current_dir_info = None
+                    elif sys.platform == 'win32':
+                        # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
+                        # python 2.5 has e.errno == EINVAL,
+                        #            and e.winerror == ERROR_DIRECTORY
+                        try:
+                            e_winerror = e.winerror
+                        except AttributeError:
+                            e_winerror = None
+                        win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
+                        if (e.errno in win_errors or e_winerror in win_errors):
+                            self.current_dir_info = None
+                        else:
+                            # Will this really raise the right exception ?
+                            raise
+                    else:
+                        raise
+                else:
+                    if self.current_dir_info[0][0] == '':
+                        # remove .bzr from iteration
+                        bzr_index = self.bisect_left(self.current_dir_list, ('.bzr',))
+                        if self.current_dir_list[bzr_index][0] != '.bzr':
+                            raise AssertionError()
+                        del self.current_dir_list[bzr_index]
+            initial_key = (self.current_root, '', '')
+            self.block_index, _ = self.state._find_block_index_from_key(initial_key)
+            if self.block_index == 0:
+                # we have processed the total root already, but because the
+                # initial key matched it we should skip it here.
+                self.block_index = self.block_index + 1
+            self._update_current_block()
+        # walk until both the directory listing and the versioned metadata
+        # are exhausted. 
+        while (self.current_dir_info is not None
+            or self.current_block is not None):
+            # Uncommon case - a missing directory or an unversioned directory:
+            if (self.current_dir_info and self.current_block
+                and self.current_dir_info[0][0] != self.current_block[0]):
+                # Work around pyrex broken heuristic - current_dirname has
+                # the same scope as current_dirname_c
+                current_dirname = self.current_dir_info[0][0]
+                current_dirname_c = PyString_AS_STRING_void(
+                    <void *>current_dirname)
+                current_blockname = self.current_block[0]
+                current_blockname_c = PyString_AS_STRING_void(
+                    <void *>current_blockname)
+                # In the python generator we evaluate this if block once per
+                # dir+block; because we reenter in the pyrex version its being
+                # evaluated once per path: we could cache the result before
+                # doing the while loop and probably save time.
+                if _cmp_by_dirs(current_dirname_c,
+                    PyString_Size(current_dirname),
+                    current_blockname_c,
+                    PyString_Size(current_blockname)) < 0:
+                    # filesystem data refers to paths not covered by the
+                    # dirblock.  this has two possibilities:
+                    # A) it is versioned but empty, so there is no block for it
+                    # B) it is not versioned.
+
+                    # if (A) then we need to recurse into it to check for
+                    # new unknown files or directories.
+                    # if (B) then we should ignore it, because we don't
+                    # recurse into unknown directories.
+                    # We are doing a loop
+                    while self.path_index < len(self.current_dir_list):
+                        current_path_info = self.current_dir_list[self.path_index]
+                        # dont descend into this unversioned path if it is
+                        # a dir
+                        if current_path_info[2] in ('directory',
+                                                    'tree-reference'):
+                            del self.current_dir_list[self.path_index]
+                            self.path_index = self.path_index - 1
+                        self.path_index = self.path_index + 1
+                        if self.want_unversioned:
+                            if current_path_info[2] == 'directory':
+                                if self.tree._directory_is_tree_reference(
+                                    self.utf8_decode(current_path_info[0])[0]):
+                                    current_path_info = current_path_info[:2] + \
+                                        ('tree-reference',) + current_path_info[3:]
+                            new_executable = bool(
+                                stat.S_ISREG(current_path_info[3].st_mode)
+                                and stat.S_IEXEC & current_path_info[3].st_mode)
+                            return (None,
+                                (None, self.utf8_decode(current_path_info[0])[0]),
+                                True,
+                                (False, False),
+                                (None, None),
+                                (None, self.utf8_decode(current_path_info[1])[0]),
+                                (None, current_path_info[2]),
+                                (None, new_executable))
+                    # This dir info has been handled, go to the next
+                    self.path_index = 0
+                    self.current_dir_list = None
+                    try:
+                        self.current_dir_info = self.dir_iterator.next()
+                        self.current_dir_list = self.current_dir_info[1]
+                    except StopIteration:
+                        self.current_dir_info = None
+                else: #(dircmp > 0)
+                    # We have a dirblock entry for this location, but there
+                    # is no filesystem path for this. This is most likely
+                    # because a directory was removed from the disk.
+                    # We don't have to report the missing directory,
+                    # because that should have already been handled, but we
+                    # need to handle all of the files that are contained
+                    # within.
+                    while self.current_block_pos < len(self.current_block_list):
+                        current_entry = self.current_block_list[self.current_block_pos]
+                        self.current_block_pos = self.current_block_pos + 1
+                        # entry referring to file not present on disk.
+                        # advance the entry only, after processing.
+                        result = self._process_entry(current_entry, None)
+                        if result is not None:
+                            if result is not self.uninteresting:
+                                return result
+                    self.block_index = self.block_index + 1
+                    self._update_current_block()
+                continue # next loop-on-block/dir
+            result = self._loop_one_block()
+            if result is not None:
+                return result
+        if len(self.search_specific_files):
+            # More supplied paths to process
+            self.current_root = None
+            return self._iter_next()
+        raise StopIteration()
+
+    cdef object _maybe_tree_ref(self, current_path_info):
+        if self.tree._directory_is_tree_reference(
+            self.utf8_decode(current_path_info[0])[0]):
+            return current_path_info[:2] + \
+                ('tree-reference',) + current_path_info[3:]
+        else:
+            return current_path_info
+
+    cdef object _loop_one_block(self):
+            # current_dir_info and current_block refer to the same directory -
+            # this is the common case code.
+            # Assign local variables for current path and entry:
+            cdef object current_entry
+            cdef object current_path_info
+            cdef int path_handled
+            cdef char minikind
+            cdef int cmp_result
+            # cdef char * temp_str
+            # cdef Py_ssize_t temp_str_length
+            # PyString_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
+            # if not strncmp(temp_str, "directory", temp_str_length):
+            if (self.current_block is not None and
+                self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
+                current_entry = PyList_GET_ITEM(self.current_block_list,
+                    self.current_block_pos)
+                # accomodate pyrex
+                Py_INCREF(current_entry)
+            else:
+                current_entry = None
+            if (self.current_dir_info is not None and
+                self.path_index < PyList_GET_SIZE(self.current_dir_list)):
+                current_path_info = PyList_GET_ITEM(self.current_dir_list,
+                    self.path_index)
+                # accomodate pyrex
+                Py_INCREF(current_path_info)
+                disk_kind = PyTuple_GET_ITEM(current_path_info, 2)
+                # accomodate pyrex
+                Py_INCREF(disk_kind)
+                if disk_kind == "directory":
+                    current_path_info = self._maybe_tree_ref(current_path_info)
+            else:
+                current_path_info = None
+            while (current_entry is not None or current_path_info is not None):
+                advance_entry = -1
+                advance_path = -1
+                result = None
+                path_handled = 0
+                if current_entry is None:
+                    # unversioned -  the check for path_handled when the path
+                    # is advanced will yield this path if needed.
+                    pass
+                elif current_path_info is None:
+                    # no path is fine: the per entry code will handle it.
+                    result = self._process_entry(current_entry, current_path_info)
+                    if result is not None:
+                        if result is self.uninteresting:
+                            result = None
+                else:
+                    minikind = _minikind_from_string(
+                        current_entry[1][self.target_index][0])
+                    cmp_result = cmp(current_path_info[1], current_entry[0][1])
+                    if (cmp_result or minikind == c'a' or minikind == c'r'):
+                        # The current path on disk doesn't match the dirblock
+                        # record. Either the dirblock record is marked as
+                        # absent/renamed, or the file on disk is not present at all
+                        # in the dirblock. Either way, report about the dirblock
+                        # entry, and let other code handle the filesystem one.
+
+                        # Compare the basename for these files to determine
+                        # which comes first
+                        if cmp_result < 0:
+                            # extra file on disk: pass for now, but only
+                            # increment the path, not the entry
+                            advance_entry = 0
+                        else:
+                            # entry referring to file not present on disk.
+                            # advance the entry only, after processing.
+                            result = self._process_entry(current_entry, None)
+                            if result is not None:
+                                if result is self.uninteresting:
+                                    result = None
+                            advance_path = 0
+                    else:
+                        # paths are the same,and the dirstate entry is not
+                        # absent or renamed.
+                        result = self._process_entry(current_entry, current_path_info)
+                        if result is not None:
+                            path_handled = -1
+                            if result is self.uninteresting:
+                                result = None
+                # >- loop control starts here:
+                # >- entry
+                if advance_entry and current_entry is not None:
+                    self.current_block_pos = self.current_block_pos + 1
+                    if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
+                        current_entry = self.current_block_list[self.current_block_pos]
+                    else:
+                        current_entry = None
+                # >- path
+                if advance_path and current_path_info is not None:
+                    if not path_handled:
+                        # unversioned in all regards
+                        if self.want_unversioned:
+                            new_executable = bool(
+                                stat.S_ISREG(current_path_info[3].st_mode)
+                                and stat.S_IEXEC & current_path_info[3].st_mode)
+                            try:
+                                relpath_unicode = self.utf8_decode(current_path_info[0])[0]
+                            except UnicodeDecodeError:
+                                raise errors.BadFilenameEncoding(
+                                    current_path_info[0], osutils._fs_enc)
+                            if result is not None:
+                                raise AssertionError(
+                                    "result is not None: %r" % result)
+                            result = (None,
+                                (None, relpath_unicode),
+                                True,
+                                (False, False),
+                                (None, None),
+                                (None, self.utf8_decode(current_path_info[1])[0]),
+                                (None, current_path_info[2]),
+                                (None, new_executable))
+                        # dont descend into this unversioned path if it is
+                        # a dir
+                        if current_path_info[2] in ('directory'):
+                            del self.current_dir_list[self.path_index]
+                            self.path_index = self.path_index - 1
+                    # dont descend the disk iterator into any tree 
+                    # paths.
+                    if current_path_info[2] == 'tree-reference':
+                        del self.current_dir_list[self.path_index]
+                        self.path_index = self.path_index - 1
+                    self.path_index = self.path_index + 1
+                    if self.path_index < len(self.current_dir_list):
+                        current_path_info = self.current_dir_list[self.path_index]
+                        if current_path_info[2] == 'directory':
+                            current_path_info = self._maybe_tree_ref(
+                                current_path_info)
+                    else:
+                        current_path_info = None
+                if result is not None:
+                    # Found a result on this pass, yield it
+                    return result
+            if self.current_block is not None:
+                self.block_index = self.block_index + 1
+                self._update_current_block()
+            if self.current_dir_info is not None:
+                self.path_index = 0
+                self.current_dir_list = None
+                try:
+                    self.current_dir_info = self.dir_iterator.next()
+                    self.current_dir_list = self.current_dir_info[1]
+                except StopIteration:
+                    self.current_dir_info = None

=== modified file 'bzrlib/_dirstate_helpers_py.py'
--- bzrlib/_dirstate_helpers_py.py	2008-09-02 18:51:03 +0000
+++ bzrlib/_dirstate_helpers_py.py	2008-09-24 03:08:14 +0000
@@ -289,4 +289,3 @@
     # To convert from format 3 => format 2
     # state._dirblocks = sorted(state._dirblocks)
     state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED
-

=== modified file 'bzrlib/_patiencediff_c.c'
--- bzrlib/_patiencediff_c.c	2008-08-17 19:39:54 +0000
+++ bzrlib/_patiencediff_c.c	2008-09-25 22:08:41 +0000
@@ -27,13 +27,7 @@
 #include <string.h>
 #include <Python.h>
 
-
-/* http://www.python.org/dev/peps/pep-0353/ */
-#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
-typedef int Py_ssize_t;
-#define PY_SSIZE_T_MAX INT_MAX
-#define PY_SSIZE_T_MIN INT_MIN
-#endif
+#include "python-compat.h"
 
 
 #if defined(__GNUC__)

=== removed file 'bzrlib/_walkdirs_win32.h'
--- bzrlib/_walkdirs_win32.h	2008-07-17 03:46:13 +0000
+++ bzrlib/_walkdirs_win32.h	1970-01-01 00:00:00 +0000
@@ -1,25 +0,0 @@
-/*
- * Copyright (C) 2007 Canonical Ltd
- * 
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- * 
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- * 
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- */
-
-/* Header includes, etc for _walkdirs_win32
- * Pyrex doesn't support #define, and defining WIN32_LEAN_AND_MEAN makes
- * importing windows.h a lot less painful.
- */
-
-#define WIN32_LEAN_AND_MEAN
-#include <windows.h>

=== modified file 'bzrlib/_walkdirs_win32.pyx'
--- bzrlib/_walkdirs_win32.pyx	2008-09-23 23:28:27 +0000
+++ bzrlib/_walkdirs_win32.pyx	2008-09-26 05:14:51 +0000
@@ -17,7 +17,7 @@
 """Helper functions for Walkdirs on win32."""
 
 
-cdef extern from "_walkdirs_win32.h":
+cdef extern from "python-compat.h":
     struct _HANDLE:
         pass
     ctypedef _HANDLE *HANDLE

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2008-09-23 02:16:55 +0000
+++ bzrlib/branch.py	2008-09-25 22:25:09 +0000
@@ -36,7 +36,7 @@
         urlutils,
         )
 from bzrlib.config import BranchConfig
-from bzrlib.repofmt.pack_repo import RepositoryFormatPackDevelopment1Subtree
+from bzrlib.repofmt.pack_repo import RepositoryFormatKnitPack5RichRoot
 from bzrlib.tag import (
     BasicTags,
     DisabledTags,
@@ -1346,7 +1346,7 @@
     def __init__(self):
         super(BzrBranchFormat7, self).__init__()
         self._matchingbzrdir.repository_format = \
-            RepositoryFormatPackDevelopment1Subtree()
+            RepositoryFormatKnitPack5RichRoot()
 
     def supports_stacking(self):
         return True

=== modified file 'bzrlib/branchbuilder.py'
--- bzrlib/branchbuilder.py	2008-07-29 16:16:36 +0000
+++ bzrlib/branchbuilder.py	2008-07-29 16:34:09 +0000
@@ -129,7 +129,7 @@
             ('add', ('path', 'file-id', 'kind', 'content' or None))
             ('modify', ('file-id', 'new-content'))
             ('unversion', 'file-id')
-            # not supported yet: ('rename', ('orig-path', 'new-path'))
+            ('rename', ('orig-path', 'new-path'))
         :param message: An optional commit message, if not supplied, a default
             commit message will be written.
         :return: The revision_id of the new commit
@@ -157,10 +157,7 @@
             to_add_kinds = []
             new_contents = {}
             to_unversion_ids = []
-            # TODO: MemoryTree doesn't support rename() or
-            #       apply_inventory_delta, so we'll postpone allowing renames
-            #       for now
-            # to_rename = []
+            to_rename = []
             for action, info in actions:
                 if action == 'add':
                     path, file_id, kind, content = info
@@ -177,6 +174,9 @@
                     new_contents[file_id] = content
                 elif action == 'unversion':
                     to_unversion_ids.append(info)
+                elif action == 'rename':
+                    from_relpath, to_relpath = info
+                    to_rename.append((from_relpath, to_relpath))
                 else:
                     raise ValueError('Unknown build action: "%s"' % (action,))
             if to_unversion_ids:
@@ -187,6 +187,8 @@
                     tree.add([path], [file_id], ['directory'])
                 else:
                     tree.mkdir(path, file_id)
+            for from_relpath, to_relpath in to_rename:
+                tree.rename_one(from_relpath, to_relpath)
             tree.add(to_add_files, to_add_file_ids, to_add_kinds)
             for file_id, content in new_contents.iteritems():
                 tree.put_file_bytes_non_atomic(file_id, content)

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2008-09-24 06:52:03 +0000
+++ bzrlib/builtins.py	2008-09-25 22:05:46 +0000
@@ -1689,6 +1689,11 @@
                    help='Show files changed in each revision.'),
             'show-ids',
             'revision',
+            Option('change',
+                   type=bzrlib.option._parse_revision_str,
+                   short_name='c',
+                   help='Show just the specified revision.'
+                   ' See also "help revisionspec".'),
             'log-format',
             Option('message',
                    short_name='m',
@@ -1709,12 +1714,22 @@
             show_ids=False,
             forward=False,
             revision=None,
+            change=None,
             log_format=None,
             message=None,
             limit=None):
         from bzrlib.log import show_log
         direction = (forward and 'forward') or 'reverse'
-        
+
+        if change is not None:
+            if len(change) > 1:
+                raise errors.RangeInChangeOption()
+            if revision is not None:
+                raise errors.BzrCommandError(
+                    '--revision and --change are mutually exclusive')
+            else:
+                revision = change
+
         # log everything
         file_id = None
         if location:
@@ -3248,10 +3263,14 @@
     def run(self, revision=None, no_backup=False, file_list=None,
             forget_merges=None):
         tree, file_list = tree_files(file_list)
-        if forget_merges:
-            tree.set_parent_ids(tree.get_parent_ids()[:1])
-        else:
-            self._revert_tree_to_revision(tree, revision, file_list, no_backup)
+        tree.lock_write()
+        try:
+            if forget_merges:
+                tree.set_parent_ids(tree.get_parent_ids()[:1])
+            else:
+                self._revert_tree_to_revision(tree, revision, file_list, no_backup)
+        finally:
+            tree.unlock()
 
     @staticmethod
     def _revert_tree_to_revision(tree, revision, file_list, no_backup):

=== modified file 'bzrlib/bzrdir.py'
--- bzrlib/bzrdir.py	2008-09-17 07:18:20 +0000
+++ bzrlib/bzrdir.py	2008-09-27 22:58:34 +0000
@@ -1157,6 +1157,10 @@
                 basis = wt.basis_tree()
                 basis.lock_read()
                 subtrees = basis.iter_references()
+            elif result_branch is not None:
+                basis = result_branch.basis_tree()
+                basis.lock_read()
+                subtrees = basis.iter_references()
             elif source_branch is not None:
                 basis = source_branch.basis_tree()
                 basis.lock_read()
@@ -3051,7 +3055,7 @@
     )
 # The following two formats should always just be aliases.
 format_registry.register_metadir('development',
-    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment1',
+    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment2',
     help='Current development format. Can convert data to and from pack-0.92 '
         '(and anything compatible with pack-0.92) format repositories. '
         'Repositories and branches in this format can only be read by bzr.dev. '
@@ -3064,7 +3068,7 @@
     alias=True,
     )
 format_registry.register_metadir('development-subtree',
-    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment1Subtree',
+    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment2Subtree',
     help='Current development format, subtree variant. Can convert data to and '
         'from pack-0.92-subtree (and anything compatible with '
         'pack-0.92-subtree) format repositories. Repositories and branches in '
@@ -3076,10 +3080,10 @@
     experimental=True,
     alias=True,
     )
-# And the development formats which the will have aliased one of follow:
-format_registry.register_metadir('development1',
-    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment1',
-    help='A branch and pack based repository that supports stacking. '
+# And the development formats above will have aliased one of the following:
+format_registry.register_metadir('development2',
+    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment2',
+    help='1.6.1 with B+Tree based index. '
         'Please read '
         'http://doc.bazaar-vcs.org/latest/developers/development-repo.html '
         'before use.',
@@ -3088,9 +3092,9 @@
     hidden=True,
     experimental=True,
     )
-format_registry.register_metadir('development1-subtree',
-    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment1Subtree',
-    help='A branch and pack based repository that supports stacking. '
+format_registry.register_metadir('development2-subtree',
+    'bzrlib.repofmt.pack_repo.RepositoryFormatPackDevelopment2Subtree',
+    help='1.6.1-subtree with B+Tree based index. '
         'Please read '
         'http://doc.bazaar-vcs.org/latest/developers/development-repo.html '
         'before use.',

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2008-08-31 19:39:01 +0000
+++ bzrlib/commands.py	2008-09-26 23:35:02 +0000
@@ -796,16 +796,7 @@
         else:
             ret = run(*run_argv)
         if 'memory' in debug.debug_flags:
-            try:
-                status_file = file('/proc/%s/status' % os.getpid(), 'rb')
-            except IOError:
-                pass
-            else:
-                status = status_file.read()
-                status_file.close()
-                trace.note("Process status after command:")
-                for line in status.splitlines():
-                    trace.note(line)
+            trace.debug_memory('Process status after command:', short=False)
         return ret or 0
     finally:
         # reset, in case we may do other commands later within the same process

=== modified file 'bzrlib/commit.py'
--- bzrlib/commit.py	2008-08-05 04:10:43 +0000
+++ bzrlib/commit.py	2008-09-19 06:53:41 +0000
@@ -690,11 +690,12 @@
                 # required after that changes.
                 if len(self.parents) > 1:
                     ie.revision = None
-                delta, version_recorded = self.builder.record_entry_contents(
+                delta, version_recorded, _ = self.builder.record_entry_contents(
                     ie, self.parent_invs, path, self.basis_tree, None)
                 if version_recorded:
                     self.any_entries_changed = True
-                if delta: self._basis_delta.append(delta)
+                if delta:
+                    self._basis_delta.append(delta)
 
     def _report_and_accumulate_deletes(self):
         # XXX: Could the list of deleted paths and ids be instead taken from
@@ -843,14 +844,18 @@
         else:
             ie = existing_ie.copy()
             ie.revision = None
-        delta, version_recorded = self.builder.record_entry_contents(ie,
-            self.parent_invs, path, self.work_tree, content_summary)
+        # For carried over entries we don't care about the fs hash - the repo
+        # isn't generating a sha, so we're not saving computation time.
+        delta, version_recorded, fs_hash = self.builder.record_entry_contents(
+            ie, self.parent_invs, path, self.work_tree, content_summary)
         if delta:
             self._basis_delta.append(delta)
         if version_recorded:
             self.any_entries_changed = True
         if report_changes:
             self._report_change(ie, path)
+        if fs_hash:
+            self.work_tree._observed_sha1(ie.file_id, path, fs_hash)
         return ie
 
     def _report_change(self, ie, path):

=== modified file 'bzrlib/delta.py'
--- bzrlib/delta.py	2008-04-24 07:22:53 +0000
+++ bzrlib/delta.py	2008-09-15 03:23:53 +0000
@@ -241,7 +241,7 @@
                                   (executable[0] != executable[1])))
         elif kind[0] != kind[1]:
             delta.kind_changed.append((path[1], file_id, kind[0], kind[1]))
-        elif content_change is True or executable[0] != executable[1]:
+        elif content_change or executable[0] != executable[1]:
             delta.modified.append((path[1], file_id, kind[1],
                                    content_change,
                                    (executable[0] != executable[1])))

=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2008-07-30 08:55:10 +0000
+++ bzrlib/dirstate.py	2008-09-26 04:57:33 +0000
@@ -222,27 +222,44 @@
     )
 
 
-def pack_stat(st, _encode=binascii.b2a_base64, _pack=struct.pack):
-    """Convert stat values into a packed representation."""
-    # jam 20060614 it isn't really worth removing more entries if we
-    # are going to leave it in packed form.
-    # With only st_mtime and st_mode filesize is 5.5M and read time is 275ms
-    # With all entries, filesize is 5.9M and read time is maybe 280ms
-    # well within the noise margin
-
-    # base64 encoding always adds a final newline, so strip it off
-    # The current version
-    return _encode(_pack('>LLLLLL'
-        , st.st_size, int(st.st_mtime), int(st.st_ctime)
-        , st.st_dev, st.st_ino & 0xFFFFFFFF, st.st_mode))[:-1]
-    # This is 0.060s / 1.520s faster by not encoding as much information
-    # return _encode(_pack('>LL', int(st.st_mtime), st.st_mode))[:-1]
-    # This is not strictly faster than _encode(_pack())[:-1]
-    # return '%X.%X.%X.%X.%X.%X' % (
-    #      st.st_size, int(st.st_mtime), int(st.st_ctime),
-    #      st.st_dev, st.st_ino, st.st_mode)
-    # Similar to the _encode(_pack('>LL'))
-    # return '%X.%X' % (int(st.st_mtime), st.st_mode)
+# This is the Windows equivalent of ENOTDIR
+# It is defined in pywin32.winerror, but we don't want a strong dependency for
+# just an error code.
+ERROR_PATH_NOT_FOUND = 3
+ERROR_DIRECTORY = 267
+
+
+if not getattr(struct, '_compile', None):
+    # Cannot pre-compile the dirstate pack_stat
+    def pack_stat(st, _encode=binascii.b2a_base64, _pack=struct.pack):
+        """Convert stat values into a packed representation."""
+        return _encode(_pack('>LLLLLL', st.st_size, int(st.st_mtime),
+            int(st.st_ctime), st.st_dev, st.st_ino & 0xFFFFFFFF,
+            st.st_mode))[:-1]
+else:
+    # compile the struct compiler we need, so as to only do it once
+    from _struct import Struct
+    _compiled_pack = Struct('>LLLLLL').pack
+    def pack_stat(st, _encode=binascii.b2a_base64, _pack=_compiled_pack):
+        """Convert stat values into a packed representation."""
+        # jam 20060614 it isn't really worth removing more entries if we
+        # are going to leave it in packed form.
+        # With only st_mtime and st_mode filesize is 5.5M and read time is 275ms
+        # With all entries, filesize is 5.9M and read time is maybe 280ms
+        # well within the noise margin
+
+        # base64 encoding always adds a final newline, so strip it off
+        # The current version
+        return _encode(_pack(st.st_size, int(st.st_mtime), int(st.st_ctime),
+            st.st_dev, st.st_ino & 0xFFFFFFFF, st.st_mode))[:-1]
+        # This is 0.060s / 1.520s faster by not encoding as much information
+        # return _encode(_pack('>LL', int(st.st_mtime), st.st_mode))[:-1]
+        # This is not strictly faster than _encode(_pack())[:-1]
+        # return '%X.%X.%X.%X.%X.%X' % (
+        #      st.st_size, int(st.st_mtime), int(st.st_ctime),
+        #      st.st_dev, st.st_ino, st.st_mode)
+        # Similar to the _encode(_pack('>LL'))
+        # return '%X.%X' % (int(st.st_mtime), st.st_mode)
 
 
 class DirState(object):
@@ -1033,6 +1050,24 @@
         self._dirblocks[0] = ('', root_block)
         self._dirblocks[1] = ('', contents_of_root_block)
 
+    def _entries_for_path(self, path):
+        """Return a list with all the entries that match path for all ids."""
+        dirname, basename = os.path.split(path)
+        key = (dirname, basename, '')
+        block_index, present = self._find_block_index_from_key(key)
+        if not present:
+            # the block which should contain path is absent.
+            return []
+        result = []
+        block = self._dirblocks[block_index][1]
+        entry_index, _ = self._find_entry_index(key, block)
+        # we may need to look at multiple entries at this path: walk while the specific_files match.
+        while (entry_index < len(block) and
+            block[entry_index][0][0:2] == key[0:2]):
+            result.append(block[entry_index])
+            entry_index += 1
+        return result
+
     def _entry_to_line(self, entry):
         """Serialize entry to a NULL delimited line ready for _get_output_lines.
 
@@ -1472,17 +1507,13 @@
                     # it is being resurrected here, so blank it out temporarily.
                     self._dirblocks[block_index][1][entry_index][1][1] = null
 
-    def update_entry(self, entry, abspath, stat_value,
-                     _stat_to_minikind=_stat_to_minikind,
-                     _pack_stat=pack_stat):
-        """Update the entry based on what is actually on disk.
+    def _observed_sha1(self, entry, sha1, stat_value,
+        _stat_to_minikind=_stat_to_minikind, _pack_stat=pack_stat):
+        """Note the sha1 of a file.
 
-        :param entry: This is the dirblock entry for the file in question.
-        :param abspath: The path on disk for this file.
-        :param stat_value: (optional) if we already have done a stat on the
-            file, re-use it.
-        :return: The sha1 hexdigest of the file (40 bytes) or link target of a
-                symlink.
+        :param entry: The entry the sha1 is for.
+        :param sha1: The observed sha1.
+        :param stat_value: The os.lstat for the file.
         """
         try:
             minikind = _stat_to_minikind[stat_value.st_mode & 0170000]
@@ -1490,60 +1521,14 @@
             # Unhandled kind
             return None
         packed_stat = _pack_stat(stat_value)
-        (saved_minikind, saved_link_or_sha1, saved_file_size,
-         saved_executable, saved_packed_stat) = entry[1][0]
-
-        if (minikind == saved_minikind
-            and packed_stat == saved_packed_stat):
-            # The stat hasn't changed since we saved, so we can re-use the
-            # saved sha hash.
-            if minikind == 'd':
-                return None
-
-            # size should also be in packed_stat
-            if saved_file_size == stat_value.st_size:
-                return saved_link_or_sha1
-
-        # If we have gotten this far, that means that we need to actually
-        # process this entry.
-        link_or_sha1 = None
         if minikind == 'f':
-            link_or_sha1 = self._sha1_file(abspath)
-            executable = self._is_executable(stat_value.st_mode,
-                                             saved_executable)
-            if self._cutoff_time is None:
-                self._sha_cutoff_time()
-            if (stat_value.st_mtime < self._cutoff_time
-                and stat_value.st_ctime < self._cutoff_time):
-                entry[1][0] = ('f', link_or_sha1, stat_value.st_size,
-                               executable, packed_stat)
-            else:
-                entry[1][0] = ('f', '', stat_value.st_size,
-                               executable, DirState.NULLSTAT)
-        elif minikind == 'd':
-            link_or_sha1 = None
-            entry[1][0] = ('d', '', 0, False, packed_stat)
-            if saved_minikind != 'd':
-                # This changed from something into a directory. Make sure we
-                # have a directory block for it. This doesn't happen very
-                # often, so this doesn't have to be super fast.
-                block_index, entry_index, dir_present, file_present = \
-                    self._get_block_entry_index(entry[0][0], entry[0][1], 0)
-                self._ensure_block(block_index, entry_index,
-                                   osutils.pathjoin(entry[0][0], entry[0][1]))
-        elif minikind == 'l':
-            link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
-            if self._cutoff_time is None:
-                self._sha_cutoff_time()
-            if (stat_value.st_mtime < self._cutoff_time
-                and stat_value.st_ctime < self._cutoff_time):
-                entry[1][0] = ('l', link_or_sha1, stat_value.st_size,
-                               False, packed_stat)
-            else:
-                entry[1][0] = ('l', '', stat_value.st_size,
-                               False, DirState.NULLSTAT)
-        self._dirblock_state = DirState.IN_MEMORY_MODIFIED
-        return link_or_sha1
+            if self._cutoff_time is None:
+                self._sha_cutoff_time()
+            if (stat_value.st_mtime < self._cutoff_time
+                and stat_value.st_ctime < self._cutoff_time):
+                entry[1][0] = ('f', sha1, entry[1][0][2], entry[1][0][3],
+                    packed_stat)
+                self._dirblock_state = DirState.IN_MEMORY_MODIFIED
 
     def _sha_cutoff_time(self):
         """Return cutoff time.
@@ -2774,6 +2759,732 @@
             raise errors.ObjectNotLocked(self)
 
 
+def py_update_entry(state, entry, abspath, stat_value,
+                 _stat_to_minikind=DirState._stat_to_minikind,
+                 _pack_stat=pack_stat):
+    """Update the entry based on what is actually on disk.
+
+    This function only calculates the sha if it needs to - if the entry is
+    uncachable, or clearly different to the first parent's entry, no sha
+    is calculated, and None is returned.
+
+    :param state: The dirstate this entry is in.
+    :param entry: This is the dirblock entry for the file in question.
+    :param abspath: The path on disk for this file.
+    :param stat_value: The stat value done on the path.
+    :return: None, or The sha1 hexdigest of the file (40 bytes) or link
+        target of a symlink.
+    """
+    try:
+        minikind = _stat_to_minikind[stat_value.st_mode & 0170000]
+    except KeyError:
+        # Unhandled kind
+        return None
+    packed_stat = _pack_stat(stat_value)
+    (saved_minikind, saved_link_or_sha1, saved_file_size,
+     saved_executable, saved_packed_stat) = entry[1][0]
+
+    if (minikind == saved_minikind
+        and packed_stat == saved_packed_stat):
+        # The stat hasn't changed since we saved, so we can re-use the
+        # saved sha hash.
+        if minikind == 'd':
+            return None
+
+        # size should also be in packed_stat
+        if saved_file_size == stat_value.st_size:
+            return saved_link_or_sha1
+
+    # If we have gotten this far, that means that we need to actually
+    # process this entry.
+    link_or_sha1 = None
+    if minikind == 'f':
+        executable = state._is_executable(stat_value.st_mode,
+                                         saved_executable)
+        if state._cutoff_time is None:
+            state._sha_cutoff_time()
+        if (stat_value.st_mtime < state._cutoff_time
+            and stat_value.st_ctime < state._cutoff_time
+            and len(entry[1]) > 1
+            and entry[1][1][0] != 'a'):
+                # Could check for size changes for further optimised
+                # avoidance of sha1's. However the most prominent case of
+                # over-shaing is during initial add, which this catches.
+            link_or_sha1 = state._sha1_file(abspath)
+            entry[1][0] = ('f', link_or_sha1, stat_value.st_size,
+                           executable, packed_stat)
+        else:
+            entry[1][0] = ('f', '', stat_value.st_size,
+                           executable, DirState.NULLSTAT)
+    elif minikind == 'd':
+        link_or_sha1 = None
+        entry[1][0] = ('d', '', 0, False, packed_stat)
+        if saved_minikind != 'd':
+            # This changed from something into a directory. Make sure we
+            # have a directory block for it. This doesn't happen very
+            # often, so this doesn't have to be super fast.
+            block_index, entry_index, dir_present, file_present = \
+                state._get_block_entry_index(entry[0][0], entry[0][1], 0)
+            state._ensure_block(block_index, entry_index,
+                               osutils.pathjoin(entry[0][0], entry[0][1]))
+    elif minikind == 'l':
+        link_or_sha1 = state._read_link(abspath, saved_link_or_sha1)
+        if state._cutoff_time is None:
+            state._sha_cutoff_time()
+        if (stat_value.st_mtime < state._cutoff_time
+            and stat_value.st_ctime < state._cutoff_time):
+            entry[1][0] = ('l', link_or_sha1, stat_value.st_size,
+                           False, packed_stat)
+        else:
+            entry[1][0] = ('l', '', stat_value.st_size,
+                           False, DirState.NULLSTAT)
+    state._dirblock_state = DirState.IN_MEMORY_MODIFIED
+    return link_or_sha1
+update_entry = py_update_entry
+
+
+class ProcessEntryPython(object):
+
+    __slots__ = ["old_dirname_to_file_id", "new_dirname_to_file_id", "uninteresting",
+        "last_source_parent", "last_target_parent", "include_unchanged",
+        "use_filesystem_for_exec", "utf8_decode", "searched_specific_files",
+        "search_specific_files", "state", "source_index", "target_index",
+        "want_unversioned", "tree"]
+
+    def __init__(self, include_unchanged, use_filesystem_for_exec,
+        search_specific_files, state, source_index, target_index,
+        want_unversioned, tree):
+        self.old_dirname_to_file_id = {}
+        self.new_dirname_to_file_id = {}
+        # Just a sentry, so that _process_entry can say that this
+        # record is handled, but isn't interesting to process (unchanged)
+        self.uninteresting = object()
+        # Using a list so that we can access the values and change them in
+        # nested scope. Each one is [path, file_id, entry]
+        self.last_source_parent = [None, None]
+        self.last_target_parent = [None, None]
+        self.include_unchanged = include_unchanged
+        self.use_filesystem_for_exec = use_filesystem_for_exec
+        self.utf8_decode = cache_utf8._utf8_decode
+        # for all search_indexs in each path at or under each element of
+        # search_specific_files, if the detail is relocated: add the id, and add the
+        # relocated path as one to search if its not searched already. If the
+        # detail is not relocated, add the id.
+        self.searched_specific_files = set()
+        self.search_specific_files = search_specific_files
+        self.state = state
+        self.source_index = source_index
+        self.target_index = target_index
+        self.want_unversioned = want_unversioned
+        self.tree = tree
+
+    def _process_entry(self, entry, path_info, pathjoin=osutils.pathjoin):
+        """Compare an entry and real disk to generate delta information.
+
+        :param path_info: top_relpath, basename, kind, lstat, abspath for
+            the path of entry. If None, then the path is considered absent.
+            (Perhaps we should pass in a concrete entry for this ?)
+            Basename is returned as a utf8 string because we expect this
+            tuple will be ignored, and don't want to take the time to
+            decode.
+        :return: None if these don't match
+                 A tuple of information about the change, or
+                 the object 'uninteresting' if these match, but are
+                 basically identical.
+        """
+        if self.source_index is None:
+            source_details = DirState.NULL_PARENT_DETAILS
+        else:
+            source_details = entry[1][self.source_index]
+        target_details = entry[1][self.target_index]
+        target_minikind = target_details[0]
+        if path_info is not None and target_minikind in 'fdlt':
+            if not (self.target_index == 0):
+                raise AssertionError()
+            link_or_sha1 = update_entry(self.state, entry,
+                abspath=path_info[4], stat_value=path_info[3])
+            # The entry may have been modified by update_entry
+            target_details = entry[1][self.target_index]
+            target_minikind = target_details[0]
+        else:
+            link_or_sha1 = None
+        file_id = entry[0][2]
+        source_minikind = source_details[0]
+        if source_minikind in 'fdltr' and target_minikind in 'fdlt':
+            # claimed content in both: diff
+            #   r    | fdlt   |      | add source to search, add id path move and perform
+            #        |        |      | diff check on source-target
+            #   r    | fdlt   |  a   | dangling file that was present in the basis.
+            #        |        |      | ???
+            if source_minikind in 'r':
+                # add the source to the search path to find any children it
+                # has.  TODO ? : only add if it is a container ?
+                if not osutils.is_inside_any(self.searched_specific_files,
+                                             source_details[1]):
+                    self.search_specific_files.add(source_details[1])
+                # generate the old path; this is needed for stating later
+                # as well.
+                old_path = source_details[1]
+                old_dirname, old_basename = os.path.split(old_path)
+                path = pathjoin(entry[0][0], entry[0][1])
+                old_entry = self.state._get_entry(self.source_index,
+                                             path_utf8=old_path)
+                # update the source details variable to be the real
+                # location.
+                if old_entry == (None, None):
+                    raise errors.CorruptDirstate(self.state._filename,
+                        "entry '%s/%s' is considered renamed from %r"
+                        " but source does not exist\n"
+                        "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
+                source_details = old_entry[1][self.source_index]
+                source_minikind = source_details[0]
+            else:
+                old_dirname = entry[0][0]
+                old_basename = entry[0][1]
+                old_path = path = None
+            if path_info is None:
+                # the file is missing on disk, show as removed.
+                content_change = True
+                target_kind = None
+                target_exec = False
+            else:
+                # source and target are both versioned and disk file is present.
+                target_kind = path_info[2]
+                if target_kind == 'directory':
+                    if path is None:
+                        old_path = path = pathjoin(old_dirname, old_basename)
+                    self.new_dirname_to_file_id[path] = file_id
+                    if source_minikind != 'd':
+                        content_change = True
+                    else:
+                        # directories have no fingerprint
+                        content_change = False
+                    target_exec = False
+                elif target_kind == 'file':
+                    if source_minikind != 'f':
+                        content_change = True
+                    else:
+                        # If the size is the same, check the sha:
+                        if target_details[2] == source_details[2]:
+                            if link_or_sha1 is None:
+                                # Stat cache miss:
+                                file_obj = file(path_info[4], 'rb')
+                                try:
+                                    statvalue = os.fstat(file_obj.fileno())
+                                    link_or_sha1 = osutils.sha_file(file_obj)
+                                finally:
+                                    file_obj.close()
+                                self.state._observed_sha1(entry, link_or_sha1,
+                                    statvalue)
+                            content_change = (link_or_sha1 != source_details[1])
+                        else:
+                            # Size changed, so must be different
+                            content_change = True
+                    # Target details is updated at update_entry time
+                    if self.use_filesystem_for_exec:
+                        # We don't need S_ISREG here, because we are sure
+                        # we are dealing with a file.
+                        target_exec = bool(stat.S_IEXEC & path_info[3].st_mode)
+                    else:
+                        target_exec = target_details[3]
+                elif target_kind == 'symlink':
+                    if source_minikind != 'l':
+                        content_change = True
+                    else:
+                        content_change = (link_or_sha1 != source_details[1])
+                    target_exec = False
+                elif target_kind == 'tree-reference':
+                    if source_minikind != 't':
+                        content_change = True
+                    else:
+                        content_change = False
+                    target_exec = False
+                else:
+                    raise Exception, "unknown kind %s" % path_info[2]
+            if source_minikind == 'd':
+                if path is None:
+                    old_path = path = pathjoin(old_dirname, old_basename)
+                self.old_dirname_to_file_id[old_path] = file_id
+            # parent id is the entry for the path in the target tree
+            if old_dirname == self.last_source_parent[0]:
+                source_parent_id = self.last_source_parent[1]
+            else:
+                try:
+                    source_parent_id = self.old_dirname_to_file_id[old_dirname]
+                except KeyError:
+                    source_parent_entry = self.state._get_entry(self.source_index,
+                                                           path_utf8=old_dirname)
+                    source_parent_id = source_parent_entry[0][2]
+                if source_parent_id == entry[0][2]:
+                    # This is the root, so the parent is None
+                    source_parent_id = None
+                else:
+                    self.last_source_parent[0] = old_dirname
+                    self.last_source_parent[1] = source_parent_id
+            new_dirname = entry[0][0]
+            if new_dirname == self.last_target_parent[0]:
+                target_parent_id = self.last_target_parent[1]
+            else:
+                try:
+                    target_parent_id = self.new_dirname_to_file_id[new_dirname]
+                except KeyError:
+                    # TODO: We don't always need to do the lookup, because the
+                    #       parent entry will be the same as the source entry.
+                    target_parent_entry = self.state._get_entry(self.target_index,
+                                                           path_utf8=new_dirname)
+                    if target_parent_entry == (None, None):
+                        raise AssertionError(
+                            "Could not find target parent in wt: %s\nparent of: %s"
+                            % (new_dirname, entry))
+                    target_parent_id = target_parent_entry[0][2]
+                if target_parent_id == entry[0][2]:
+                    # This is the root, so the parent is None
+                    target_parent_id = None
+                else:
+                    self.last_target_parent[0] = new_dirname
+                    self.last_target_parent[1] = target_parent_id
+
+            source_exec = source_details[3]
+            if (self.include_unchanged
+                or content_change
+                or source_parent_id != target_parent_id
+                or old_basename != entry[0][1]
+                or source_exec != target_exec
+                ):
+                if old_path is None:
+                    old_path = path = pathjoin(old_dirname, old_basename)
+                    old_path_u = self.utf8_decode(old_path)[0]
+                    path_u = old_path_u
+                else:
+                    old_path_u = self.utf8_decode(old_path)[0]
+                    if old_path == path:
+                        path_u = old_path_u
+                    else:
+                        path_u = self.utf8_decode(path)[0]
+                source_kind = DirState._minikind_to_kind[source_minikind]
+                return (entry[0][2],
+                       (old_path_u, path_u),
+                       content_change,
+                       (True, True),
+                       (source_parent_id, target_parent_id),
+                       (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),
+                       (source_kind, target_kind),
+                       (source_exec, target_exec))
+            else:
+                return self.uninteresting
+        elif source_minikind in 'a' and target_minikind in 'fdlt':
+            # looks like a new file
+            path = pathjoin(entry[0][0], entry[0][1])
+            # parent id is the entry for the path in the target tree
+            # TODO: these are the same for an entire directory: cache em.
+            parent_id = self.state._get_entry(self.target_index,
+                                         path_utf8=entry[0][0])[0][2]
+            if parent_id == entry[0][2]:
+                parent_id = None
+            if path_info is not None:
+                # Present on disk:
+                if self.use_filesystem_for_exec:
+                    # We need S_ISREG here, because we aren't sure if this
+                    # is a file or not.
+                    target_exec = bool(
+                        stat.S_ISREG(path_info[3].st_mode)
+                        and stat.S_IEXEC & path_info[3].st_mode)
+                else:
+                    target_exec = target_details[3]
+                return (entry[0][2],
+                       (None, self.utf8_decode(path)[0]),
+                       True,
+                       (False, True),
+                       (None, parent_id),
+                       (None, self.utf8_decode(entry[0][1])[0]),
+                       (None, path_info[2]),
+                       (None, target_exec))
+            else:
+                # Its a missing file, report it as such.
+                return (entry[0][2],
+                       (None, self.utf8_decode(path)[0]),
+                       False,
+                       (False, True),
+                       (None, parent_id),
+                       (None, self.utf8_decode(entry[0][1])[0]),
+                       (None, None),
+                       (None, False))
+        elif source_minikind in 'fdlt' and target_minikind in 'a':
+            # unversioned, possibly, or possibly not deleted: we dont care.
+            # if its still on disk, *and* theres no other entry at this
+            # path [we dont know this in this routine at the moment -
+            # perhaps we should change this - then it would be an unknown.
+            old_path = pathjoin(entry[0][0], entry[0][1])
+            # parent id is the entry for the path in the target tree
+            parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
+            if parent_id == entry[0][2]:
+                parent_id = None
+            return (entry[0][2],
+                   (self.utf8_decode(old_path)[0], None),
+                   True,
+                   (True, False),
+                   (parent_id, None),
+                   (self.utf8_decode(entry[0][1])[0], None),
+                   (DirState._minikind_to_kind[source_minikind], None),
+                   (source_details[3], None))
+        elif source_minikind in 'fdlt' and target_minikind in 'r':
+            # a rename; could be a true rename, or a rename inherited from
+            # a renamed parent. TODO: handle this efficiently. Its not
+            # common case to rename dirs though, so a correct but slow
+            # implementation will do.
+            if not osutils.is_inside_any(self.searched_specific_files, target_details[1]):
+                self.search_specific_files.add(target_details[1])
+        elif source_minikind in 'ra' and target_minikind in 'ra':
+            # neither of the selected trees contain this file,
+            # so skip over it. This is not currently directly tested, but
+            # is indirectly via test_too_much.TestCommands.test_conflicts.
+            pass
+        else:
+            raise AssertionError("don't know how to compare "
+                "source_minikind=%r, target_minikind=%r"
+                % (source_minikind, target_minikind))
+            ## import pdb;pdb.set_trace()
+        return None
+
+    def __iter__(self):
+        return self
+
+    def iter_changes(self):
+        """Iterate over the changes."""
+        utf8_decode = cache_utf8._utf8_decode
+        _cmp_by_dirs = cmp_by_dirs
+        _process_entry = self._process_entry
+        uninteresting = self.uninteresting
+        search_specific_files = self.search_specific_files
+        searched_specific_files = self.searched_specific_files
+        splitpath = osutils.splitpath
+        # sketch: 
+        # compare source_index and target_index at or under each element of search_specific_files.
+        # follow the following comparison table. Note that we only want to do diff operations when
+        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
+        # for the target.
+        # cases:
+        # 
+        # Source | Target | disk | action
+        #   r    | fdlt   |      | add source to search, add id path move and perform
+        #        |        |      | diff check on source-target
+        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
+        #        |        |      | ???
+        #   r    |  a     |      | add source to search
+        #   r    |  a     |  a   | 
+        #   r    |  r     |      | this path is present in a non-examined tree, skip.
+        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
+        #   a    | fdlt   |      | add new id
+        #   a    | fdlt   |  a   | dangling locally added file, skip
+        #   a    |  a     |      | not present in either tree, skip
+        #   a    |  a     |  a   | not present in any tree, skip
+        #   a    |  r     |      | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #   a    |  r     |  a   | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #  fdlt  | fdlt   |      | content in both: diff them
+        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
+        #  fdlt  |  a     |      | unversioned: output deleted id for now
+        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
+        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+
+        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
+        #       keeping a cache of directories that we have seen.
+
+        while search_specific_files:
+            # TODO: the pending list should be lexically sorted?  the
+            # interface doesn't require it.
+            current_root = search_specific_files.pop()
+            current_root_unicode = current_root.decode('utf8')
+            searched_specific_files.add(current_root)
+            # process the entries for this containing directory: the rest will be
+            # found by their parents recursively.
+            root_entries = self.state._entries_for_path(current_root)
+            root_abspath = self.tree.abspath(current_root_unicode)
+            try:
+                root_stat = os.lstat(root_abspath)
+            except OSError, e:
+                if e.errno == errno.ENOENT:
+                    # the path does not exist: let _process_entry know that.
+                    root_dir_info = None
+                else:
+                    # some other random error: hand it up.
+                    raise
+            else:
+                root_dir_info = ('', current_root,
+                    osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
+                    root_abspath)
+                if root_dir_info[2] == 'directory':
+                    if self.tree._directory_is_tree_reference(
+                        current_root.decode('utf8')):
+                        root_dir_info = root_dir_info[:2] + \
+                            ('tree-reference',) + root_dir_info[3:]
+
+            if not root_entries and not root_dir_info:
+                # this specified path is not present at all, skip it.
+                continue
+            path_handled = False
+            for entry in root_entries:
+                result = _process_entry(entry, root_dir_info)
+                if result is not None:
+                    path_handled = True
+                    if result is not uninteresting:
+                        yield result
+            if self.want_unversioned and not path_handled and root_dir_info:
+                new_executable = bool(
+                    stat.S_ISREG(root_dir_info[3].st_mode)
+                    and stat.S_IEXEC & root_dir_info[3].st_mode)
+                yield (None,
+                       (None, current_root_unicode),
+                       True,
+                       (False, False),
+                       (None, None),
+                       (None, splitpath(current_root_unicode)[-1]),
+                       (None, root_dir_info[2]),
+                       (None, new_executable)
+                      )
+            initial_key = (current_root, '', '')
+            block_index, _ = self.state._find_block_index_from_key(initial_key)
+            if block_index == 0:
+                # we have processed the total root already, but because the
+                # initial key matched it we should skip it here.
+                block_index +=1
+            if root_dir_info and root_dir_info[2] == 'tree-reference':
+                current_dir_info = None
+            else:
+                dir_iterator = osutils._walkdirs_utf8(root_abspath, prefix=current_root)
+                try:
+                    current_dir_info = dir_iterator.next()
+                except OSError, e:
+                    # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
+                    # python 2.5 has e.errno == EINVAL,
+                    #            and e.winerror == ERROR_DIRECTORY
+                    e_winerror = getattr(e, 'winerror', None)
+                    win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
+                    # there may be directories in the inventory even though
+                    # this path is not a file on disk: so mark it as end of
+                    # iterator
+                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
+                        current_dir_info = None
+                    elif (sys.platform == 'win32'
+                          and (e.errno in win_errors
+                               or e_winerror in win_errors)):
+                        current_dir_info = None
+                    else:
+                        raise
+                else:
+                    if current_dir_info[0][0] == '':
+                        # remove .bzr from iteration
+                        bzr_index = bisect.bisect_left(current_dir_info[1], ('.bzr',))
+                        if current_dir_info[1][bzr_index][0] != '.bzr':
+                            raise AssertionError()
+                        del current_dir_info[1][bzr_index]
+            # walk until both the directory listing and the versioned metadata
+            # are exhausted. 
+            if (block_index < len(self.state._dirblocks) and
+                osutils.is_inside(current_root, self.state._dirblocks[block_index][0])):
+                current_block = self.state._dirblocks[block_index]
+            else:
+                current_block = None
+            while (current_dir_info is not None or
+                   current_block is not None):
+                if (current_dir_info and current_block
+                    and current_dir_info[0][0] != current_block[0]):
+                    if _cmp_by_dirs(current_dir_info[0][0], current_block[0]) < 0:
+                        # filesystem data refers to paths not covered by the dirblock.
+                        # this has two possibilities:
+                        # A) it is versioned but empty, so there is no block for it
+                        # B) it is not versioned.
+
+                        # if (A) then we need to recurse into it to check for
+                        # new unknown files or directories.
+                        # if (B) then we should ignore it, because we don't
+                        # recurse into unknown directories.
+                        path_index = 0
+                        while path_index < len(current_dir_info[1]):
+                                current_path_info = current_dir_info[1][path_index]
+                                if self.want_unversioned:
+                                    if current_path_info[2] == 'directory':
+                                        if self.tree._directory_is_tree_reference(
+                                            current_path_info[0].decode('utf8')):
+                                            current_path_info = current_path_info[:2] + \
+                                                ('tree-reference',) + current_path_info[3:]
+                                    new_executable = bool(
+                                        stat.S_ISREG(current_path_info[3].st_mode)
+                                        and stat.S_IEXEC & current_path_info[3].st_mode)
+                                    yield (None,
+                                        (None, utf8_decode(current_path_info[0])[0]),
+                                        True,
+                                        (False, False),
+                                        (None, None),
+                                        (None, utf8_decode(current_path_info[1])[0]),
+                                        (None, current_path_info[2]),
+                                        (None, new_executable))
+                                # dont descend into this unversioned path if it is
+                                # a dir
+                                if current_path_info[2] in ('directory',
+                                                            'tree-reference'):
+                                    del current_dir_info[1][path_index]
+                                    path_index -= 1
+                                path_index += 1
+
+                        # This dir info has been handled, go to the next
+                        try:
+                            current_dir_info = dir_iterator.next()
+                        except StopIteration:
+                            current_dir_info = None
+                    else:
+                        # We have a dirblock entry for this location, but there
+                        # is no filesystem path for this. This is most likely
+                        # because a directory was removed from the disk.
+                        # We don't have to report the missing directory,
+                        # because that should have already been handled, but we
+                        # need to handle all of the files that are contained
+                        # within.
+                        for current_entry in current_block[1]:
+                            # entry referring to file not present on disk.
+                            # advance the entry only, after processing.
+                            result = _process_entry(current_entry, None)
+                            if result is not None:
+                                if result is not uninteresting:
+                                    yield result
+                        block_index +=1
+                        if (block_index < len(self.state._dirblocks) and
+                            osutils.is_inside(current_root,
+                                              self.state._dirblocks[block_index][0])):
+                            current_block = self.state._dirblocks[block_index]
+                        else:
+                            current_block = None
+                    continue
+                entry_index = 0
+                if current_block and entry_index < len(current_block[1]):
+                    current_entry = current_block[1][entry_index]
+                else:
+                    current_entry = None
+                advance_entry = True
+                path_index = 0
+                if current_dir_info and path_index < len(current_dir_info[1]):
+                    current_path_info = current_dir_info[1][path_index]
+                    if current_path_info[2] == 'directory':
+                        if self.tree._directory_is_tree_reference(
+                            current_path_info[0].decode('utf8')):
+                            current_path_info = current_path_info[:2] + \
+                                ('tree-reference',) + current_path_info[3:]
+                else:
+                    current_path_info = None
+                advance_path = True
+                path_handled = False
+                while (current_entry is not None or
+                    current_path_info is not None):
+                    if current_entry is None:
+                        # the check for path_handled when the path is adnvaced
+                        # will yield this path if needed.
+                        pass
+                    elif current_path_info is None:
+                        # no path is fine: the per entry code will handle it.
+                        result = _process_entry(current_entry, current_path_info)
+                        if result is not None:
+                            if result is not uninteresting:
+                                yield result
+                    elif (current_entry[0][1] != current_path_info[1]
+                          or current_entry[1][self.target_index][0] in 'ar'):
+                        # The current path on disk doesn't match the dirblock
+                        # record. Either the dirblock is marked as absent, or
+                        # the file on disk is not present at all in the
+                        # dirblock. Either way, report about the dirblock
+                        # entry, and let other code handle the filesystem one.
+
+                        # Compare the basename for these files to determine
+                        # which comes first
+                        if current_path_info[1] < current_entry[0][1]:
+                            # extra file on disk: pass for now, but only
+                            # increment the path, not the entry
+                            advance_entry = False
+                        else:
+                            # entry referring to file not present on disk.
+                            # advance the entry only, after processing.
+                            result = _process_entry(current_entry, None)
+                            if result is not None:
+                                if result is not uninteresting:
+                                    yield result
+                            advance_path = False
+                    else:
+                        result = _process_entry(current_entry, current_path_info)
+                        if result is not None:
+                            path_handled = True
+                            if result is not uninteresting:
+                                yield result
+                    if advance_entry and current_entry is not None:
+                        entry_index += 1
+                        if entry_index < len(current_block[1]):
+                            current_entry = current_block[1][entry_index]
+                        else:
+                            current_entry = None
+                    else:
+                        advance_entry = True # reset the advance flaga
+                    if advance_path and current_path_info is not None:
+                        if not path_handled:
+                            # unversioned in all regards
+                            if self.want_unversioned:
+                                new_executable = bool(
+                                    stat.S_ISREG(current_path_info[3].st_mode)
+                                    and stat.S_IEXEC & current_path_info[3].st_mode)
+                                try:
+                                    relpath_unicode = utf8_decode(current_path_info[0])[0]
+                                except UnicodeDecodeError:
+                                    raise errors.BadFilenameEncoding(
+                                        current_path_info[0], osutils._fs_enc)
+                                yield (None,
+                                    (None, relpath_unicode),
+                                    True,
+                                    (False, False),
+                                    (None, None),
+                                    (None, utf8_decode(current_path_info[1])[0]),
+                                    (None, current_path_info[2]),
+                                    (None, new_executable))
+                            # dont descend into this unversioned path if it is
+                            # a dir
+                            if current_path_info[2] in ('directory'):
+                                del current_dir_info[1][path_index]
+                                path_index -= 1
+                        # dont descend the disk iterator into any tree 
+                        # paths.
+                        if current_path_info[2] == 'tree-reference':
+                            del current_dir_info[1][path_index]
+                            path_index -= 1
+                        path_index += 1
+                        if path_index < len(current_dir_info[1]):
+                            current_path_info = current_dir_info[1][path_index]
+                            if current_path_info[2] == 'directory':
+                                if self.tree._directory_is_tree_reference(
+                                    current_path_info[0].decode('utf8')):
+                                    current_path_info = current_path_info[:2] + \
+                                        ('tree-reference',) + current_path_info[3:]
+                        else:
+                            current_path_info = None
+                        path_handled = False
+                    else:
+                        advance_path = True # reset the advance flagg.
+                if current_block is not None:
+                    block_index += 1
+                    if (block_index < len(self.state._dirblocks) and
+                        osutils.is_inside(current_root, self.state._dirblocks[block_index][0])):
+                        current_block = self.state._dirblocks[block_index]
+                    else:
+                        current_block = None
+                if current_dir_info is not None:
+                    try:
+                        current_dir_info = dir_iterator.next()
+                    except StopIteration:
+                        current_dir_info = None
+_process_entry = ProcessEntryPython
+
+
 # Try to load the compiled form if possible
 try:
     from bzrlib._dirstate_helpers_c import (
@@ -2782,6 +3493,8 @@
         _bisect_path_left_c as _bisect_path_left,
         _bisect_path_right_c as _bisect_path_right,
         cmp_by_dirs_c as cmp_by_dirs,
+        ProcessEntryC as _process_entry,
+        update_entry as update_entry,
         )
 except ImportError:
     from bzrlib._dirstate_helpers_py import (

=== modified file 'bzrlib/help_topics/__init__.py'
--- bzrlib/help_topics/__init__.py	2008-09-04 18:37:59 +0000
+++ bzrlib/help_topics/__init__.py	2008-09-23 19:15:53 +0000
@@ -711,7 +711,8 @@
 topic_registry.register('criss-cross', _criss_cross,
                         'Information on criss-cross merging', SECT_CONCEPT)
 topic_registry.register('sync-for-reconfigure', _branches_out_of_sync,
-                        'Information on criss-cross merging', SECT_CONCEPT)
+                        'Steps to resolve "out-of-sync" when reconfiguring',
+                        SECT_CONCEPT)
 
 
 class HelpTopicIndex(object):

=== modified file 'bzrlib/index.py'
--- bzrlib/index.py	2008-09-03 22:30:56 +0000
+++ bzrlib/index.py	2008-09-21 14:48:37 +0000
@@ -366,7 +366,7 @@
         self._keys_by_offset = {}
         # ready-to-return key:value or key:value, node_ref_lists
         self._nodes = {}
-        self._nodes_by_key = {}
+        self._nodes_by_key = None
         trailers = 0
         pos = stream.tell()
         lines = stream.read().split('\n')
@@ -381,26 +381,30 @@
             else:
                 node_value = value
             self._nodes[key] = node_value
-            if self._key_length > 1:
-                # TODO: We may want to do this lazily, but if we are calling
-                #       _buffer_all, we are likely to be doing
-                #       iter_entries_prefix
-                key_dict = self._nodes_by_key
-                if self.node_ref_lists:
-                    key_value = key, node_value[0], node_value[1]
-                else:
-                    key_value = key, node_value
-                # For a key of (foo, bar, baz) create
-                # _nodes_by_key[foo][bar][baz] = key_value
-                for subkey in key[:-1]:
-                    key_dict = key_dict.setdefault(subkey, {})
-                key_dict[key[-1]] = key_value
         # cache the keys for quick set intersections
         self._keys = set(self._nodes)
         if trailers != 1:
             # there must be one line - the empty trailer line.
             raise errors.BadIndexData(self)
 
+    def _get_nodes_by_key(self):
+        if self._nodes_by_key is None:
+            nodes_by_key = {}
+            if self.node_ref_lists:
+                for key, (value, references) in self._nodes.iteritems():
+                    key_dict = nodes_by_key
+                    for subkey in key[:-1]:
+                        key_dict = key_dict.setdefault(subkey, {})
+                    key_dict[key[-1]] = key, value, references
+            else:
+                for key, value in self._nodes.iteritems():
+                    key_dict = nodes_by_key
+                    for subkey in key[:-1]:
+                        key_dict = key_dict.setdefault(subkey, {})
+                    key_dict[key[-1]] = key, value
+            self._nodes_by_key = nodes_by_key
+        return self._nodes_by_key
+
     def iter_all_entries(self):
         """Iterate over all keys within the index.
 
@@ -593,6 +597,7 @@
                 else:
                     yield self, key, self._nodes[key]
             return
+        nodes_by_key = self._get_nodes_by_key()
         for key in keys:
             # sanity check
             if key[0] is None:
@@ -600,7 +605,7 @@
             if len(key) != self._key_length:
                 raise errors.BadIndexKey(key)
             # find what it refers to:
-            key_dict = self._nodes_by_key
+            key_dict = nodes_by_key
             elements = list(key)
             # find the subdict whose contents should be returned.
             try:
@@ -973,7 +978,7 @@
                 raise errors.BadIndexData(self)
             # keys are tuples. Each element is a string that may occur many
             # times, so we intern them to save space. AB, RC, 200807
-            key = tuple(intern(element) for element in elements[:self._key_length])
+            key = tuple([intern(element) for element in elements[:self._key_length]])
             if first_key is None:
                 first_key = key
             absent, references, value = elements[-3:]

=== modified file 'bzrlib/log.py'
--- bzrlib/log.py	2008-09-11 17:13:46 +0000
+++ bzrlib/log.py	2008-09-22 20:09:07 +0000
@@ -268,8 +268,8 @@
     if specific_fileid:
         view_revisions = _filter_revisions_touching_file_id(branch,
                                                          specific_fileid,
-                                                         mainline_revs,
-                                                         view_revisions)
+                                                         view_revisions,
+                                                         direction)
 
     # rebase merge_depth - unless there are no revisions or 
     # either the first or last revision have merge_depth = 0.
@@ -536,75 +536,88 @@
     return view_revisions
 
 
-def _filter_revisions_touching_file_id(branch, file_id, mainline_revisions,
-                                       view_revs_iter):
-    """Return the list of revision ids which touch a given file id.
+def _filter_revisions_touching_file_id(branch, file_id, view_revisions,
+                                       direction):
+    r"""Return the list of revision ids which touch a given file id.
 
     The function filters view_revisions and returns a subset.
     This includes the revisions which directly change the file id,
     and the revisions which merge these changes. So if the
     revision graph is::
-        A
-        |\
-        B C
+        A-.
+        |\ \
+        B C E
+        |/ /
+        D |
+        |\|
+        | F
         |/
-        D
-
-    And 'C' changes a file, then both C and D will be returned.
-
-    This will also can be restricted based on a subset of the mainline.
-
+        G
+
+    And 'C' changes a file, then both C and D will be returned. F will not be
+    returned even though it brings the changes to C into the branch starting
+    with E. (Note that if we were using F as the tip instead of G, then we
+    would see C, D, F.)
+
+    This will also be restricted based on a subset of the mainline.
+
+    :param branch: The branch where we can get text revision information.
+    :param file_id: Filter out revisions that do not touch file_id.
+    :param view_revisions: A list of (revision_id, dotted_revno, merge_depth)
+        tuples. This is the list of revisions which will be filtered. It is
+        assumed that view_revisions is in merge_sort order (either forward or
+        reverse).
+    :param direction: The direction of view_revisions.  See also
+        reverse_by_depth, and get_view_revisions
     :return: A list of (revision_id, dotted_revno, merge_depth) tuples.
     """
-    # find all the revisions that change the specific file
-    # build the ancestry of each revision in the graph
-    # - only listing the ancestors that change the specific file.
-    graph = branch.repository.get_graph()
-    # This asks for all mainline revisions, which means we only have to spider
-    # sideways, rather than depth history. That said, its still size-of-history
-    # and should be addressed.
-    # mainline_revisions always includes an extra revision at the beginning, so
-    # don't request it.
-    parent_map = dict(((key, value) for key, value in
-        graph.iter_ancestry(mainline_revisions[1:]) if value is not None))
-    sorted_rev_list = tsort.topo_sort(parent_map.items())
-    text_keys = [(file_id, rev_id) for rev_id in sorted_rev_list]
-    modified_text_versions = branch.repository.texts.get_parent_map(text_keys)
-    ancestry = {}
-    for rev in sorted_rev_list:
-        text_key = (file_id, rev)
-        parents = parent_map[rev]
-        if text_key not in modified_text_versions and len(parents) == 1:
-            # We will not be adding anything new, so just use a reference to
-            # the parent ancestry.
-            rev_ancestry = ancestry[parents[0]]
+    # Lookup all possible text keys to determine which ones actually modified
+    # the file.
+    text_keys = [(file_id, rev_id) for rev_id, revno, depth in view_revisions]
+    # Looking up keys in batches of 1000 can cut the time in half, as well as
+    # memory consumption. GraphIndex *does* like to look for a few keys in
+    # parallel, it just doesn't like looking for *lots* of keys in parallel.
+    # TODO: This code needs to be re-evaluated periodically as we tune the
+    #       indexing layer. We might consider passing in hints as to the known
+    #       access pattern (sparse/clustered, high success rate/low success
+    #       rate). This particular access is clustered with a low success rate.
+    get_parent_map = branch.repository.texts.get_parent_map
+    modified_text_revisions = set()
+    chunk_size = 1000
+    for start in xrange(0, len(text_keys), chunk_size):
+        next_keys = text_keys[start:start + chunk_size]
+        # Only keep the revision_id portion of the key
+        modified_text_revisions.update(
+            [k[1] for k in get_parent_map(next_keys)])
+    del text_keys, next_keys
+
+    result = []
+    if direction == 'forward':
+        # TODO: The algorithm for finding 'merges' of file changes expects
+        #       'reverse' order (the default from 'merge_sort()'). Instead of
+        #       forcing this, we could just use the reverse_by_depth order.
+        view_revisions = reverse_by_depth(view_revisions)
+    # Track what revisions will merge the current revision, replace entries
+    # with 'None' when they have been added to result
+    current_merge_stack = [None]
+    for info in view_revisions:
+        rev_id, revno, depth = info
+        if depth == len(current_merge_stack):
+            current_merge_stack.append(info)
         else:
-            rev_ancestry = set()
-            if text_key in modified_text_versions:
-                rev_ancestry.add(rev)
-            for parent in parents:
-                if parent not in ancestry:
-                    # parent is a Ghost, which won't be present in
-                    # sorted_rev_list, but we may access it later, so create an
-                    # empty node for it
-                    ancestry[parent] = set()
-                rev_ancestry = rev_ancestry.union(ancestry[parent])
-        ancestry[rev] = rev_ancestry
-
-    def is_merging_rev(r):
-        parents = parent_map[r]
-        if len(parents) > 1:
-            leftparent = parents[0]
-            for rightparent in parents[1:]:
-                if not ancestry[leftparent].issuperset(
-                        ancestry[rightparent]):
-                    return True
-        return False
-
-    # filter from the view the revisions that did not change or merge 
-    # the specific file
-    return [(r, n, d) for r, n, d in view_revs_iter
-            if (file_id, r) in modified_text_versions or is_merging_rev(r)]
+            del current_merge_stack[depth + 1:]
+            current_merge_stack[-1] = info
+
+        if rev_id in modified_text_revisions:
+            # This needs to be logged, along with the extra revisions
+            for idx in xrange(len(current_merge_stack)):
+                node = current_merge_stack[idx]
+                if node is not None:
+                    result.append(node)
+                    current_merge_stack[idx] = None
+    if direction == 'forward':
+        result = reverse_by_depth(result)
+    return result
 
 
 def get_view_revisions(mainline_revs, rev_nos, branch, direction,

=== modified file 'bzrlib/memorytree.py'
--- bzrlib/memorytree.py	2008-09-02 00:29:28 +0000
+++ bzrlib/memorytree.py	2008-09-24 12:51:22 +0000
@@ -21,10 +21,12 @@
 
 
 from copy import deepcopy
+import os
 
 from bzrlib import (
     errors,
     mutabletree,
+    osutils,
     revision as _mod_revision,
     )
 from bzrlib.decorators import needs_read_lock, needs_write_lock
@@ -99,6 +101,14 @@
             return None, False, None
         return entry.kind, entry.executable, None
 
+    @needs_tree_write_lock
+    def rename_one(self, from_rel, to_rel):
+        file_id = self.path2id(from_rel)
+        to_dir, to_tail = os.path.split(to_rel)
+        to_parent_id = self.path2id(to_dir)
+        self._file_transport.move(from_rel, to_rel)
+        self._inventory.rename(file_id, to_parent_id, to_tail)
+
     def path_content_summary(self, path):
         """See Tree.path_content_summary."""
         id = self.path2id(path)

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2008-09-05 05:12:35 +0000
+++ bzrlib/merge.py	2008-09-24 12:53:22 +0000
@@ -28,6 +28,7 @@
     patiencediff,
     registry,
     revision as _mod_revision,
+    tree as _mod_tree,
     tsort,
     )
 from bzrlib.branch import Branch
@@ -96,6 +97,8 @@
         self._revision_graph = revision_graph
         self._base_is_ancestor = None
         self._base_is_other_ancestor = None
+        self._is_criss_cross = None
+        self._lca_trees = None
 
     @property
     def revision_graph(self):
@@ -353,15 +356,45 @@
                      ensure_null(self.other_basis)]
         if NULL_REVISION in revisions:
             self.base_rev_id = NULL_REVISION
+            self.base_tree = self.revision_tree(self.base_rev_id)
+            self._is_criss_cross = False
         else:
-            self.base_rev_id, steps = self.revision_graph.find_unique_lca(
-                revisions[0], revisions[1], count_steps=True)
+            lcas = self.revision_graph.find_lca(revisions[0], revisions[1])
+            self._is_criss_cross = False
+            if len(lcas) == 0:
+                self.base_rev_id = NULL_REVISION
+            elif len(lcas) == 1:
+                self.base_rev_id = list(lcas)[0]
+            else: # len(lcas) > 1
+                if len(lcas) > 2:
+                    # find_unique_lca can only handle 2 nodes, so we have to
+                    # start back at the beginning. It is a shame to traverse
+                    # the graph again, but better than re-implementing
+                    # find_unique_lca.
+                    self.base_rev_id = self.revision_graph.find_unique_lca(
+                                            revisions[0], revisions[1])
+                else:
+                    self.base_rev_id = self.revision_graph.find_unique_lca(
+                                            *lcas)
+                self._is_criss_cross = True
             if self.base_rev_id == NULL_REVISION:
                 raise UnrelatedBranches()
-            if steps > 1:
+            if self._is_criss_cross:
                 warning('Warning: criss-cross merge encountered.  See bzr'
                         ' help criss-cross.')
-        self.base_tree = self.revision_tree(self.base_rev_id)
+                interesting_revision_ids = [self.base_rev_id]
+                interesting_revision_ids.extend(lcas)
+                interesting_trees = dict((t.get_revision_id(), t)
+                    for t in self.this_branch.repository.revision_trees(
+                        interesting_revision_ids))
+                self._cached_trees.update(interesting_trees)
+                self.base_tree = interesting_trees.pop(self.base_rev_id)
+                sorted_lca_keys = self.revision_graph.find_merge_order(
+                    revisions[0], lcas)
+                self._lca_trees = [interesting_trees[key]
+                                   for key in sorted_lca_keys]
+            else:
+                self.base_tree = self.revision_tree(self.base_rev_id)
         self.base_is_ancestor = True
         self.base_is_other_ancestor = True
 
@@ -409,6 +442,9 @@
         if self.merge_type.supports_cherrypick:
             kwargs['cherrypick'] = (not self.base_is_ancestor or
                                     not self.base_is_other_ancestor)
+        if self._is_criss_cross and getattr(self.merge_type,
+                                            'supports_lca_trees', False):
+            kwargs['lca_trees'] = self._lca_trees
         return self.merge_type(pb=self._pb,
                                change_reporter=self.change_reporter,
                                **kwargs)
@@ -460,6 +496,23 @@
         return len(merge.cooked_conflicts)
 
 
+class _InventoryNoneEntry(object):
+    """This represents an inventory entry which *isn't there*.
+
+    It simplifies the merging logic if we always have an InventoryEntry, even
+    if it isn't actually present
+    """
+    executable = None
+    kind = None
+    name = None
+    parent_id = None
+    revision = None
+    symlink_target = None
+    text_sha1 = None
+
+_none_entry = _InventoryNoneEntry()
+
+
 class Merge3Merger(object):
     """Three-way merger that uses the merge3 text merger"""
     requires_base = True
@@ -469,12 +522,13 @@
     supports_cherrypick = True
     supports_reverse_cherrypick = True
     winner_idx = {"this": 2, "other": 1, "conflict": 1}
+    supports_lca_trees = True
 
     def __init__(self, working_tree, this_tree, base_tree, other_tree, 
                  interesting_ids=None, reprocess=False, show_base=False,
                  pb=DummyProgress(), pp=None, change_reporter=None,
                  interesting_files=None, do_merge=True,
-                 cherrypick=False):
+                 cherrypick=False, lca_trees=None):
         """Initialize the merger object and perform the merge.
 
         :param working_tree: The working tree to apply the merge to
@@ -496,6 +550,9 @@
             be combined with interesting_ids.  If neither interesting_files nor
             interesting_ids is specified, all files may participate in the
             merge.
+        :param lca_trees: Can be set to a dictionary of {revision_id:rev_tree}
+            if the ancestry was found to include a criss-cross merge.
+            Otherwise should be None.
         """
         object.__init__(self)
         if interesting_files is not None and interesting_ids is not None:
@@ -510,6 +567,12 @@
         self.cooked_conflicts = []
         self.reprocess = reprocess
         self.show_base = show_base
+        self._lca_trees = lca_trees
+        # Uncommenting this will change the default algorithm to always use
+        # _entries_lca. This can be useful for running the test suite and
+        # making sure we haven't missed any corner cases.
+        # if lca_trees is None:
+        #     self._lca_trees = [self.base_tree]
         self.pb = pb
         self.pp = pp
         self.change_reporter = change_reporter
@@ -556,19 +619,24 @@
         return self.tt
 
     def _compute_transform(self):
-        entries = self._entries3()
+        if self._lca_trees is None:
+            entries = self._entries3()
+            resolver = self._three_way
+        else:
+            entries = self._entries_lca()
+            resolver = self._lca_multi_way
         child_pb = ui.ui_factory.nested_progress_bar()
         try:
             for num, (file_id, changed, parents3, names3,
                       executable3) in enumerate(entries):
                 child_pb.update('Preparing file merge', num, len(entries))
-                self._merge_names(file_id, parents3, names3)
+                self._merge_names(file_id, parents3, names3, resolver=resolver)
                 if changed:
                     file_status = self.merge_contents(file_id)
                 else:
                     file_status = 'unmodified'
                 self._merge_executable(file_id,
-                    executable3, file_status)
+                    executable3, file_status, resolver=resolver)
         finally:
             child_pb.finished()
         self.fix_root()
@@ -620,6 +688,177 @@
             result.append((file_id, changed, parents3, names3, executable3))
         return result
 
+    def _entries_lca(self):
+        """Gather data about files modified between multiple trees.
+
+        This compares OTHER versus all LCA trees, and for interesting entries,
+        it then compares with THIS and BASE.
+
+        For the multi-valued entries, the format will be (BASE, [lca1, lca2])
+        :return: [(file_id, changed, parents, names, executable)]
+            file_id     Simple file_id of the entry
+            changed     Boolean, True if the kind or contents changed
+                        else False
+            parents     ((base, [parent_id, in, lcas]), parent_id_other,
+                         parent_id_this)
+            names       ((base, [name, in, lcas]), name_in_other, name_in_this)
+            executable  ((base, [exec, in, lcas]), exec_in_other, exec_in_this)
+        """
+        if self.interesting_files is not None:
+            lookup_trees = [self.this_tree, self.base_tree]
+            lookup_trees.extend(self._lca_trees)
+            # I think we should include the lca trees as well
+            interesting_ids = self.other_tree.paths2ids(self.interesting_files,
+                                                        lookup_trees)
+        else:
+            interesting_ids = self.interesting_ids
+        result = []
+        walker = _mod_tree.MultiWalker(self.other_tree, self._lca_trees)
+
+        base_inventory = self.base_tree.inventory
+        this_inventory = self.this_tree.inventory
+        for path, file_id, other_ie, lca_values in walker.iter_all():
+            # Is this modified at all from any of the other trees?
+            if other_ie is None:
+                other_ie = _none_entry
+            if interesting_ids is not None and file_id not in interesting_ids:
+                continue
+
+            # If other_revision is found in any of the lcas, that means this
+            # node is uninteresting. This is because when merging, if there are
+            # multiple heads(), we have to create a new node. So if we didn't,
+            # we know that the ancestry is linear, and that OTHER did not
+            # modify anything
+            # See doc/developers/lca_merge_resolution.txt for details
+            other_revision = other_ie.revision
+            if other_revision is not None:
+                # We can't use this shortcut when other_revision is None,
+                # because it may be None because things are WorkingTrees, and
+                # not because it is *actually* None.
+                is_unmodified = False
+                for lca_path, ie in lca_values:
+                    if ie is not None and ie.revision == other_revision:
+                        is_unmodified = True
+                        break
+                if is_unmodified:
+                    continue
+
+            lca_entries = []
+            for lca_path, lca_ie in lca_values:
+                if lca_ie is None:
+                    lca_entries.append(_none_entry)
+                else:
+                    lca_entries.append(lca_ie)
+
+            if file_id in base_inventory:
+                base_ie = base_inventory[file_id]
+            else:
+                base_ie = _none_entry
+
+            if file_id in this_inventory:
+                this_ie = this_inventory[file_id]
+            else:
+                this_ie = _none_entry
+
+            lca_kinds = []
+            lca_parent_ids = []
+            lca_names = []
+            lca_executable = []
+            for lca_ie in lca_entries:
+                lca_kinds.append(lca_ie.kind)
+                lca_parent_ids.append(lca_ie.parent_id)
+                lca_names.append(lca_ie.name)
+                lca_executable.append(lca_ie.executable)
+
+            kind_winner = self._lca_multi_way(
+                (base_ie.kind, lca_kinds),
+                other_ie.kind, this_ie.kind)
+            parent_id_winner = self._lca_multi_way(
+                (base_ie.parent_id, lca_parent_ids),
+                other_ie.parent_id, this_ie.parent_id)
+            name_winner = self._lca_multi_way(
+                (base_ie.name, lca_names),
+                other_ie.name, this_ie.name)
+
+            content_changed = True
+            if kind_winner == 'this':
+                # No kind change in OTHER, see if there are *any* changes
+                if other_ie.kind == None:
+                    # No content and 'this' wins the kind, so skip this?
+                    # continue
+                    pass
+                elif other_ie.kind == 'directory':
+                    if parent_id_winner == 'this' and name_winner == 'this':
+                        # No change for this directory in OTHER, skip
+                        continue
+                    content_changed = False
+                elif other_ie.kind == 'file':
+                    def get_sha1(ie, tree):
+                        if ie.kind != 'file':
+                            return None
+                        return tree.get_file_sha1(file_id)
+                    base_sha1 = get_sha1(base_ie, self.base_tree)
+                    lca_sha1s = [get_sha1(ie, tree) for ie, tree
+                                 in zip(lca_entries, self._lca_trees)]
+                    this_sha1 = get_sha1(this_ie, self.this_tree)
+                    other_sha1 = get_sha1(other_ie, self.other_tree)
+                    sha1_winner = self._lca_multi_way(
+                        (base_sha1, lca_sha1s), other_sha1, this_sha1,
+                        allow_overriding_lca=False)
+                    exec_winner = self._lca_multi_way(
+                        (base_ie.executable, lca_executable),
+                        other_ie.executable, this_ie.executable)
+                    if (parent_id_winner == 'this' and name_winner == 'this'
+                        and sha1_winner == 'this' and exec_winner == 'this'):
+                        # No kind, parent, name, exec, or content change for
+                        # OTHER, so this node is not considered interesting
+                        continue
+                    if sha1_winner == 'this':
+                        content_changed = False
+                elif other_ie.kind == 'symlink':
+                    def get_target(ie, tree):
+                        if ie.kind != 'symlink':
+                            return None
+                        return tree.get_symlink_target(file_id)
+                    base_target = get_target(base_ie, self.base_tree)
+                    lca_targets = [get_target(ie, tree) for ie, tree
+                                   in zip(lca_entries, self._lca_trees)]
+                    this_target = get_target(this_ie, self.this_tree)
+                    other_target = get_target(other_ie, self.other_tree)
+                    target_winner = self._lca_multi_way(
+                        (base_target, lca_targets),
+                        other_target, this_target)
+                    if (parent_id_winner == 'this' and name_winner == 'this'
+                        and target_winner == 'this'):
+                        # No kind, parent, name, or symlink target change
+                        # not interesting
+                        continue
+                    if target_winner == 'this':
+                        content_changed = False
+                elif other_ie.kind == 'tree-reference':
+                    # The 'changed' information seems to be handled at a higher
+                    # level. At least, _entries3 returns False for content
+                    # changed, even when at a new revision_id.
+                    content_changed = False
+                    if (parent_id_winner == 'this' and name_winner == 'this'):
+                        # Nothing interesting
+                        continue
+                else:
+                    raise AssertionError('unhandled kind: %s' % other_ie.kind)
+                # XXX: We need to handle kind == 'symlink'
+
+            # If we have gotten this far, that means something has changed
+            result.append((file_id, content_changed,
+                           ((base_ie.parent_id, lca_parent_ids),
+                            other_ie.parent_id, this_ie.parent_id),
+                           ((base_ie.name, lca_names),
+                            other_ie.name, this_ie.name),
+                           ((base_ie.executable, lca_executable),
+                            other_ie.executable, this_ie.executable)
+                          ))
+        return result
+
+
     def fix_root(self):
         try:
             self.tt.final_kind(self.tt.root)
@@ -714,6 +953,56 @@
             return "other"
 
     @staticmethod
+    def _lca_multi_way(bases, other, this, allow_overriding_lca=True):
+        """Consider LCAs when determining whether a change has occurred.
+
+        If LCAS are all identical, this is the same as a _three_way comparison.
+
+        :param bases: value in (BASE, [LCAS])
+        :param other: value in OTHER
+        :param this: value in THIS
+        :param allow_overriding_lca: If there is more than one unique lca
+            value, allow OTHER to override THIS if it has a new value, and
+            THIS only has an lca value, or vice versa. This is appropriate for
+            truly scalar values, not as much for non-scalars.
+        :return: 'this', 'other', or 'conflict' depending on whether an entry
+            changed or not.
+        """
+        # See doc/developers/lca_merge_resolution.txt for details about this
+        # algorithm.
+        if other == this:
+            # Either Ambiguously clean, or nothing was actually changed. We
+            # don't really care
+            return 'this'
+        base_val, lca_vals = bases
+        # Remove 'base_val' from the lca_vals, because it is not interesting
+        filtered_lca_vals = [lca_val for lca_val in lca_vals
+                                      if lca_val != base_val]
+        if len(filtered_lca_vals) == 0:
+            return Merge3Merger._three_way(base_val, other, this)
+
+        unique_lca_vals = set(filtered_lca_vals)
+        if len(unique_lca_vals) == 1:
+            return Merge3Merger._three_way(unique_lca_vals.pop(), other, this)
+
+        if allow_overriding_lca:
+            if other in unique_lca_vals:
+                if this in unique_lca_vals:
+                    # Each side picked a different lca, conflict
+                    return 'conflict'
+                else:
+                    # This has a value which supersedes both lca values, and
+                    # other only has an lca value
+                    return 'this'
+            elif this in unique_lca_vals:
+                # OTHER has a value which supersedes both lca values, and this
+                # only has an lca value
+                return 'other'
+
+        # At this point, the lcas disagree, and the tips disagree
+        return 'conflict'
+
+    @staticmethod
     def scalar_three_way(this_tree, base_tree, other_tree, file_id, key):
         """Do a three-way test on a scalar.
         Return "this", "other" or "conflict", depending whether a value wins.
@@ -751,16 +1040,17 @@
             else:
                 names.append(entry.name)
                 parents.append(entry.parent_id)
-        return self._merge_names(file_id, parents, names)
+        return self._merge_names(file_id, parents, names,
+                                 resolver=self._three_way)
 
-    def _merge_names(self, file_id, parents, names):
+    def _merge_names(self, file_id, parents, names, resolver):
         """Perform a merge on file_id names and parents"""
         base_name, other_name, this_name = names
         base_parent, other_parent, this_parent = parents
 
-        name_winner = self._three_way(*names)
+        name_winner = resolver(*names)
 
-        parent_id_winner = self._three_way(*parents)
+        parent_id_winner = resolver(*parents)
         if this_name is None:
             if name_winner == "this":
                 name_winner = "other"
@@ -954,14 +1244,16 @@
         """Perform a merge on the execute bit."""
         executable = [self.executable(t, file_id) for t in (self.base_tree,
                       self.other_tree, self.this_tree)]
-        self._merge_executable(file_id, executable, file_status)
+        self._merge_executable(file_id, executable, file_status,
+                               resolver=self._three_way)
 
-    def _merge_executable(self, file_id, executable, file_status):
+    def _merge_executable(self, file_id, executable, file_status,
+                          resolver):
         """Perform a merge on the execute bit."""
         base_executable, other_executable, this_executable = executable
         if file_status == "deleted":
             return
-        winner = self._three_way(*executable)
+        winner = resolver(*executable)
         if winner == "conflict":
         # There must be a None in here, if we have a conflict, but we
         # need executability since file status was not deleted.

=== modified file 'bzrlib/mutabletree.py'
--- bzrlib/mutabletree.py	2008-05-08 04:33:38 +0000
+++ bzrlib/mutabletree.py	2008-09-22 05:15:20 +0000
@@ -201,6 +201,20 @@
         """Helper function for add - sets the entries of kinds."""
         raise NotImplementedError(self._gather_kinds)
 
+    def get_file_with_stat(self, file_id, path=None):
+        """Get a file handle and stat object for file_id.
+
+        The default implementation returns (self.get_file, None) for backwards
+        compatibility.
+
+        :param file_id: The file id to read.
+        :param path: The path of the file, if it is known.
+        :return: A tuple (file_handle, stat_value_or_None). If the tree has
+            no stat facility, or need for a stat cache feedback during commit,
+            it may return None for the second element of the tuple.
+        """
+        return (self.get_file(file_id, path), None)
+
     @needs_read_lock
     def last_revision(self):
         """Return the revision id of the last commit performed in this tree.
@@ -247,6 +261,23 @@
         """
         raise NotImplementedError(self.mkdir)
 
+    def _observed_sha1(self, file_id, path, (sha1, stat_value)):
+        """Tell the tree we have observed a paths sha1.
+
+        The intent of this function is to allow trees that have a hashcache to
+        update the hashcache during commit. If the observed file is too new
+        (based on the stat_value) to be safely hash-cached the tree will ignore
+        it. 
+
+        The default implementation does nothing.
+
+        :param file_id: The file id
+        :param path: The file path
+        :param sha1: The sha 1 that was observed.
+        :param stat_value: A stat result for the file the sha1 was read from.
+        :return: None
+        """
+
     @needs_write_lock
     def put_file_bytes_non_atomic(self, file_id, bytes):
         """Update the content of a file in the tree.

=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2008-09-26 07:09:50 +0000
+++ bzrlib/osutils.py	2008-09-29 07:09:43 +0000
@@ -55,10 +55,10 @@
 # sha and md5 modules are deprecated in python2.6 but hashlib is available as
 # of 2.5
 if sys.version_info < (2, 5):
-    import md5 as _md5
-    md5 = _md5.new
-    import sha as _sha
-    sha = _sha.new
+    import md5 as _mod_md5
+    md5 = _mod_md5.new
+    import sha as _mod_sha
+    sha = _mod_sha.new
 else:
     from hashlib import (
         md5,

=== modified file 'bzrlib/python-compat.h'
--- bzrlib/python-compat.h	2008-09-24 04:15:03 +0000
+++ bzrlib/python-compat.h	2008-09-26 05:14:51 +0000
@@ -25,12 +25,33 @@
 #ifndef _BZR_PYTHON_COMPAT_H
 #define _BZR_PYTHON_COMPAT_H
 
-#if PY_VERSION_HEX < 0x02050000
-  typedef int Py_ssize_t;
+/* http://www.python.org/dev/peps/pep-0353/ */
+#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
+    typedef int Py_ssize_t;
     #define PY_SSIZE_T_MAX INT_MAX
     #define PY_SSIZE_T_MIN INT_MIN
     #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
     #define PyInt_AsSsize_t(o) PyInt_AsLong(o)
 #endif
 
+#if defined(_WIN32) || defined(WIN32)
+    /* Needed for htonl */
+    #include "Winsock.h"
+
+    /* Defining WIN32_LEAN_AND_MEAN makes including windows quite a bit
+     * lighter weight.
+     */
+    #define WIN32_LEAN_AND_MEAN
+    #include <windows.h>
+
+    /* sys/stat.h doesn't have S_ISLNK on win32, so we fake it by just always
+     * returning False
+     */
+    #define S_ISLNK(mode) (0)
+#else /* Not win32 */
+    /* For htonl */
+    #include "arpa/inet.h"
 #endif
+
+
+#endif /* _BZR_PYTHON_COMPAT_H */

=== modified file 'bzrlib/repofmt/pack_repo.py'
--- bzrlib/repofmt/pack_repo.py	2008-09-26 07:09:50 +0000
+++ bzrlib/repofmt/pack_repo.py	2008-09-29 07:03:55 +0000
@@ -54,6 +54,14 @@
     )
 
 from bzrlib.decorators import needs_write_lock
+from bzrlib.btree_index import (
+    BTreeGraphIndex,
+    BTreeBuilder,
+    )
+from bzrlib.index import (
+    GraphIndex,
+    InMemoryGraphIndex,
+    )
 from bzrlib.repofmt.knitrepo import KnitRepository
 from bzrlib.repository import (
     CommitBuilder,
@@ -215,7 +223,8 @@
         }
 
     def __init__(self, upload_transport, index_transport, pack_transport,
-        upload_suffix='', file_mode=None):
+        upload_suffix='', file_mode=None, index_builder_class=None,
+        index_class=None):
         """Create a NewPack instance.
 
         :param upload_transport: A writable transport for the pack to be
@@ -228,24 +237,30 @@
         :param upload_suffix: An optional suffix to be given to any temporary
             files created during the pack creation. e.g '.autopack'
         :param file_mode: An optional file mode to create the new files with.
+        :param index_builder_class: Required keyword parameter - the class of
+            index builder to use.
+        :param index_class: Required keyword parameter - the class of index
+            object to use.
         """
         # The relative locations of the packs are constrained, but all are
         # passed in because the caller has them, so as to avoid object churn.
         Pack.__init__(self,
             # Revisions: parents list, no text compression.
-            InMemoryGraphIndex(reference_lists=1),
+            index_builder_class(reference_lists=1),
             # Inventory: We want to map compression only, but currently the
             # knit code hasn't been updated enough to understand that, so we
             # have a regular 2-list index giving parents and compression
             # source.
-            InMemoryGraphIndex(reference_lists=2),
+            index_builder_class(reference_lists=2),
             # Texts: compression and per file graph, for all fileids - so two
             # reference lists and two elements in the key tuple.
-            InMemoryGraphIndex(reference_lists=2, key_elements=2),
+            index_builder_class(reference_lists=2, key_elements=2),
             # Signatures: Just blobs to store, no compression, no parents
             # listing.
-            InMemoryGraphIndex(reference_lists=0),
+            index_builder_class(reference_lists=0),
             )
+        # When we make readonly indices, we need this.
+        self.index_class = index_class
         # where should the new pack be opened
         self.upload_transport = upload_transport
         # where are indices written out to
@@ -392,7 +407,7 @@
 
     def _replace_index_with_readonly(self, index_type):
         setattr(self, index_type + '_index',
-            GraphIndex(self.index_transport,
+            self.index_class(self.index_transport,
                 self.index_name(index_type, self.name),
                 self.index_sizes[self.index_offset(index_type)]))
 
@@ -591,7 +606,9 @@
         return NewPack(self._pack_collection._upload_transport,
             self._pack_collection._index_transport,
             self._pack_collection._pack_transport, upload_suffix=self.suffix,
-            file_mode=self._pack_collection.repo.bzrdir._get_file_mode())
+            file_mode=self._pack_collection.repo.bzrdir._get_file_mode(),
+            index_builder_class=self._pack_collection._index_builder_class,
+            index_class=self._pack_collection._index_class)
 
     def _copy_revision_texts(self):
         """Copy revision data to the new pack."""
@@ -1104,7 +1121,7 @@
     """
 
     def __init__(self, repo, transport, index_transport, upload_transport,
-                 pack_transport):
+                 pack_transport, index_builder_class, index_class):
         """Create a new RepositoryPackCollection.
 
         :param transport: Addresses the repository base directory 
@@ -1113,12 +1130,16 @@
         :param upload_transport: Addresses the directory into which packs are written
             while they're being created.
         :param pack_transport: Addresses the directory of existing complete packs.
+        :param index_builder_class: The index builder class to use.
+        :param index_class: The index class to use.
         """
         self.repo = repo
         self.transport = transport
         self._index_transport = index_transport
         self._upload_transport = upload_transport
         self._pack_transport = pack_transport
+        self._index_builder_class = index_builder_class
+        self._index_class = index_class
         self._suffix_offsets = {'.rix': 0, '.iix': 1, '.tix': 2, '.six': 3}
         self.packs = []
         # name:Pack mapping
@@ -1275,8 +1296,8 @@
         # plan out what packs to keep, and what to reorganise
         while len(existing_packs):
             # take the largest pack, and if its less than the head of the
-            # distribution chart we will include its contents in the new pack for
-            # that position. If its larger, we remove its size from the
+            # distribution chart we will include its contents in the new pack
+            # for that position. If its larger, we remove its size from the
             # distribution chart
             next_pack_rev_count, next_pack = existing_packs.pop(0)
             if next_pack_rev_count >= pack_distribution[0]:
@@ -1299,8 +1320,18 @@
                     # this pack is used up, shift left.
                     del pack_distribution[0]
                     pack_operations.append([0, []])
-        
-        return pack_operations
+        # Now that we know which pack files we want to move, shove them all
+        # into a single pack file.
+        final_rev_count = 0
+        final_pack_list = []
+        for num_revs, pack_files in pack_operations:
+            final_rev_count += num_revs
+            final_pack_list.extend(pack_files)
+        if len(final_pack_list) == 1:
+            raise AssertionError('We somehow generated an autopack with a'
+                ' single pack file being moved.')
+            return []
+        return [[final_rev_count, final_pack_list]]
 
     def ensure_loaded(self):
         # NB: if you see an assertion error here, its probably access against
@@ -1359,14 +1390,14 @@
         detect updates from others during our write operation.
         :return: An iterator of the index contents.
         """
-        return GraphIndex(self.transport, 'pack-names', None
+        return self._index_class(self.transport, 'pack-names', None
                 ).iter_all_entries()
 
     def _make_index(self, name, suffix):
         size_offset = self._suffix_offsets[suffix]
         index_name = name + suffix
         index_size = self._names[name][size_offset]
-        return GraphIndex(
+        return self._index_class(
             self._index_transport, index_name, index_size)
 
     def _max_pack_count(self, total_revisions):
@@ -1535,7 +1566,7 @@
         """
         self.lock_names()
         try:
-            builder = GraphIndexBuilder()
+            builder = self._index_builder_class()
             # load the disk nodes across
             disk_nodes = set()
             for index, key, value in self._iter_disk_pack_index():
@@ -1607,7 +1638,9 @@
             raise errors.NotWriteLocked(self)
         self._new_pack = NewPack(self._upload_transport, self._index_transport,
             self._pack_transport, upload_suffix='.pack',
-            file_mode=self.repo.bzrdir._get_file_mode())
+            file_mode=self.repo.bzrdir._get_file_mode(),
+            index_builder_class=self._index_builder_class,
+            index_class=self._index_class)
         # allow writing: queue writes to a new index
         self.revision_index.add_writable_index(self._new_pack.revision_index,
             self._new_pack)
@@ -1682,7 +1715,9 @@
         self._pack_collection = RepositoryPackCollection(self, self._transport,
             index_transport,
             self._transport.clone('upload'),
-            self._transport.clone('packs'))
+            self._transport.clone('packs'),
+            _format.index_builder_class,
+            _format.index_class)
         self.inventories = KnitVersionedFiles(
             _KnitGraphIndex(self._pack_collection.inventory_index.combined_index,
                 add_callback=self._pack_collection.inventory_index.add_callback,
@@ -1906,6 +1941,9 @@
     _serializer = None
     # External references are not supported in pack repositories yet.
     supports_external_lookups = False
+    # What index classes to use
+    index_builder_class = None
+    index_class = None
 
     def initialize(self, a_bzrdir, shared=False):
         """Create a pack based repository.
@@ -1917,7 +1955,7 @@
         """
         mutter('creating repository in %s.', a_bzrdir.transport.base)
         dirs = ['indices', 'obsolete_packs', 'packs', 'upload']
-        builder = GraphIndexBuilder()
+        builder = self.index_builder_class()
         files = [('pack-names', builder.finish())]
         utf8_files = [('format', self.get_format_string())]
         
@@ -1955,6 +1993,9 @@
     repository_class = KnitPackRepository
     _commit_builder_class = PackCommitBuilder
     _serializer = xml5.serializer_v5
+    # What index classes to use
+    index_builder_class = InMemoryGraphIndex
+    index_class = GraphIndex
 
     def _get_matching_bzrdir(self):
         return bzrdir.format_registry.make_bzrdir('pack-0.92')
@@ -1991,6 +2032,9 @@
     rich_root_data = True
     supports_tree_reference = True
     _serializer = xml7.serializer_v7
+    # What index classes to use
+    index_builder_class = InMemoryGraphIndex
+    index_class = GraphIndex
 
     def _get_matching_bzrdir(self):
         return bzrdir.format_registry.make_bzrdir(
@@ -2032,6 +2076,9 @@
     rich_root_data = True
     supports_tree_reference = False
     _serializer = xml6.serializer_v6
+    # What index classes to use
+    index_builder_class = InMemoryGraphIndex
+    index_class = GraphIndex
 
     def _get_matching_bzrdir(self):
         return bzrdir.format_registry.make_bzrdir(
@@ -2070,9 +2117,12 @@
     _commit_builder_class = PackCommitBuilder
     _serializer = xml5.serializer_v5
     supports_external_lookups = True
+    # What index classes to use
+    index_builder_class = InMemoryGraphIndex
+    index_class = GraphIndex
 
     def _get_matching_bzrdir(self):
-        return bzrdir.format_registry.make_bzrdir('development1')
+        return bzrdir.format_registry.make_bzrdir('1.6')
 
     def _ignore_setting_bzrdir(self, format):
         pass
@@ -2106,6 +2156,9 @@
     supports_tree_reference = False # no subtrees
     _serializer = xml6.serializer_v6
     supports_external_lookups = True
+    # What index classes to use
+    index_builder_class = InMemoryGraphIndex
+    index_class = GraphIndex
 
     def _get_matching_bzrdir(self):
         return bzrdir.format_registry.make_bzrdir(
@@ -2149,10 +2202,13 @@
     _serializer = xml7.serializer_v7
 
     supports_external_lookups = True
+    # What index classes to use
+    index_builder_class = InMemoryGraphIndex
+    index_class = GraphIndex
 
     def _get_matching_bzrdir(self):
         return bzrdir.format_registry.make_bzrdir(
-            'development1-subtree')
+            '1.6.1-rich-root')
 
     def _ignore_setting_bzrdir(self, format):
         pass
@@ -2173,22 +2229,24 @@
                 " (deprecated)")
 
 
-class RepositoryFormatPackDevelopment1(RepositoryFormatPack):
+class RepositoryFormatPackDevelopment2(RepositoryFormatPack):
     """A no-subtrees development repository.
 
-    This format should be retained until the second release after bzr 1.5.
+    This format should be retained until the second release after bzr 1.7.
 
-    Supports external lookups, which results in non-truncated ghosts after
-    reconcile compared to pack-0.92 formats.
+    This is pack-1.6.1 with B+Tree indices.
     """
 
     repository_class = KnitPackRepository
     _commit_builder_class = PackCommitBuilder
     _serializer = xml5.serializer_v5
     supports_external_lookups = True
+    # What index classes to use
+    index_builder_class = BTreeBuilder
+    index_class = BTreeGraphIndex
 
     def _get_matching_bzrdir(self):
-        return bzrdir.format_registry.make_bzrdir('development1')
+        return bzrdir.format_registry.make_bzrdir('development2')
 
     def _ignore_setting_bzrdir(self, format):
         pass
@@ -2197,24 +2255,23 @@
 
     def get_format_string(self):
         """See RepositoryFormat.get_format_string()."""
-        return "Bazaar development format 1 (needs bzr.dev from before 1.6)\n"
+        return "Bazaar development format 2 (needs bzr.dev from before 1.8)\n"
 
     def get_format_description(self):
         """See RepositoryFormat.get_format_description()."""
         return ("Development repository format, currently the same as "
-            "pack-0.92 with external reference support.\n")
+            "1.6.1 with B+Trees.\n")
 
     def check_conversion_target(self, target_format):
         pass
 
 
-class RepositoryFormatPackDevelopment1Subtree(RepositoryFormatPack):
+class RepositoryFormatPackDevelopment2Subtree(RepositoryFormatPack):
     """A subtrees development repository.
 
-    This format should be retained until the second release after bzr 1.5.
+    This format should be retained until the second release after bzr 1.7.
 
-    Supports external lookups, which results in non-truncated ghosts after
-    reconcile compared to pack-0.92 formats.
+    1.6.1-subtree[as it might have been] with B+Tree indices.
     """
 
     repository_class = KnitPackRepository
@@ -2223,10 +2280,13 @@
     supports_tree_reference = True
     _serializer = xml7.serializer_v7
     supports_external_lookups = True
+    # What index classes to use
+    index_builder_class = BTreeBuilder
+    index_class = BTreeGraphIndex
 
     def _get_matching_bzrdir(self):
         return bzrdir.format_registry.make_bzrdir(
-            'development1-subtree')
+            'development2-subtree')
 
     def _ignore_setting_bzrdir(self, format):
         pass
@@ -2243,10 +2303,10 @@
             
     def get_format_string(self):
         """See RepositoryFormat.get_format_string()."""
-        return ("Bazaar development format 1 with subtree support "
-            "(needs bzr.dev from before 1.6)\n")
+        return ("Bazaar development format 2 with subtree support "
+            "(needs bzr.dev from before 1.8)\n")
 
     def get_format_description(self):
         """See RepositoryFormat.get_format_description()."""
         return ("Development repository format, currently the same as "
-            "pack-0.92-subtree with external reference support.\n")
+            "1.6.1-subtree with B+Tree indices.\n")

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-09-04 14:22:34 +0000
+++ bzrlib/repository.py	2008-09-27 22:58:34 +0000
@@ -241,12 +241,16 @@
             content - stat, length, exec, sha/link target. This is only
             accessed when the entry has a revision of None - that is when it is
             a candidate to commit.
-        :return: A tuple (change_delta, version_recorded). change_delta is 
-            an inventory_delta change for this entry against the basis tree of
-            the commit, or None if no change occured against the basis tree.
+        :return: A tuple (change_delta, version_recorded, fs_hash).
+            change_delta is an inventory_delta change for this entry against
+            the basis tree of the commit, or None if no change occured against
+            the basis tree.
             version_recorded is True if a new version of the entry has been
             recorded. For instance, committing a merge where a file was only
             changed on the other side will return (delta, False).
+            fs_hash is either None, or the hash details for the path (currently
+            a tuple of the contents sha1 and the statvalue returned by
+            tree.get_file_with_stat()).
         """
         if self.new_inventory.root is None:
             if ie.parent_id is not None:
@@ -287,7 +291,7 @@
                 else:
                     # add
                     delta = (None, path, ie.file_id, ie)
-                return delta, False
+                return delta, False, None
             else:
                 # we don't need to commit this, because the caller already
                 # determined that an existing revision of this file is
@@ -298,7 +302,7 @@
                     raise AssertionError("Impossible situation, a skipped "
                         "inventory entry (%r) claims to be modified in this "
                         "commit (%r).", (ie, self._new_revision_id))
-                return None, False
+                return None, False, None
         # XXX: Friction: parent_candidates should return a list not a dict
         #      so that we don't have to walk the inventories again.
         parent_candiate_entries = ie.parent_candidates(parent_invs)
@@ -334,6 +338,9 @@
             # if the kind changed the content obviously has
             if kind != parent_entry.kind:
                 store = True
+        # Stat cache fingerprint feedback for the caller - None as we usually
+        # don't generate one.
+        fingerprint = None
         if kind == 'file':
             if content_summary[2] is None:
                 raise ValueError("Files must not have executable = None")
@@ -350,7 +357,7 @@
                     ie.text_size = parent_entry.text_size
                     ie.text_sha1 = parent_entry.text_sha1
                     ie.executable = parent_entry.executable
-                    return self._get_delta(ie, basis_inv, path), False
+                    return self._get_delta(ie, basis_inv, path), False, None
                 else:
                     # Either there is only a hash change(no hash cache entry,
                     # or same size content change), or there is no change on
@@ -363,10 +370,16 @@
                 # absence of a content change in the file.
                 nostore_sha = None
             ie.executable = content_summary[2]
-            lines = tree.get_file(ie.file_id, path).readlines()
+            file_obj, stat_value = tree.get_file_with_stat(ie.file_id, path)
+            try:
+                lines = file_obj.readlines()
+            finally:
+                file_obj.close()
             try:
                 ie.text_sha1, ie.text_size = self._add_text_to_weave(
                     ie.file_id, lines, heads, nostore_sha)
+                # Let the caller know we generated a stat fingerprint.
+                fingerprint = (ie.text_sha1, stat_value)
             except errors.ExistingContent:
                 # Turns out that the file content was unchanged, and we were
                 # only going to store a new node if it was changed. Carry over
@@ -375,13 +388,13 @@
                 ie.text_size = parent_entry.text_size
                 ie.text_sha1 = parent_entry.text_sha1
                 ie.executable = parent_entry.executable
-                return self._get_delta(ie, basis_inv, path), False
+                return self._get_delta(ie, basis_inv, path), False, None
         elif kind == 'directory':
             if not store:
                 # all data is meta here, nothing specific to directory, so
                 # carry over:
                 ie.revision = parent_entry.revision
-                return self._get_delta(ie, basis_inv, path), False
+                return self._get_delta(ie, basis_inv, path), False, None
             lines = []
             self._add_text_to_weave(ie.file_id, lines, heads, None)
         elif kind == 'symlink':
@@ -395,7 +408,7 @@
                 # unchanged, carry over.
                 ie.revision = parent_entry.revision
                 ie.symlink_target = parent_entry.symlink_target
-                return self._get_delta(ie, basis_inv, path), False
+                return self._get_delta(ie, basis_inv, path), False, None
             ie.symlink_target = current_link_target
             lines = []
             self._add_text_to_weave(ie.file_id, lines, heads, None)
@@ -407,14 +420,14 @@
                 # unchanged, carry over.
                 ie.reference_revision = parent_entry.reference_revision
                 ie.revision = parent_entry.revision
-                return self._get_delta(ie, basis_inv, path), False
+                return self._get_delta(ie, basis_inv, path), False, None
             ie.reference_revision = content_summary[3]
             lines = []
             self._add_text_to_weave(ie.file_id, lines, heads, None)
         else:
             raise NotImplementedError('unknown kind')
         ie.revision = self._new_revision_id
-        return self._get_delta(ie, basis_inv, path), True
+        return self._get_delta(ie, basis_inv, path), True, fingerprint
 
     def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
         # Note: as we read the content directly from the tree, we know its not
@@ -2319,19 +2332,18 @@
     )
 
 # Development formats. 
-# 1.5->1.6
-format_registry.register_lazy(
-    "Bazaar development format 1 (needs bzr.dev from before 1.6)\n",
-    'bzrlib.repofmt.pack_repo',
-    'RepositoryFormatPackDevelopment1',
-    )
-format_registry.register_lazy(
-    ("Bazaar development format 1 with subtree support "
-        "(needs bzr.dev from before 1.6)\n"),
-    'bzrlib.repofmt.pack_repo',
-    'RepositoryFormatPackDevelopment1Subtree',
-    )
-# 1.6->1.7 go below here
+# 1.7->1.8 go below here
+format_registry.register_lazy(
+    "Bazaar development format 2 (needs bzr.dev from before 1.8)\n",
+    'bzrlib.repofmt.pack_repo',
+    'RepositoryFormatPackDevelopment2',
+    )
+format_registry.register_lazy(
+    ("Bazaar development format 2 with subtree support "
+        "(needs bzr.dev from before 1.8)\n"),
+    'bzrlib.repofmt.pack_repo',
+    'RepositoryFormatPackDevelopment2Subtree',
+    )
 
 
 class InterRepository(InterObject):
@@ -2915,13 +2927,13 @@
                 RepositoryFormatKnitPack4,
                 RepositoryFormatKnitPack5,
                 RepositoryFormatKnitPack5RichRoot,
-                RepositoryFormatPackDevelopment1,
-                RepositoryFormatPackDevelopment1Subtree,
+                RepositoryFormatPackDevelopment2,
+                RepositoryFormatPackDevelopment2Subtree,
                 )
             norichroot = (
                 RepositoryFormatKnit1,            # no rr, no subtree
                 RepositoryFormatKnitPack1,        # no rr, no subtree
-                RepositoryFormatPackDevelopment1, # no rr, no subtree
+                RepositoryFormatPackDevelopment2, # no rr, no subtree
                 RepositoryFormatKnitPack5,        # no rr, no subtree
                 )
             richroot = (
@@ -2929,7 +2941,7 @@
                 RepositoryFormatKnitPack3,        # rr, subtree
                 RepositoryFormatKnitPack4,        # rr, no subtree
                 RepositoryFormatKnitPack5RichRoot,# rr, no subtree
-                RepositoryFormatPackDevelopment1Subtree, # rr, subtree
+                RepositoryFormatPackDevelopment2Subtree, # rr, subtree
                 )
             for format in norichroot:
                 if format.rich_root_data:

=== modified file 'bzrlib/revisionspec.py'
--- bzrlib/revisionspec.py	2008-09-25 16:02:34 +0000
+++ bzrlib/revisionspec.py	2008-09-29 07:09:43 +0000
@@ -137,17 +137,6 @@
     prefix = None
     wants_revision_history = True
 
-    def __new__(cls, spec, _internal=False):
-        if _internal:
-            return object.__new__(cls)
-
-        symbol_versioning.warn('Creating a RevisionSpec directly has'
-                               ' been deprecated in version 0.11. Use'
-                               ' RevisionSpec.from_string()'
-                               ' instead.',
-                               DeprecationWarning, stacklevel=2)
-        return RevisionSpec.from_string(spec)
-
     @staticmethod
     def from_string(spec):
         """Parse a revision spec string into a RevisionSpec object.

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2008-09-23 16:56:38 +0000
+++ bzrlib/tests/__init__.py	2008-09-25 01:11:30 +0000
@@ -877,6 +877,21 @@
         self.assertEqual(mode, mode_test,
                          'mode mismatch %o != %o' % (mode, mode_test))
 
+    def assertEqualStat(self, expected, actual):
+        """assert that expected and actual are the same stat result.
+
+        :param expected: A stat result.
+        :param actual: A stat result.
+        :raises AssertionError: If the expected and actual stat values differ
+            other than by atime.
+        """
+        self.assertEqual(expected.st_size, actual.st_size)
+        self.assertEqual(expected.st_mtime, actual.st_mtime)
+        self.assertEqual(expected.st_ctime, actual.st_ctime)
+        self.assertEqual(expected.st_dev, actual.st_dev)
+        self.assertEqual(expected.st_ino, actual.st_ino)
+        self.assertEqual(expected.st_mode, actual.st_mode)
+
     def assertPositive(self, val):
         """Assert that val is greater than 0."""
         self.assertTrue(val > 0, 'expected a positive value, but got %s' % val)

=== modified file 'bzrlib/tests/blackbox/test_info.py'
--- bzrlib/tests/blackbox/test_info.py	2008-08-29 05:18:56 +0000
+++ bzrlib/tests/blackbox/test_info.py	2008-09-25 22:25:09 +0000
@@ -1376,13 +1376,13 @@
     def test_info_stacked(self):
         # We have a mainline
         trunk_tree = self.make_branch_and_tree('mainline',
-            format='development1')
+            format='1.6')
         trunk_tree.commit('mainline')
         # and a branch from it which is stacked
         new_dir = trunk_tree.bzrdir.sprout('newbranch', stacked=True)
         out, err = self.run_bzr('info newbranch')
         self.assertEqual(
-"""Standalone tree (format: development1)
+"""Standalone tree (format: 1.6)
 Location:
   branch root: newbranch
 

=== modified file 'bzrlib/tests/blackbox/test_log.py'
--- bzrlib/tests/blackbox/test_log.py	2008-09-01 00:56:53 +0000
+++ bzrlib/tests/blackbox/test_log.py	2008-09-24 13:43:48 +0000
@@ -138,7 +138,25 @@
         self.assertTrue('revno: 2\n' not in log)
         self.assertTrue('branch nick: branch2\n' in log)
         self.assertTrue('branch nick: branch1\n' not in log)
-        
+
+    def test_log_change_revno(self):
+        self._prepare()
+        expected_log = self.run_bzr("log -r 1")[0]
+        log = self.run_bzr("log -c 1")[0]
+        self.assertEqualDiff(expected_log, log)
+
+    def test_log_change_single_revno(self):
+        self._prepare()
+        self.run_bzr_error('bzr: ERROR: Option --change does not'
+                           ' accept revision ranges',
+                           ['log', '--change', '2..3'])
+
+    def test_log_change_incompatible_with_revision(self):
+        self._prepare()
+        self.run_bzr_error('bzr: ERROR: --revision and --change'
+                           ' are mutually exclusive',
+                           ['log', '--change', '2', '--revision', '3'])
+
     def test_log_nonexistent_file(self):
         # files that don't exist in either the basis tree or working tree
         # should give an error

=== modified file 'bzrlib/tests/blackbox/test_push.py'
--- bzrlib/tests/blackbox/test_push.py	2008-08-23 17:57:41 +0000
+++ bzrlib/tests/blackbox/test_push.py	2008-09-25 22:25:09 +0000
@@ -343,9 +343,9 @@
         self.assertFalse(self.get_transport('published').has('.'))
 
     def test_push_notifies_default_stacking(self):
-        self.make_branch('stack_on', format='development1')
+        self.make_branch('stack_on', format='1.6')
         self.make_bzrdir('.').get_config().set_default_stack_on('stack_on')
-        self.make_branch('from', format='development1')
+        self.make_branch('from', format='1.6')
         out, err = self.run_bzr('push -d from to')
         self.assertContainsRe(err,
                               'Using default stacking branch stack_on at .*')

=== modified file 'bzrlib/tests/intertree_implementations/__init__.py'
--- bzrlib/tests/intertree_implementations/__init__.py	2008-07-31 04:51:46 +0000
+++ bzrlib/tests/intertree_implementations/__init__.py	2008-09-24 01:54:18 +0000
@@ -22,6 +22,7 @@
  - tests/test_workingtree.py
 """
 
+import bzrlib
 import bzrlib.errors as errors
 from bzrlib.transport import get_transport
 from bzrlib.transform import TransformPreview
@@ -72,11 +73,11 @@
              mutable_trees_to_test_trees)
         """
         result = []
-        for (intertree_class,
+        for (label, intertree_class,
             workingtree_format,
             workingtree_format_to,
             mutable_trees_to_test_trees) in formats:
-            scenario = (intertree_class.__name__, {
+            scenario = (label, {
                 "transport_server":self._transport_server,
                 "transport_readonly_server":self._transport_readonly_server,
                 "bzrdir_format":workingtree_format._matchingbzrdir,
@@ -109,18 +110,37 @@
         ]
     test_intertree_permutations = [
         # test InterTree with two default-format working trees.
-        (InterTree, default_tree_format, default_tree_format,
+        (InterTree.__name__, InterTree, default_tree_format, default_tree_format,
          return_provided_trees)]
     for optimiser in InterTree._optimisers:
-        test_intertree_permutations.append(
-            (optimiser,
-             optimiser._matching_from_tree_format,
-             optimiser._matching_to_tree_format,
-             optimiser._test_mutable_trees_to_test_trees))
-
-    # test PreviewTree
+        if optimiser is bzrlib.workingtree_4.InterDirStateTree:
+            # Its a little ugly to be conditional here, but less so than having
+            # the optimiser listed twice.
+            # Add once, compiled version
+            test_intertree_permutations.append(
+                (optimiser.__name__ + "(C)",
+                 optimiser,
+                 optimiser._matching_from_tree_format,
+                 optimiser._matching_to_tree_format,
+                 optimiser.make_source_parent_tree_compiled_dirstate))
+            # python version
+            test_intertree_permutations.append(
+                (optimiser.__name__ + "(PY)",
+                 optimiser,
+                 optimiser._matching_from_tree_format,
+                 optimiser._matching_to_tree_format,
+                 optimiser.make_source_parent_tree_python_dirstate))
+        else:
+            test_intertree_permutations.append(
+                (optimiser.__name__,
+                 optimiser,
+                 optimiser._matching_from_tree_format,
+                 optimiser._matching_to_tree_format,
+                 optimiser._test_mutable_trees_to_test_trees))
+    # PreviewTree does not have an InterTree optimiser class.
     test_intertree_permutations.append(
-        (InterTree,
+        (InterTree.__name__ + "(PreviewTree)",
+         InterTree,
          default_tree_format,
          default_tree_format,
          mutable_trees_to_preview_trees))

=== modified file 'bzrlib/tests/per_repository/test_commit_builder.py'
--- bzrlib/tests/per_repository/test_commit_builder.py	2008-09-04 20:32:04 +0000
+++ bzrlib/tests/per_repository/test_commit_builder.py	2008-09-22 05:15:20 +0000
@@ -151,7 +151,7 @@
         try:
             ie = inventory.make_entry('directory', '', None,
                     tree.get_root_id())
-            delta, version_recorded = builder.record_entry_contents(
+            delta, version_recorded, fs_hash = builder.record_entry_contents(
                 ie, [parent_tree.inventory], '', tree,
                 tree.path_content_summary(''))
             self.assertFalse(version_recorded)
@@ -164,6 +164,8 @@
                     delta)
             else:
                 self.assertEqual(None, delta)
+            # Directories do not get hashed.
+            self.assertEqual(None, fs_hash)
             builder.abort()
         except:
             builder.abort()
@@ -288,10 +290,12 @@
         os.symlink('target', 'link')
         self._add_commit_check_unchanged(tree, 'link')
 
-    def _add_commit_renamed_check_changed(self, tree, name):
+    def _add_commit_renamed_check_changed(self, tree, name,
+        expect_fs_hash=False):
         def rename():
             tree.rename_one(name, 'new_' + name)
-        self._add_commit_change_check_changed(tree, name, rename)
+        self._add_commit_change_check_changed(tree, name, rename,
+            expect_fs_hash=expect_fs_hash)
 
     def test_last_modified_revision_after_rename_dir_changes(self):
         # renaming a dir changes the last modified.
@@ -303,7 +307,8 @@
         # renaming a file changes the last modified.
         tree = self.make_branch_and_tree('.')
         self.build_tree(['file'])
-        self._add_commit_renamed_check_changed(tree, 'file')
+        self._add_commit_renamed_check_changed(tree, 'file',
+            expect_fs_hash=True)
 
     def test_last_modified_revision_after_rename_link_changes(self):
         # renaming a link changes the last modified.
@@ -312,12 +317,14 @@
         os.symlink('target', 'link')
         self._add_commit_renamed_check_changed(tree, 'link')
 
-    def _add_commit_reparent_check_changed(self, tree, name):
+    def _add_commit_reparent_check_changed(self, tree, name,
+        expect_fs_hash=False):
         self.build_tree(['newparent/'])
         tree.add(['newparent'])
         def reparent():
             tree.rename_one(name, 'newparent/new_' + name)
-        self._add_commit_change_check_changed(tree, name, reparent)
+        self._add_commit_change_check_changed(tree, name, reparent,
+            expect_fs_hash=expect_fs_hash)
 
     def test_last_modified_revision_after_reparent_dir_changes(self):
         # reparenting a dir changes the last modified.
@@ -329,7 +336,8 @@
         # reparenting a file changes the last modified.
         tree = self.make_branch_and_tree('.')
         self.build_tree(['file'])
-        self._add_commit_reparent_check_changed(tree, 'file')
+        self._add_commit_reparent_check_changed(tree, 'file',
+            expect_fs_hash=True)
 
     def test_last_modified_revision_after_reparent_link_changes(self):
         # reparenting a link changes the last modified.
@@ -338,11 +346,13 @@
         os.symlink('target', 'link')
         self._add_commit_reparent_check_changed(tree, 'link')
 
-    def _add_commit_change_check_changed(self, tree, name, changer):
+    def _add_commit_change_check_changed(self, tree, name, changer,
+        expect_fs_hash=False):
         tree.add([name], [name + 'id'])
         rev1 = tree.commit('')
         changer()
-        rev2 = self.mini_commit(tree, name, tree.id2path(name + 'id'))
+        rev2 = self.mini_commit(tree, name, tree.id2path(name + 'id'),
+            expect_fs_hash=expect_fs_hash)
         tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
         self.assertEqual(rev1, tree1.inventory[name + 'id'].revision)
         self.assertEqual(rev2, tree2.inventory[name + 'id'].revision)
@@ -353,7 +363,7 @@
         self.assertFileGraph(expected_graph, tree, (file_id, rev2))
 
     def mini_commit(self, tree, name, new_name, records_version=True,
-        delta_against_basis=True):
+        delta_against_basis=True, expect_fs_hash=False):
         """Perform a miniature commit looking for record entry results.
         
         :param tree: The tree to commit.
@@ -363,6 +373,8 @@
             record a new version.
         :param delta_against_basis: True of the commit of new_name is expected
             to have a delta against the basis.
+        :param expect_fs_hash: True or false to indicate whether we expect a
+            file hash to be returned from the record_entry_contents call.
         """
         tree.lock_write()
         try:
@@ -396,18 +408,24 @@
                 commit_id(parent_id)
             # because a change of some sort is meant to have occurred,
             # recording the entry must return True.
-            delta, version_recorded = commit_id(file_id)
+            delta, version_recorded, fs_hash = commit_id(file_id)
             if records_version:
                 self.assertTrue(version_recorded)
             else:
                 self.assertFalse(version_recorded)
+            if expect_fs_hash:
+                tree_file_stat = tree.get_file_with_stat(file_id)
+                tree_file_stat[0].close()
+                self.assertEqual(2, len(fs_hash))
+                self.assertEqual(tree.get_file_sha1(file_id), fs_hash[0])
+                self.assertEqualStat(tree_file_stat[1], fs_hash[1])
+            else:
+                self.assertEqual(None, fs_hash)
             new_entry = builder.new_inventory[file_id]
             if delta_against_basis:
                 expected_delta = (name, new_name, file_id, new_entry)
             else:
                 expected_delta = None
-            if expected_delta != delta:
-                import pdb;pdb.set_trace()
             self.assertEqual(expected_delta, delta)
             builder.finish_inventory()
             rev2 = builder.commit('')
@@ -434,7 +452,8 @@
         self.build_tree(['file'])
         def change_file():
             tree.put_file_bytes_non_atomic('fileid', 'new content')
-        self._add_commit_change_check_changed(tree, 'file', change_file)
+        self._add_commit_change_check_changed(tree, 'file', change_file,
+            expect_fs_hash=True)
 
     def test_last_modified_revision_after_content_link_changes(self):
         # changing a link changes the last modified.
@@ -455,13 +474,14 @@
         tree.rename_one(name, 'new_' + name)
         return tree.commit('')
 
-    def _commit_sprout_rename_merge(self, tree1, name):
+    def _commit_sprout_rename_merge(self, tree1, name, expect_fs_hash=False):
         rev1, tree2 = self._commit_sprout(tree1, name)
         # change both sides equally
         rev2 = self._rename_in_tree(tree1, name)
         rev3 = self._rename_in_tree(tree2, name)
         tree1.merge_from_branch(tree2.branch)
-        rev4 = self.mini_commit(tree1, 'new_' + name, 'new_' + name)
+        rev4 = self.mini_commit(tree1, 'new_' + name, 'new_' + name,
+            expect_fs_hash=expect_fs_hash)
         tree3, = self._get_revtrees(tree1, [rev4])
         self.assertEqual(rev4, tree3.inventory[name + 'id'].revision)
         file_id = name + 'id'
@@ -482,7 +502,7 @@
         # merge a file changes the last modified.
         tree1 = self.make_branch_and_tree('t1')
         self.build_tree(['t1/file'])
-        self._commit_sprout_rename_merge(tree1, 'file')
+        self._commit_sprout_rename_merge(tree1, 'file', expect_fs_hash=True)
 
     def test_last_modified_revision_after_merge_link_changes(self):
         # merge a link changes the last modified.
@@ -534,7 +554,7 @@
         self.requireFeature(tests.SymlinkFeature)
         os.symlink('target', name)
 
-    def _check_kind_change(self, make_before, make_after):
+    def _check_kind_change(self, make_before, make_after, expect_fs_hash=False):
         tree = self.make_branch_and_tree('.')
         path = 'name'
         make_before(path)
@@ -543,16 +563,19 @@
             osutils.delete_any(path)
             make_after(path)
 
-        self._add_commit_change_check_changed(tree, path, change_kind)
+        self._add_commit_change_check_changed(tree, path, change_kind,
+            expect_fs_hash=expect_fs_hash)
 
     def test_last_modified_dir_file(self):
-        self._check_kind_change(self.make_dir, self.make_file)
+        self._check_kind_change(self.make_dir, self.make_file,
+            expect_fs_hash=True)
 
     def test_last_modified_dir_link(self):
         self._check_kind_change(self.make_dir, self.make_link)
 
     def test_last_modified_link_file(self):
-        self._check_kind_change(self.make_link, self.make_file)
+        self._check_kind_change(self.make_link, self.make_file,
+            expect_fs_hash=True)
 
     def test_last_modified_link_dir(self):
         self._check_kind_change(self.make_link, self.make_dir)

=== modified file 'bzrlib/tests/test__dirstate_helpers.py'
--- bzrlib/tests/test__dirstate_helpers.py	2008-09-02 18:51:03 +0000
+++ bzrlib/tests/test__dirstate_helpers.py	2008-09-25 01:54:42 +0000
@@ -18,12 +18,16 @@
 
 import bisect
 import os
+import time
 
 from bzrlib import (
     dirstate,
     errors,
     tests,
     )
+from bzrlib.tests import (
+        SymlinkFeature,
+        )
 from bzrlib.tests import test_dirstate
 
 
@@ -765,3 +769,405 @@
             from bzrlib._dirstate_helpers_py import _read_dirblocks_py
             self.assertIs(_read_dirblocks_py, dirstate._read_dirblocks)
 
+    def test_update_entry(self):
+        if CompiledDirstateHelpersFeature.available():
+            from bzrlib._dirstate_helpers_c import update_entry
+            self.assertIs(update_entry, dirstate.update_entry)
+        else:
+            from bzrlib.dirstate import py_update_entry
+            self.assertIs(py_update_entry, dirstate.py_update_entry)
+
+    def test_process_entry(self):
+        if CompiledDirstateHelpersFeature.available():
+            from bzrlib._dirstate_helpers_c import ProcessEntryC
+            self.assertIs(ProcessEntryC, dirstate._process_entry)
+        else:
+            from bzrlib.dirstate import ProcessEntryPython
+            self.assertIs(ProcessEntryPython, dirstate._process_entry)
+
+
+class TestUpdateEntry(test_dirstate.TestCaseWithDirState):
+    """Test the DirState.update_entry functions"""
+
+    def get_state_with_a(self):
+        """Create a DirState tracking a single object named 'a'"""
+        state = test_dirstate.InstrumentedDirState.initialize('dirstate')
+        self.addCleanup(state.unlock)
+        state.add('a', 'a-id', 'file', None, '')
+        entry = state._get_entry(0, path_utf8='a')
+        self.set_update_entry()
+        return state, entry
+
+    def set_update_entry(self):
+        self.update_entry = dirstate.py_update_entry
+
+    def test_observed_sha1_cachable(self):
+        state, entry = self.get_state_with_a()
+        atime = time.time() - 10
+        self.build_tree(['a'])
+        statvalue = os.lstat('a')
+        statvalue = test_dirstate._FakeStat(statvalue.st_size, atime, atime,
+            statvalue.st_dev, statvalue.st_ino, statvalue.st_mode)
+        state._observed_sha1(entry, "foo", statvalue)
+        self.assertEqual('foo', entry[1][0][1])
+        packed_stat = dirstate.pack_stat(statvalue)
+        self.assertEqual(packed_stat, entry[1][0][4])
+
+    def test_observed_sha1_not_cachable(self):
+        state, entry = self.get_state_with_a()
+        oldval = entry[1][0][1]
+        oldstat = entry[1][0][4]
+        self.build_tree(['a'])
+        statvalue = os.lstat('a')
+        state._observed_sha1(entry, "foo", statvalue)
+        self.assertEqual(oldval, entry[1][0][1])
+        self.assertEqual(oldstat, entry[1][0][4])
+
+    def test_update_entry(self):
+        state, _ = self.get_state_with_a()
+        tree = self.make_branch_and_tree('tree')
+        tree.lock_write()
+        empty_revid = tree.commit('empty')
+        self.build_tree(['tree/a'])
+        tree.add(['a'], ['a-id'])
+        with_a_id = tree.commit('with_a')
+        self.addCleanup(tree.unlock)
+        state.set_parent_trees(
+            [(empty_revid, tree.branch.repository.revision_tree(empty_revid))],
+            [])
+        entry = state._get_entry(0, path_utf8='a')
+        self.build_tree(['a'])
+        # Add one where we don't provide the stat or sha already
+        self.assertEqual(('', 'a', 'a-id'), entry[0])
+        self.assertEqual(('f', '', 0, False, dirstate.DirState.NULLSTAT),
+                         entry[1][0])
+        # Flush the buffers to disk
+        state.save()
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+
+        stat_value = os.lstat('a')
+        packed_stat = dirstate.pack_stat(stat_value)
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual(None, link_or_sha1)
+
+        # The dirblock entry should not have cached the file's sha1 (too new)
+        self.assertEqual(('f', '', 14, False, dirstate.DirState.NULLSTAT),
+                         entry[1][0])
+        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
+                         state._dirblock_state)
+        mode = stat_value.st_mode
+        self.assertEqual([('is_exec', mode, False)], state._log)
+
+        state.save()
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+
+        # If we do it again right away, we don't know if the file has changed
+        # so we will re-read the file. Roll the clock back so the file is
+        # guaranteed to look too new.
+        state.adjust_time(-10)
+        del state._log[:]
+
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual([('is_exec', mode, False)], state._log)
+        self.assertEqual(None, link_or_sha1)
+        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
+                         state._dirblock_state)
+        self.assertEqual(('f', '', 14, False, dirstate.DirState.NULLSTAT),
+                         entry[1][0])
+        state.save()
+
+        # If it is cachable (the clock has moved forward) but new it still
+        # won't calculate the sha or cache it.
+        state.adjust_time(+20)
+        del state._log[:]
+        link_or_sha1 = dirstate.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual(None, link_or_sha1)
+        self.assertEqual([('is_exec', mode, False)], state._log)
+        self.assertEqual(('f', '', 14, False, dirstate.DirState.NULLSTAT),
+                         entry[1][0])
+
+        # If the file is no longer new, and the clock has been moved forward
+        # sufficiently, it will cache the sha.
+        del state._log[:]
+        state.set_parent_trees(
+            [(with_a_id, tree.branch.repository.revision_tree(with_a_id))],
+            [])
+        entry = state._get_entry(0, path_utf8='a')
+
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
+                         link_or_sha1)
+        self.assertEqual([('is_exec', mode, False), ('sha1', 'a')],
+                          state._log)
+        self.assertEqual(('f', link_or_sha1, 14, False, packed_stat),
+                         entry[1][0])
+
+        # Subsequent calls will just return the cached value
+        del state._log[:]
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
+                         link_or_sha1)
+        self.assertEqual([], state._log)
+        self.assertEqual(('f', link_or_sha1, 14, False, packed_stat),
+                         entry[1][0])
+
+    def test_update_entry_symlink(self):
+        """Update entry should read symlinks."""
+        self.requireFeature(SymlinkFeature)
+        state, entry = self.get_state_with_a()
+        state.save()
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+        os.symlink('target', 'a')
+
+        state.adjust_time(-10) # Make the symlink look new
+        stat_value = os.lstat('a')
+        packed_stat = dirstate.pack_stat(stat_value)
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual('target', link_or_sha1)
+        self.assertEqual([('read_link', 'a', '')], state._log)
+        # Dirblock is not updated (the link is too new)
+        self.assertEqual([('l', '', 6, False, dirstate.DirState.NULLSTAT)],
+                         entry[1])
+        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
+                         state._dirblock_state)
+
+        # Because the stat_value looks new, we should re-read the target
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual('target', link_or_sha1)
+        self.assertEqual([('read_link', 'a', ''),
+                          ('read_link', 'a', ''),
+                         ], state._log)
+        self.assertEqual([('l', '', 6, False, dirstate.DirState.NULLSTAT)],
+                         entry[1])
+        state.adjust_time(+20) # Skip into the future, all files look old
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual('target', link_or_sha1)
+        # We need to re-read the link because only now can we cache it
+        self.assertEqual([('read_link', 'a', ''),
+                          ('read_link', 'a', ''),
+                          ('read_link', 'a', ''),
+                         ], state._log)
+        self.assertEqual([('l', 'target', 6, False, packed_stat)],
+                         entry[1])
+
+        # Another call won't re-read the link
+        self.assertEqual([('read_link', 'a', ''),
+                          ('read_link', 'a', ''),
+                          ('read_link', 'a', ''),
+                         ], state._log)
+        link_or_sha1 = self.update_entry(state, entry, abspath='a',
+                                          stat_value=stat_value)
+        self.assertEqual('target', link_or_sha1)
+        self.assertEqual([('l', 'target', 6, False, packed_stat)],
+                         entry[1])
+
+    def do_update_entry(self, state, entry, abspath):
+        stat_value = os.lstat(abspath)
+        return self.update_entry(state, entry, abspath, stat_value)
+
+    def test_update_entry_dir(self):
+        state, entry = self.get_state_with_a()
+        self.build_tree(['a/'])
+        self.assertIs(None, self.do_update_entry(state, entry, 'a'))
+
+    def test_update_entry_dir_unchanged(self):
+        state, entry = self.get_state_with_a()
+        self.build_tree(['a/'])
+        state.adjust_time(+20)
+        self.assertIs(None, self.do_update_entry(state, entry, 'a'))
+        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
+                         state._dirblock_state)
+        state.save()
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+        self.assertIs(None, self.do_update_entry(state, entry, 'a'))
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+
+    def test_update_entry_file_unchanged(self):
+        state, _ = self.get_state_with_a()
+        tree = self.make_branch_and_tree('tree')
+        tree.lock_write()
+        self.build_tree(['tree/a'])
+        tree.add(['a'], ['a-id'])
+        with_a_id = tree.commit('witha')
+        self.addCleanup(tree.unlock)
+        state.set_parent_trees(
+            [(with_a_id, tree.branch.repository.revision_tree(with_a_id))],
+            [])
+        entry = state._get_entry(0, path_utf8='a')
+        self.build_tree(['a'])
+        sha1sum = 'b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6'
+        state.adjust_time(+20)
+        self.assertEqual(sha1sum, self.do_update_entry(state, entry, 'a'))
+        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
+                         state._dirblock_state)
+        state.save()
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+        self.assertEqual(sha1sum, self.do_update_entry(state, entry, 'a'))
+        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
+                         state._dirblock_state)
+
+    def create_and_test_file(self, state, entry):
+        """Create a file at 'a' and verify the state finds it during update.
+
+        The state should already be versioning *something* at 'a'. This makes
+        sure that state.update_entry recognizes it as a file.
+        """
+        self.build_tree(['a'])
+        stat_value = os.lstat('a')
+        packed_stat = dirstate.pack_stat(stat_value)
+
+        link_or_sha1 = self.do_update_entry(state, entry, abspath='a')
+        self.assertEqual(None, link_or_sha1)
+        self.assertEqual([('f', '', 14, False, dirstate.DirState.NULLSTAT)],
+                         entry[1])
+        return packed_stat
+
+    def create_and_test_dir(self, state, entry):
+        """Create a directory at 'a' and verify the state finds it.
+
+        The state should already be versioning *something* at 'a'. This makes
+        sure that state.update_entry recognizes it as a directory.
+        """
+        self.build_tree(['a/'])
+        stat_value = os.lstat('a')
+        packed_stat = dirstate.pack_stat(stat_value)
+
+        link_or_sha1 = self.do_update_entry(state, entry, abspath='a')
+        self.assertIs(None, link_or_sha1)
+        self.assertEqual([('d', '', 0, False, packed_stat)], entry[1])
+
+        return packed_stat
+
+    def create_and_test_symlink(self, state, entry):
+        """Create a symlink at 'a' and verify the state finds it.
+
+        The state should already be versioning *something* at 'a'. This makes
+        sure that state.update_entry recognizes it as a symlink.
+
+        This should not be called if this platform does not have symlink
+        support.
+        """
+        # caller should care about skipping test on platforms without symlinks
+        os.symlink('path/to/foo', 'a')
+
+        stat_value = os.lstat('a')
+        packed_stat = dirstate.pack_stat(stat_value)
+
+        link_or_sha1 = self.do_update_entry(state, entry, abspath='a')
+        self.assertEqual('path/to/foo', link_or_sha1)
+        self.assertEqual([('l', 'path/to/foo', 11, False, packed_stat)],
+                         entry[1])
+        return packed_stat
+
+    def test_update_file_to_dir(self):
+        """If a file changes to a directory we return None for the sha.
+        We also update the inventory record.
+        """
+        state, entry = self.get_state_with_a()
+        # The file sha1 won't be cached unless the file is old
+        state.adjust_time(+10)
+        self.create_and_test_file(state, entry)
+        os.remove('a')
+        self.create_and_test_dir(state, entry)
+
+    def test_update_file_to_symlink(self):
+        """File becomes a symlink"""
+        self.requireFeature(SymlinkFeature)
+        state, entry = self.get_state_with_a()
+        # The file sha1 won't be cached unless the file is old
+        state.adjust_time(+10)
+        self.create_and_test_file(state, entry)
+        os.remove('a')
+        self.create_and_test_symlink(state, entry)
+
+    def test_update_dir_to_file(self):
+        """Directory becoming a file updates the entry."""
+        state, entry = self.get_state_with_a()
+        # The file sha1 won't be cached unless the file is old
+        state.adjust_time(+10)
+        self.create_and_test_dir(state, entry)
+        os.rmdir('a')
+        self.create_and_test_file(state, entry)
+
+    def test_update_dir_to_symlink(self):
+        """Directory becomes a symlink"""
+        self.requireFeature(SymlinkFeature)
+        state, entry = self.get_state_with_a()
+        # The symlink target won't be cached if it isn't old
+        state.adjust_time(+10)
+        self.create_and_test_dir(state, entry)
+        os.rmdir('a')
+        self.create_and_test_symlink(state, entry)
+
+    def test_update_symlink_to_file(self):
+        """Symlink becomes a file"""
+        self.requireFeature(SymlinkFeature)
+        state, entry = self.get_state_with_a()
+        # The symlink and file info won't be cached unless old
+        state.adjust_time(+10)
+        self.create_and_test_symlink(state, entry)
+        os.remove('a')
+        self.create_and_test_file(state, entry)
+
+    def test_update_symlink_to_dir(self):
+        """Symlink becomes a directory"""
+        self.requireFeature(SymlinkFeature)
+        state, entry = self.get_state_with_a()
+        # The symlink target won't be cached if it isn't old
+        state.adjust_time(+10)
+        self.create_and_test_symlink(state, entry)
+        os.remove('a')
+        self.create_and_test_dir(state, entry)
+
+    def test__is_executable_win32(self):
+        state, entry = self.get_state_with_a()
+        self.build_tree(['a'])
+
+        # Make sure we are using the win32 implementation of _is_executable
+        state._is_executable = state._is_executable_win32
+
+        # The file on disk is not executable, but we are marking it as though
+        # it is. With _is_executable_win32 we ignore what is on disk.
+        entry[1][0] = ('f', '', 0, True, dirstate.DirState.NULLSTAT)
+
+        stat_value = os.lstat('a')
+        packed_stat = dirstate.pack_stat(stat_value)
+
+        state.adjust_time(-10) # Make sure everything is new
+        self.update_entry(state, entry, abspath='a', stat_value=stat_value)
+
+        # The row is updated, but the executable bit stays set.
+        self.assertEqual([('f', '', 14, True, dirstate.DirState.NULLSTAT)],
+                         entry[1])
+
+        # Make the disk object look old enough to cache (but it won't cache the sha
+        # as it is a new file).
+        state.adjust_time(+20)
+        digest = 'b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6'
+        self.update_entry(state, entry, abspath='a', stat_value=stat_value)
+        self.assertEqual([('f', '', 14, True, dirstate.DirState.NULLSTAT)],
+            entry[1])
+
+
+class TestCompiledUpdateEntry(TestUpdateEntry):
+    """Test the pyrex implementation of _read_dirblocks"""
+
+    _test_needs_features = [CompiledDirstateHelpersFeature]
+
+    def set_update_entry(self):
+        from bzrlib._dirstate_helpers_c import update_entry
+        self.update_entry = update_entry

=== modified file 'bzrlib/tests/test_branchbuilder.py'
--- bzrlib/tests/test_branchbuilder.py	2008-07-29 16:16:36 +0000
+++ bzrlib/tests/test_branchbuilder.py	2008-07-29 16:34:09 +0000
@@ -197,8 +197,23 @@
             builder.build_snapshot, 'B-id', None, [('weirdo', ('foo',))])
         self.assertEqual('Unknown build action: "weirdo"', str(e))
 
-    # TODO: rename a file/directory, but rename isn't supported by the
-    #       MemoryTree api yet, so for now we wait until it is used
+    def test_rename(self):
+        builder = self.build_a_rev()
+        builder.build_snapshot('B-id', None,
+            [('rename', ('a', 'b'))])
+        rev_tree = builder.get_branch().repository.revision_tree('B-id')
+        self.assertTreeShape([(u'', 'a-root-id', 'directory'),
+                              (u'b', 'a-id', 'file')], rev_tree)
+
+    def test_rename_into_subdir(self):
+        builder = self.build_a_rev()
+        builder.build_snapshot('B-id', None,
+            [('add', ('dir', 'dir-id', 'directory', None)),
+             ('rename', ('a', 'dir/a'))])
+        rev_tree = builder.get_branch().repository.revision_tree('B-id')
+        self.assertTreeShape([(u'', 'a-root-id', 'directory'),
+                              (u'dir', 'dir-id', 'directory'),
+                              (u'dir/a', 'a-id', 'file')], rev_tree)
 
     def test_set_parent(self):
         builder = self.build_a_rev()

=== modified file 'bzrlib/tests/test_bzrdir.py'
--- bzrlib/tests/test_bzrdir.py	2008-09-02 15:28:44 +0000
+++ bzrlib/tests/test_bzrdir.py	2008-09-25 22:25:09 +0000
@@ -442,7 +442,7 @@
         self.assertEqual(parent_bzrdir.root_transport.base,
                          repo_policy._stack_on_pwd)
 
-    def prepare_default_stacking(self, child_format='development1'):
+    def prepare_default_stacking(self, child_format='1.6'):
         parent_bzrdir = self.make_bzrdir('.')
         child_branch = self.make_branch('child', format=child_format)
         parent_bzrdir.get_config().set_default_stack_on(child_branch.base)
@@ -507,28 +507,28 @@
         self.assertTrue(repo.supports_rich_root())
 
     def test_add_fallback_repo_handles_absolute_urls(self):
-        stack_on = self.make_branch('stack_on', format='development1')
-        repo = self.make_repository('repo', format='development1')
+        stack_on = self.make_branch('stack_on', format='1.6')
+        repo = self.make_repository('repo', format='1.6')
         policy = bzrdir.UseExistingRepository(repo, stack_on.base)
         policy._add_fallback(repo)
 
     def test_add_fallback_repo_handles_relative_urls(self):
-        stack_on = self.make_branch('stack_on', format='development1')
-        repo = self.make_repository('repo', format='development1')
+        stack_on = self.make_branch('stack_on', format='1.6')
+        repo = self.make_repository('repo', format='1.6')
         policy = bzrdir.UseExistingRepository(repo, '.', stack_on.base)
         policy._add_fallback(repo)
 
     def test_configure_relative_branch_stacking_url(self):
-        stack_on = self.make_branch('stack_on', format='development1')
-        stacked = self.make_branch('stack_on/stacked', format='development1')
+        stack_on = self.make_branch('stack_on', format='1.6')
+        stacked = self.make_branch('stack_on/stacked', format='1.6')
         policy = bzrdir.UseExistingRepository(stacked.repository,
             '.', stack_on.base)
         policy.configure_branch(stacked)
         self.assertEqual('..', stacked.get_stacked_on_url())
 
     def test_relative_branch_stacking_to_absolute(self):
-        stack_on = self.make_branch('stack_on', format='development1')
-        stacked = self.make_branch('stack_on/stacked', format='development1')
+        stack_on = self.make_branch('stack_on', format='1.6')
+        stacked = self.make_branch('stack_on/stacked', format='1.6')
         policy = bzrdir.UseExistingRepository(stacked.repository,
             '.', self.get_readonly_url('stack_on'))
         policy.configure_branch(stacked)

=== modified file 'bzrlib/tests/test_dirstate.py'
--- bzrlib/tests/test_dirstate.py	2008-09-01 14:03:34 +0000
+++ bzrlib/tests/test_dirstate.py	2008-09-25 01:54:42 +0000
@@ -18,7 +18,6 @@
 
 import bisect
 import os
-import time
 
 from bzrlib import (
     dirstate,
@@ -564,20 +563,22 @@
         state.lock_read()
         try:
             entry = state._get_entry(0, path_utf8='a-file')
-            # The current sha1 sum should be empty
-            self.assertEqual('', entry[1][0][1])
+            # The current size should be 0 (default)
+            self.assertEqual(0, entry[1][0][2])
             # We should have a real entry.
             self.assertNotEqual((None, None), entry)
             # Make sure everything is old enough
             state._sha_cutoff_time()
             state._cutoff_time += 10
-            sha1sum = state.update_entry(entry, 'a-file', os.lstat('a-file'))
-            # We should have gotten a real sha1
-            self.assertEqual('ecc5374e9ed82ad3ea3b4d452ea995a5fd3e70e3',
-                             sha1sum)
+            # Change the file length
+            self.build_tree_contents([('a-file', 'shorter')])
+            sha1sum = dirstate.update_entry(state, entry, 'a-file',
+                os.lstat('a-file'))
+            # new file, no cached sha:
+            self.assertEqual(None, sha1sum)
 
             # The dirblock has been updated
-            self.assertEqual(sha1sum, entry[1][0][1])
+            self.assertEqual(7, entry[1][0][2])
             self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
                              state._dirblock_state)
 
@@ -593,7 +594,7 @@
         state.lock_read()
         try:
             entry = state._get_entry(0, path_utf8='a-file')
-            self.assertEqual(sha1sum, entry[1][0][1])
+            self.assertEqual(7, entry[1][0][2])
         finally:
             state.unlock()
 
@@ -612,10 +613,10 @@
         state.lock_read()
         try:
             entry = state._get_entry(0, path_utf8='a-file')
-            sha1sum = state.update_entry(entry, 'a-file', os.lstat('a-file'))
-            # We should have gotten a real sha1
-            self.assertEqual('ecc5374e9ed82ad3ea3b4d452ea995a5fd3e70e3',
-                             sha1sum)
+            sha1sum = dirstate.update_entry(state, entry, 'a-file',
+                os.lstat('a-file'))
+            # No sha - too new
+            self.assertEqual(None, sha1sum)
             self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
                              state._dirblock_state)
 
@@ -1666,326 +1667,6 @@
         self.st_mode = mode
 
 
-class TestUpdateEntry(TestCaseWithDirState):
-    """Test the DirState.update_entry functions"""
-
-    def get_state_with_a(self):
-        """Create a DirState tracking a single object named 'a'"""
-        state = InstrumentedDirState.initialize('dirstate')
-        self.addCleanup(state.unlock)
-        state.add('a', 'a-id', 'file', None, '')
-        entry = state._get_entry(0, path_utf8='a')
-        return state, entry
-
-    def test_update_entry(self):
-        state, entry = self.get_state_with_a()
-        self.build_tree(['a'])
-        # Add one where we don't provide the stat or sha already
-        self.assertEqual(('', 'a', 'a-id'), entry[0])
-        self.assertEqual([('f', '', 0, False, dirstate.DirState.NULLSTAT)],
-                         entry[1])
-        # Flush the buffers to disk
-        state.save()
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-
-        stat_value = os.lstat('a')
-        packed_stat = dirstate.pack_stat(stat_value)
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
-                         link_or_sha1)
-
-        # The dirblock entry should not cache the file's sha1
-        self.assertEqual([('f', '', 14, False, dirstate.DirState.NULLSTAT)],
-                         entry[1])
-        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
-                         state._dirblock_state)
-        mode = stat_value.st_mode
-        self.assertEqual([('sha1', 'a'), ('is_exec', mode, False)], state._log)
-
-        state.save()
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-
-        # If we do it again right away, we don't know if the file has changed
-        # so we will re-read the file. Roll the clock back so the file is
-        # guaranteed to look too new.
-        state.adjust_time(-10)
-
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual([('sha1', 'a'), ('is_exec', mode, False),
-                          ('sha1', 'a'), ('is_exec', mode, False),
-                         ], state._log)
-        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
-                         link_or_sha1)
-        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
-                         state._dirblock_state)
-        self.assertEqual([('f', '', 14, False, dirstate.DirState.NULLSTAT)],
-                         entry[1])
-        state.save()
-
-        # However, if we move the clock forward so the file is considered
-        # "stable", it should just cache the value.
-        state.adjust_time(+20)
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
-                         link_or_sha1)
-        self.assertEqual([('sha1', 'a'), ('is_exec', mode, False),
-                          ('sha1', 'a'), ('is_exec', mode, False),
-                          ('sha1', 'a'), ('is_exec', mode, False),
-                         ], state._log)
-        self.assertEqual([('f', link_or_sha1, 14, False, packed_stat)],
-                         entry[1])
-
-        # Subsequent calls will just return the cached value
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
-                         link_or_sha1)
-        self.assertEqual([('sha1', 'a'), ('is_exec', mode, False),
-                          ('sha1', 'a'), ('is_exec', mode, False),
-                          ('sha1', 'a'), ('is_exec', mode, False),
-                         ], state._log)
-        self.assertEqual([('f', link_or_sha1, 14, False, packed_stat)],
-                         entry[1])
-
-    def test_update_entry_symlink(self):
-        """Update entry should read symlinks."""
-        self.requireFeature(SymlinkFeature)
-        state, entry = self.get_state_with_a()
-        state.save()
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-        os.symlink('target', 'a')
-
-        state.adjust_time(-10) # Make the symlink look new
-        stat_value = os.lstat('a')
-        packed_stat = dirstate.pack_stat(stat_value)
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('target', link_or_sha1)
-        self.assertEqual([('read_link', 'a', '')], state._log)
-        # Dirblock is not updated (the link is too new)
-        self.assertEqual([('l', '', 6, False, dirstate.DirState.NULLSTAT)],
-                         entry[1])
-        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
-                         state._dirblock_state)
-
-        # Because the stat_value looks new, we should re-read the target
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('target', link_or_sha1)
-        self.assertEqual([('read_link', 'a', ''),
-                          ('read_link', 'a', ''),
-                         ], state._log)
-        self.assertEqual([('l', '', 6, False, dirstate.DirState.NULLSTAT)],
-                         entry[1])
-        state.adjust_time(+20) # Skip into the future, all files look old
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('target', link_or_sha1)
-        # We need to re-read the link because only now can we cache it
-        self.assertEqual([('read_link', 'a', ''),
-                          ('read_link', 'a', ''),
-                          ('read_link', 'a', ''),
-                         ], state._log)
-        self.assertEqual([('l', 'target', 6, False, packed_stat)],
-                         entry[1])
-
-        # Another call won't re-read the link
-        self.assertEqual([('read_link', 'a', ''),
-                          ('read_link', 'a', ''),
-                          ('read_link', 'a', ''),
-                         ], state._log)
-        link_or_sha1 = state.update_entry(entry, abspath='a',
-                                          stat_value=stat_value)
-        self.assertEqual('target', link_or_sha1)
-        self.assertEqual([('l', 'target', 6, False, packed_stat)],
-                         entry[1])
-
-    def do_update_entry(self, state, entry, abspath):
-        stat_value = os.lstat(abspath)
-        return state.update_entry(entry, abspath, stat_value)
-
-    def test_update_entry_dir(self):
-        state, entry = self.get_state_with_a()
-        self.build_tree(['a/'])
-        self.assertIs(None, self.do_update_entry(state, entry, 'a'))
-
-    def test_update_entry_dir_unchanged(self):
-        state, entry = self.get_state_with_a()
-        self.build_tree(['a/'])
-        state.adjust_time(+20)
-        self.assertIs(None, self.do_update_entry(state, entry, 'a'))
-        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
-                         state._dirblock_state)
-        state.save()
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-        self.assertIs(None, self.do_update_entry(state, entry, 'a'))
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-
-    def test_update_entry_file_unchanged(self):
-        state, entry = self.get_state_with_a()
-        self.build_tree(['a'])
-        sha1sum = 'b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6'
-        state.adjust_time(+20)
-        self.assertEqual(sha1sum, self.do_update_entry(state, entry, 'a'))
-        self.assertEqual(dirstate.DirState.IN_MEMORY_MODIFIED,
-                         state._dirblock_state)
-        state.save()
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-        self.assertEqual(sha1sum, self.do_update_entry(state, entry, 'a'))
-        self.assertEqual(dirstate.DirState.IN_MEMORY_UNMODIFIED,
-                         state._dirblock_state)
-
-    def create_and_test_file(self, state, entry):
-        """Create a file at 'a' and verify the state finds it.
-
-        The state should already be versioning *something* at 'a'. This makes
-        sure that state.update_entry recognizes it as a file.
-        """
-        self.build_tree(['a'])
-        stat_value = os.lstat('a')
-        packed_stat = dirstate.pack_stat(stat_value)
-
-        link_or_sha1 = self.do_update_entry(state, entry, abspath='a')
-        self.assertEqual('b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6',
-                         link_or_sha1)
-        self.assertEqual([('f', link_or_sha1, 14, False, packed_stat)],
-                         entry[1])
-        return packed_stat
-
-    def create_and_test_dir(self, state, entry):
-        """Create a directory at 'a' and verify the state finds it.
-
-        The state should already be versioning *something* at 'a'. This makes
-        sure that state.update_entry recognizes it as a directory.
-        """
-        self.build_tree(['a/'])
-        stat_value = os.lstat('a')
-        packed_stat = dirstate.pack_stat(stat_value)
-
-        link_or_sha1 = self.do_update_entry(state, entry, abspath='a')
-        self.assertIs(None, link_or_sha1)
-        self.assertEqual([('d', '', 0, False, packed_stat)], entry[1])
-
-        return packed_stat
-
-    def create_and_test_symlink(self, state, entry):
-        """Create a symlink at 'a' and verify the state finds it.
-
-        The state should already be versioning *something* at 'a'. This makes
-        sure that state.update_entry recognizes it as a symlink.
-
-        This should not be called if this platform does not have symlink
-        support.
-        """
-        # caller should care about skipping test on platforms without symlinks
-        os.symlink('path/to/foo', 'a')
-
-        stat_value = os.lstat('a')
-        packed_stat = dirstate.pack_stat(stat_value)
-
-        link_or_sha1 = self.do_update_entry(state, entry, abspath='a')
-        self.assertEqual('path/to/foo', link_or_sha1)
-        self.assertEqual([('l', 'path/to/foo', 11, False, packed_stat)],
-                         entry[1])
-        return packed_stat
-
-    def test_update_file_to_dir(self):
-        """If a file changes to a directory we return None for the sha.
-        We also update the inventory record.
-        """
-        state, entry = self.get_state_with_a()
-        # The file sha1 won't be cached unless the file is old
-        state.adjust_time(+10)
-        self.create_and_test_file(state, entry)
-        os.remove('a')
-        self.create_and_test_dir(state, entry)
-
-    def test_update_file_to_symlink(self):
-        """File becomes a symlink"""
-        self.requireFeature(SymlinkFeature)
-        state, entry = self.get_state_with_a()
-        # The file sha1 won't be cached unless the file is old
-        state.adjust_time(+10)
-        self.create_and_test_file(state, entry)
-        os.remove('a')
-        self.create_and_test_symlink(state, entry)
-
-    def test_update_dir_to_file(self):
-        """Directory becoming a file updates the entry."""
-        state, entry = self.get_state_with_a()
-        # The file sha1 won't be cached unless the file is old
-        state.adjust_time(+10)
-        self.create_and_test_dir(state, entry)
-        os.rmdir('a')
-        self.create_and_test_file(state, entry)
-
-    def test_update_dir_to_symlink(self):
-        """Directory becomes a symlink"""
-        self.requireFeature(SymlinkFeature)
-        state, entry = self.get_state_with_a()
-        # The symlink target won't be cached if it isn't old
-        state.adjust_time(+10)
-        self.create_and_test_dir(state, entry)
-        os.rmdir('a')
-        self.create_and_test_symlink(state, entry)
-
-    def test_update_symlink_to_file(self):
-        """Symlink becomes a file"""
-        self.requireFeature(SymlinkFeature)
-        state, entry = self.get_state_with_a()
-        # The symlink and file info won't be cached unless old
-        state.adjust_time(+10)
-        self.create_and_test_symlink(state, entry)
-        os.remove('a')
-        self.create_and_test_file(state, entry)
-
-    def test_update_symlink_to_dir(self):
-        """Symlink becomes a directory"""
-        self.requireFeature(SymlinkFeature)
-        state, entry = self.get_state_with_a()
-        # The symlink target won't be cached if it isn't old
-        state.adjust_time(+10)
-        self.create_and_test_symlink(state, entry)
-        os.remove('a')
-        self.create_and_test_dir(state, entry)
-
-    def test__is_executable_win32(self):
-        state, entry = self.get_state_with_a()
-        self.build_tree(['a'])
-
-        # Make sure we are using the win32 implementation of _is_executable
-        state._is_executable = state._is_executable_win32
-
-        # The file on disk is not executable, but we are marking it as though
-        # it is. With _is_executable_win32 we ignore what is on disk.
-        entry[1][0] = ('f', '', 0, True, dirstate.DirState.NULLSTAT)
-
-        stat_value = os.lstat('a')
-        packed_stat = dirstate.pack_stat(stat_value)
-
-        state.adjust_time(-10) # Make sure everything is new
-        state.update_entry(entry, abspath='a', stat_value=stat_value)
-
-        # The row is updated, but the executable bit stays set.
-        self.assertEqual([('f', '', 14, True, dirstate.DirState.NULLSTAT)],
-                         entry[1])
-
-        # Make the disk object look old enough to cache
-        state.adjust_time(+20)
-        digest = 'b50e5406bb5e153ebbeb20268fcf37c87e1ecfb6'
-        state.update_entry(entry, abspath='a', stat_value=stat_value)
-        self.assertEqual([('f', digest, 14, True, packed_stat)], entry[1])
-
-
 class TestPackStat(TestCaseWithTransport):
 
     def assertPackStat(self, expected, stat_value):

=== modified file 'bzrlib/tests/test_log.py'
--- bzrlib/tests/test_log.py	2008-08-21 04:12:02 +0000
+++ bzrlib/tests/test_log.py	2008-09-22 20:09:07 +0000
@@ -986,10 +986,10 @@
         view_revs_iter = log.get_view_revisions(mainline, revnos, tree.branch,
                                                 'reverse', True)
         actual_revs = log._filter_revisions_touching_file_id(
-                            tree.branch, 
+                            tree.branch,
                             file_id,
-                            mainline,
-                            list(view_revs_iter))
+                            list(view_revs_iter),
+                            'reverse')
         self.assertEqual(revisions, [r for r, revno, depth in actual_revs])
 
     def test_file_id_f1(self):

=== modified file 'bzrlib/tests/test_memorytree.py'
--- bzrlib/tests/test_memorytree.py	2008-07-22 18:12:25 +0000
+++ bzrlib/tests/test_memorytree.py	2008-07-23 02:35:09 +0000
@@ -182,3 +182,42 @@
         rev_id = tree.commit('first post')
         tree.unlock()
         self.assertEqual(rev_id, tree.last_revision())
+
+    def test_rename_file(self):
+        tree = self.make_branch_and_memory_tree('branch')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.add(['', 'foo'], ['root-id', 'foo-id'], ['directory', 'file'])
+        tree.put_file_bytes_non_atomic('foo-id', 'content\n')
+        tree.commit('one', rev_id='rev-one')
+        tree.rename_one('foo', 'bar')
+        self.assertEqual('bar', tree.id2path('foo-id'))
+        self.assertEqual('content\n', tree._file_transport.get_bytes('bar'))
+        self.assertRaises(errors.NoSuchFile,
+                          tree._file_transport.get_bytes, 'foo')
+        tree.commit('two', rev_id='rev-two')
+        self.assertEqual('content\n', tree._file_transport.get_bytes('bar'))
+        self.assertRaises(errors.NoSuchFile,
+                          tree._file_transport.get_bytes, 'foo')
+
+        rev_tree2 = tree.branch.repository.revision_tree('rev-two')
+        self.assertEqual('bar', rev_tree2.id2path('foo-id'))
+        self.assertEqual('content\n', rev_tree2.get_file_text('foo-id'))
+
+    def test_rename_file_to_subdir(self):
+        tree = self.make_branch_and_memory_tree('branch')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.add('')
+        tree.mkdir('subdir', 'subdir-id')
+        tree.add('foo', 'foo-id', 'file')
+        tree.put_file_bytes_non_atomic('foo-id', 'content\n')
+        tree.commit('one', rev_id='rev-one')
+
+        tree.rename_one('foo', 'subdir/bar')
+        self.assertEqual('subdir/bar', tree.id2path('foo-id'))
+        self.assertEqual('content\n',
+                         tree._file_transport.get_bytes('subdir/bar'))
+        tree.commit('two', rev_id='rev-two')
+        rev_tree2 = tree.branch.repository.revision_tree('rev-two')
+        self.assertEqual('subdir/bar', rev_tree2.id2path('foo-id'))

=== modified file 'bzrlib/tests/test_merge.py'
--- bzrlib/tests/test_merge.py	2008-08-27 20:32:22 +0000
+++ bzrlib/tests/test_merge.py	2008-09-05 03:11:40 +0000
@@ -21,9 +21,11 @@
     conflicts,
     errors,
     knit,
+    memorytree,
     merge as _mod_merge,
     option,
     progress,
+    tests,
     transform,
     versionedfile,
     )
@@ -1107,3 +1109,1638 @@
     def test_merge_move_and_change(self):
         self.expectFailure("lca merge doesn't conflict for move and change",
             super(TestLCAMerge, self).test_merge_move_and_change)
+
+
+class LoggingMerger(object):
+    # These seem to be the required attributes
+    requires_base = False
+    supports_reprocess = False
+    supports_show_base = False
+    supports_cherrypick = False
+    # We intentionally do not define supports_lca_trees
+
+    def __init__(self, *args, **kwargs):
+        self.args = args
+        self.kwargs = kwargs
+
+
+class TestMergerBase(TestCaseWithMemoryTransport):
+    """Common functionality for Merger tests that don't write to disk."""
+
+    def get_builder(self):
+        builder = self.make_branch_builder('path')
+        builder.start_series()
+        self.addCleanup(builder.finish_series)
+        return builder
+
+    def setup_simple_graph(self):
+        """Create a simple 3-node graph.
+
+        :return: A BranchBuilder
+        """
+        #
+        #  A
+        #  |\
+        #  B C
+        #
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', ('', None, 'directory', None))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        return builder
+
+    def setup_criss_cross_graph(self):
+        """Create a 5-node graph with a criss-cross.
+
+        :return: A BranchBuilder
+        """
+        # A
+        # |\
+        # B C
+        # |X|
+        # D E
+        builder = self.setup_simple_graph()
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        return builder
+
+    def make_Merger(self, builder, other_revision_id,
+                    interesting_files=None, interesting_ids=None):
+        """Make a Merger object from a branch builder"""
+        mem_tree = memorytree.MemoryTree.create_on_branch(builder.get_branch())
+        mem_tree.lock_write()
+        self.addCleanup(mem_tree.unlock)
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            mem_tree, other_revision_id)
+        merger.set_interesting_files(interesting_files)
+        # It seems there is no matching function for set_interesting_ids
+        merger.interesting_ids = interesting_ids
+        merger.merge_type = _mod_merge.Merge3Merger
+        return merger
+
+
+class TestMergerInMemory(TestMergerBase):
+
+    def test_find_base(self):
+        merger = self.make_Merger(self.setup_simple_graph(), 'C-id')
+        self.assertEqual('A-id', merger.base_rev_id)
+        self.assertFalse(merger._is_criss_cross)
+        self.assertIs(None, merger._lca_trees)
+
+    def test_find_base_criss_cross(self):
+        builder = self.setup_criss_cross_graph()
+        merger = self.make_Merger(builder, 'E-id')
+        self.assertEqual('A-id', merger.base_rev_id)
+        self.assertTrue(merger._is_criss_cross)
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merger._lca_trees])
+        # If we swap the order, we should get a different lca order
+        builder.build_snapshot('F-id', ['E-id'], [])
+        merger = self.make_Merger(builder, 'D-id')
+        self.assertEqual(['C-id', 'B-id'], [t.get_revision_id()
+                                            for t in merger._lca_trees])
+
+    def test_find_base_triple_criss_cross(self):
+        #       A-.
+        #      / \ \
+        #     B   C F # F is merged into both branches
+        #     |\ /| |
+        #     | X | |\
+        #     |/ \| | :
+        #   : D   E |
+        #    \|   |/
+        #     G   H
+        builder = self.setup_criss_cross_graph()
+        builder.build_snapshot('F-id', ['A-id'], [])
+        builder.build_snapshot('H-id', ['E-id', 'F-id'], [])
+        builder.build_snapshot('G-id', ['D-id', 'F-id'], [])
+        merger = self.make_Merger(builder, 'H-id')
+        self.assertEqual(['B-id', 'C-id', 'F-id'],
+                         [t.get_revision_id() for t in merger._lca_trees])
+
+    def test_no_criss_cross_passed_to_merge_type(self):
+        class LCATreesMerger(LoggingMerger):
+            supports_lca_trees = True
+
+        merger = self.make_Merger(self.setup_simple_graph(), 'C-id')
+        merger.merge_type = LCATreesMerger
+        merge_obj = merger.make_merger()
+        self.assertIsInstance(merge_obj, LCATreesMerger)
+        self.assertFalse('lca_trees' in merge_obj.kwargs)
+
+    def test_criss_cross_passed_to_merge_type(self):
+        merger = self.make_Merger(self.setup_criss_cross_graph(), 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merger._lca_trees])
+
+    def test_criss_cross_not_supported_merge_type(self):
+        merger = self.make_Merger(self.setup_criss_cross_graph(), 'E-id')
+        # We explicitly do not define supports_lca_trees
+        merger.merge_type = LoggingMerger
+        merge_obj = merger.make_merger()
+        self.assertIsInstance(merge_obj, LoggingMerger)
+        self.assertFalse('lca_trees' in merge_obj.kwargs)
+
+    def test_criss_cross_unsupported_merge_type(self):
+        class UnsupportedLCATreesMerger(LoggingMerger):
+            supports_lca_trees = False
+
+        merger = self.make_Merger(self.setup_criss_cross_graph(), 'E-id')
+        merger.merge_type = UnsupportedLCATreesMerger
+        merge_obj = merger.make_merger()
+        self.assertIsInstance(merge_obj, UnsupportedLCATreesMerger)
+        self.assertFalse('lca_trees' in merge_obj.kwargs)
+
+
+class TestMergerEntriesLCA(TestMergerBase):
+
+    def make_merge_obj(self, builder, other_revision_id,
+                       interesting_files=None, interesting_ids=None):
+        merger = self.make_Merger(builder, other_revision_id,
+            interesting_files=interesting_files,
+            interesting_ids=interesting_ids)
+        return merger.make_merger()
+
+    def test_simple(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('a-id', 'a\nb\nC\nc\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('a-id', 'a\nB\nb\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'a\nB\nb\nC\nc\nE\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('modify', ('a-id', 'a\nB\nb\nC\nc\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+        entries = list(merge_obj._entries_lca())
+
+        # (file_id, changed, parents, names, executable)
+        # BASE, lca1, lca2, OTHER, THIS
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'a']), u'a', u'a'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_not_in_base(self):
+        # LCAs all have the same last-modified revision for the file, as do
+        # the tips, but the base has something different
+        #       A    base, doesn't have the file
+        #       |\
+        #       B C  B introduces 'foo', C introduces 'bar'
+        #       |X|
+        #       D E  D and E now both have 'foo' and 'bar'
+        #       |X|
+        #       F G  the files are now in F, G, D and E, but not in A
+        #            G modifies 'bar'
+
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('add', (u'bar', 'bar-id', 'file', 'd\ne\nf\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('add', (u'bar', 'bar-id', 'file', 'd\ne\nf\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('G-id', ['E-id', 'D-id'],
+            [('modify', (u'bar-id', 'd\ne\nf\nG\n'))])
+        builder.build_snapshot('F-id', ['D-id', 'E-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        self.assertEqual(['D-id', 'E-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('bar-id', True,
+                           ((None, [root_id, root_id]), root_id, root_id),
+                           ((None, [u'bar', u'bar']), u'bar', u'bar'),
+                           ((None, [False, False]), False, False)),
+                         ], entries)
+
+    def test_not_in_this(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('a-id', 'a\nB\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('a-id', 'a\nb\nC\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'a\nB\nb\nC\nc\nE\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('unversion', 'a-id')])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), root_id, None),
+                           ((u'a', [u'a', u'a']), u'a', None),
+                           ((False, [False, False]), False, None)),
+                         ], entries)
+
+    def test_file_not_in_one_lca(self):
+        #   A   # just root
+        #   |\
+        #   B C # B no file, C introduces a file
+        #   |X|
+        #   D E # D and E both have the file, unchanged from C
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], []) # Inherited from C
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], # Merged from C
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([], entries)
+
+    def test_not_in_other(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('unversion', 'a-id')])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), None, root_id),
+                           ((u'a', [u'a', u'a']), None, u'a'),
+                           ((False, [False, False]), None, False)),
+                         ], entries)
+
+    def test_not_in_other_or_lca(self):
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B nothing, C deletes foo
+        #       |X|
+        #       D E  D restores foo (same as B), E leaves it deleted
+        # We should emit an entry for this
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('unversion', 'foo-id')])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, None]), None, root_id),
+                           ((u'foo', [u'foo', None]), None, 'foo'),
+                           ((False, [False, None]), None, False)),
+                         ], entries)
+
+    def test_only_in_one_lca(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('unversion', 'a-id')])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((None, [None, root_id]), None, None),
+                           ((None, [None, u'a']), None, None),
+                           ((None, [None, False]), None, None)),
+                         ], entries)
+
+    def test_only_in_other(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((None, [None, None]), root_id, None),
+                           ((None, [None, None]), u'a', None),
+                           ((None, [None, None]), False, None)),
+                         ], entries)
+
+    def test_one_lca_supersedes(self):
+        # One LCA supersedes the other LCAs last modified value, but the
+        # value is not the same as BASE.
+        #       A    base, introduces 'foo', last mod A
+        #       |\
+        #       B C  B modifies 'foo' (mod B), C does nothing (mod A)
+        #       |X|
+        #       D E  D does nothing (mod B), E updates 'foo' (mod E)
+        #       |X|
+        #       F G  F updates 'foo' (mod F). G does nothing (mod E)
+        #
+        #   At this point, G should not be considered to modify 'foo', even
+        #   though its LCAs disagree. This is because the modification in E
+        #   completely supersedes the value in D.
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('foo-id', 'E content\n'))])
+        builder.build_snapshot('G-id', ['E-id', 'D-id'], [])
+        builder.build_snapshot('F-id', ['D-id', 'E-id'],
+            [('modify', ('foo-id', 'F content\n'))])
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        self.assertEqual([], list(merge_obj._entries_lca()))
+
+    def test_one_lca_supersedes_path(self):
+        # Double-criss-cross merge, the ultimate base value is different from
+        # the intermediate.
+        #   A    value 'foo'
+        #   |\
+        #   B C  B value 'bar', C = 'foo'
+        #   |X|
+        #   D E  D = 'bar', E supersedes to 'bing'
+        #   |X|
+        #   F G  F = 'bing', G supersedes to 'barry'
+        #
+        # In this case, we technically should not care about the value 'bar' for
+        # D, because it was clearly superseded by E's 'bing'. The
+        # per-file/attribute graph would actually look like:
+        #   A
+        #   |
+        #   B
+        #   |
+        #   E
+        #   |
+        #   G
+        #
+        # Because the other side of the merge never modifies the value, it just
+        # takes the value from the merge.
+        #
+        # ATM this fails because we will prune 'foo' from the LCAs, but we
+        # won't prune 'bar'. This is getting far off into edge-case land, so we
+        # aren't supporting it yet.
+        #
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', ('foo', 'bing'))]) # override to bing
+        builder.build_snapshot('G-id', ['E-id', 'D-id'],
+            [('rename', ('bing', 'barry'))]) # override to barry
+        builder.build_snapshot('F-id', ['D-id', 'E-id'],
+            [('rename', ('bar', 'bing'))]) # Merge in E's change
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        self.expectFailure("We don't do an actual heads() check on lca values,"
+            " or use the per-attribute graph",
+            self.assertEqual, [], list(merge_obj._entries_lca()))
+
+    def test_one_lca_accidentally_pruned(self):
+        # Another incorrect resolution from the same basic flaw:
+        #   A    value 'foo'
+        #   |\
+        #   B C  B value 'bar', C = 'foo'
+        #   |X|
+        #   D E  D = 'bar', E reverts to 'foo'
+        #   |X|
+        #   F G  F = 'bing', G switches to 'bar'
+        #
+        # 'bar' will not be seen as an interesting change, because 'foo' will
+        # be pruned from the LCAs, even though it was newly introduced by E
+        # (superseding B).
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('G-id', ['E-id', 'D-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('F-id', ['D-id', 'E-id'],
+            [('rename', ('bar', 'bing'))]) # should end up conflicting
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.expectFailure("We prune values from BASE even when relevant.",
+            self.assertEqual,
+                [('foo-id', False,
+                  ((root_id, [root_id, root_id]), root_id, root_id),
+                  ((u'foo', [u'bar', u'foo']), u'bar', u'bing'),
+                  ((False, [False, False]), False, False)),
+                ], entries)
+
+    def test_both_sides_revert(self):
+        # Both sides of a criss-cross revert the text to the lca
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D reverts to B, E reverts to C
+        # This should conflict
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_different_lca_resolve_one_side_updates_content(self):
+        # Both sides converge, but then one side updates the text.
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D reverts to B, E reverts to C
+        #       |
+        #       F    F updates to a new value
+        # We need to emit an entry for 'foo', because D & E differed on the
+        # merge resolution
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('F-id', ['D-id'],
+            [('modify', ('foo-id', 'F content\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_same_lca_resolution_one_side_updates_content(self):
+        # Both sides converge, but then one side updates the text.
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D and E use C's value
+        #       |
+        #       F    F updates to a new value
+        # I think it is a bug that this conflicts, but we don't have a way to
+        # detect otherwise. And because of:
+        #   test_different_lca_resolve_one_side_updates_content
+        # We need to conflict.
+
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('modify', ('foo-id', 'C content\n'))]) # Same as E
+        builder.build_snapshot('F-id', ['D-id'],
+            [('modify', ('foo-id', 'F content\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.expectFailure("We don't detect that LCA resolution was the"
+                           " same on both sides",
+            self.assertEqual, [], entries)
+
+    def test_only_path_changed(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', (u'a', u'b'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        # The content was not changed, only the path
+        self.assertEqual([('a-id', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'a']), u'b', u'a'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_kind_changed(self):
+        # Identical content, except 'D' changes a-id into a directory
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('unversion', 'a-id'),
+             ('add', (u'a', 'a-id', 'directory', None))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        # Only the kind was changed (content)
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'a']), u'a', u'a'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_this_changed_kind(self):
+        # Identical content, but THIS changes a file to a directory
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('unversion', 'a-id'),
+             ('add', (u'a', 'a-id', 'directory', None))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        # Only the kind was changed (content)
+        self.assertEqual([], entries)
+
+    def test_interesting_files(self):
+        # Two files modified, but we should filter one of them
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['b'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'b', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_file_in_this(self):
+        # This renamed the file, but it should still match the entry in other
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('rename', ('b', 'c'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['c'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'b', u'b']), u'b', u'c'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_file_in_base(self):
+        # This renamed the file, but it should still match the entry in BASE
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'c', 'c-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('c', 'b'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('rename', ('c', 'b'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('c-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['c'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('c-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'c', [u'b', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_file_in_lca(self):
+        # This renamed the file, but it should still match the entry in LCA
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('b', 'c'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('rename', ('c', 'b'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['c'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'c', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_ids(self):
+        # Two files modified, but we should filter one of them
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_ids=['b-id'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'b', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+
+
+class TestMergerEntriesLCAOnDisk(tests.TestCaseWithTransport):
+
+    def get_builder(self):
+        builder = self.make_branch_builder('path')
+        builder.start_series()
+        self.addCleanup(builder.finish_series)
+        return builder
+
+    def get_wt_from_builder(self, builder):
+        """Get a real WorkingTree from the builder."""
+        the_branch = builder.get_branch()
+        wt = the_branch.bzrdir.create_workingtree()
+        # Note: This is a little bit ugly, but we are holding the branch
+        #       write-locked as part of the build process, and we would like to
+        #       maintain that. So we just force the WT to re-use the same
+        #       branch object.
+        wt._branch = the_branch
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        return wt
+
+    def do_merge(self, builder, other_revision_id):
+        wt = self.get_wt_from_builder(builder)
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, other_revision_id)
+        merger.merge_type = _mod_merge.Merge3Merger
+        return wt, merger.do_merge()
+
+    def test_simple_lca(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('modify', ('a-id', 'a\nb\nc\nd\ne\nf\n'))])
+        wt, conflicts = self.do_merge(builder, 'E-id')
+        self.assertEqual(0, conflicts)
+        # The merge should have simply update the contents of 'a'
+        self.assertEqual('a\nb\nc\nd\ne\nf\n', wt.get_file_text('a-id'))
+
+    def test_conflict_without_lca(self):
+        # This test would cause a merge conflict, unless we use the lca trees
+        # to determine the real ancestry
+        #   A       Path at 'foo'
+        #  / \
+        # B   C     Path renamed to 'bar' in B
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     Path at 'bar' in D and E
+        #     |
+        #     F     Path at 'baz' in F, which supersedes 'bar' and 'foo'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], # merge the rename
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('F-id', ['E-id'],
+            [('rename', ('bar', 'baz'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        # The merge should simply recognize that the final rename takes
+        # precedence
+        self.assertEqual('baz', wt.id2path('foo-id'))
+
+    def test_other_deletes_lca_renames(self):
+        # This test would cause a merge conflict, unless we use the lca trees
+        # to determine the real ancestry
+        #   A       Path at 'foo'
+        #  / \
+        # B   C     Path renamed to 'bar' in B
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     Path at 'bar' in D and E
+        #     |
+        #     F     F deletes 'bar'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], # merge the rename
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('F-id', ['E-id'],
+            [('unversion', 'foo-id')])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        self.assertRaises(errors.NoSuchId, wt.id2path, 'foo-id')
+
+    def test_executable_changes(self):
+        #   A       Path at 'foo'
+        #  / \
+        # B   C
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E
+        #     |
+        #     F     Executable bit changed
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        # Have to use a real WT, because BranchBuilder doesn't support exec bit
+        wt = self.get_wt_from_builder(builder)
+        tt = transform.TreeTransform(wt)
+        try:
+            tt.set_executability(True, tt.trans_id_tree_file_id('foo-id'))
+            tt.apply()
+        except:
+            tt.finalize()
+            raise
+        self.assertTrue(wt.is_executable('foo-id'))
+        wt.commit('F-id', rev_id='F-id')
+        # Reset to D, so that we can merge F
+        wt.set_parent_ids(['D-id'])
+        wt.branch.set_last_revision_info(3, 'D-id')
+        wt.revert()
+        self.assertFalse(wt.is_executable('foo-id'))
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.assertEqual(0, conflicts)
+        self.assertTrue(wt.is_executable('foo-id'))
+
+    def test_create_symlink(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A
+        #  / \
+        # B   C
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E
+        #     |
+        #     F     Add a symlink 'foo' => 'bar'
+        # Have to use a real WT, because BranchBuilder and MemoryTree don't
+        # have symlink support
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        # Have to use a real WT, because BranchBuilder doesn't support exec bit
+        wt = self.get_wt_from_builder(builder)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+        wt.commit('add symlink', rev_id='F-id')
+        # Reset to D, so that we can merge F
+        wt.set_parent_ids(['D-id'])
+        wt.branch.set_last_revision_info(3, 'D-id')
+        wt.revert()
+        self.assertIs(None, wt.path2id('foo'))
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('foo-id', wt.path2id('foo'))
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+
+    def test_both_sides_revert(self):
+        # Both sides of a criss-cross revert the text to the lca
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D reverts to B, E reverts to C
+        # This should conflict
+        # This must be done with a real WorkingTree, because normally their
+        # inventory contains "None" rather than a real sha1
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'E-id')
+        self.assertEqual(1, conflicts)
+        self.assertEqualDiff('<<<<<<< TREE\n'
+                             'B content\n'
+                             '=======\n'
+                             'C content\n'
+                             '>>>>>>> MERGE-SOURCE\n',
+                             wt.get_file_text('foo-id'))
+
+    def test_modified_symlink(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Create symlink foo => bar
+        #  / \
+        # B   C     B relinks foo => baz
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D & E have foo => baz
+        #     |
+        #     F     F changes it to bing
+        #
+        # Merging D & F should result in F cleanly overriding D, because D's
+        # value actually comes from B
+
+        # Have to use a real WT, because BranchBuilder and MemoryTree don't
+        # have symlink support
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink', rev_id='A-id')
+        os.remove('path/foo')
+        os.symlink('baz', 'path/foo')
+        wt.commit('foo => baz', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('baz', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        os.remove('path/foo')
+        os.symlink('bing', 'path/foo')
+        wt.commit('F foo => bing', rev_id='F-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.expectFailure("Merge3Merger doesn't use lcas for symlink content",
+            self.assertEqual, 0, conflicts)
+        self.assertEqual('bing', wt.get_symlink_target('foo-id'))
+
+    def test_renamed_symlink(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Create symlink foo => bar
+        #  / \
+        # B   C     B renames foo => barry
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D & E have barry
+        #     |
+        #     F     F renames barry to blah
+        #
+        # Merging D & F should result in F cleanly overriding D, because D's
+        # value actually comes from B
+
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('A add symlink', rev_id='A-id')
+        wt.rename_one('foo', 'barry')
+        wt.commit('B foo => barry', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('barry', wt.id2path('foo-id'))
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        wt.rename_one('barry', 'blah')
+        wt.commit('F barry => blah', rev_id='F-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        self.assertEqual('barry', wt.id2path('foo-id'))
+        # Check the output of the Merger object directly
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'F-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        root_id = wt.path2id('')
+        entries = list(merge_obj._entries_lca())
+        # No content change, just a path change
+        self.assertEqual([('foo-id', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'barry', u'foo']), u'blah', u'barry'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('blah', wt.id2path('foo-id'))
+
+    def test_symlink_no_content_change(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Create symlink foo => bar
+        #  / \
+        # B   C     B relinks foo => baz
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D & E have foo => baz
+        # |
+        # F         F has foo => bing
+        #
+        # Merging E into F should not cause a conflict, because E doesn't have
+        # a content change relative to the LCAs (it does relative to A)
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink', rev_id='A-id')
+        os.remove('path/foo')
+        os.symlink('baz', 'path/foo')
+        wt.commit('foo => baz', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('baz', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        os.remove('path/foo')
+        os.symlink('bing', 'path/foo')
+        wt.commit('F foo => bing', rev_id='F-id')
+
+        # Check the output of the Merger object directly
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        # Nothing interesting happened in OTHER relative to BASE
+        self.assertEqual([], list(merge_obj._entries_lca()))
+        # Now do a real merge, just to test the rest of the stack
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='E-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('bing', wt.get_symlink_target('foo-id'))
+
+    def test_symlink_this_changed_kind(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Nothing
+        #  / \
+        # B   C     B creates symlink foo => bar
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D changes foo into a file, E has foo => bing
+        #
+        # Mostly, this is trying to test that we don't try to os.readlink() on
+        # a file, or when there is nothing there
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        wt.commit('base', rev_id='A-id')
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink foo => bar', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+        os.remove('path/foo')
+        # We have to change the link in E, or it won't try to do a comparison
+        os.symlink('bing', 'path/foo')
+        wt.commit('E merges C & B, overrides to bing', rev_id='E-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        os.remove('path/foo')
+        self.build_tree_contents([('path/foo', 'file content\n')])
+        # XXX: workaround, WT doesn't detect kind changes unless you do
+        # iter_changes()
+        list(wt.iter_changes(wt.basis_tree()))
+        wt.commit('D merges B & C, makes it a file', rev_id='D-id')
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        root_id = wt.path2id('')
+        self.assertEqual([('foo-id', True,
+                           ((None, [root_id, None]), root_id, root_id),
+                           ((None, [u'foo', None]), u'foo', u'foo'),
+                           ((None, [False, None]), False, False)),
+                         ], entries)
+
+    def test_symlink_all_wt(self):
+        """Check behavior if all trees are Working Trees."""
+        self.requireFeature(tests.SymlinkFeature)
+        # The big issue is that entry.symlink_target is None for WorkingTrees.
+        # So we need to make sure we handle that case correctly.
+        #   A   foo => bar
+        #   |\
+        #   B C B relinks foo => baz
+        #   |X|
+        #   D E D & E have foo => baz
+        #     |
+        #     F F changes it to bing
+        # Merging D & F should result in F cleanly overriding D, because D's
+        # value actually comes from B
+
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink', rev_id='A-id')
+        os.remove('path/foo')
+        os.symlink('baz', 'path/foo')
+        wt.commit('foo => baz', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('baz', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        os.remove('path/foo')
+        os.symlink('bing', 'path/foo')
+        wt.commit('F foo => bing', rev_id='F-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        wt_base = wt.bzrdir.sprout('base', 'A-id').open_workingtree()
+        wt_base.lock_read()
+        self.addCleanup(wt_base.unlock)
+        wt_lca1 = wt.bzrdir.sprout('b-tree', 'B-id').open_workingtree()
+        wt_lca1.lock_read()
+        self.addCleanup(wt_lca1.unlock)
+        wt_lca2 = wt.bzrdir.sprout('c-tree', 'C-id').open_workingtree()
+        wt_lca2.lock_read()
+        self.addCleanup(wt_lca2.unlock)
+        wt_other = wt.bzrdir.sprout('other', 'F-id').open_workingtree()
+        wt_other.lock_read()
+        self.addCleanup(wt_other.unlock)
+        merge_obj = _mod_merge.Merge3Merger(wt, wt, wt_base,
+            wt_other, lca_trees=[wt_lca1, wt_lca2], do_merge=False)
+        entries = list(merge_obj._entries_lca())
+        root_id = wt.path2id('')
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_other_reverted_path_to_base(self):
+        #   A       Path at 'foo'
+        #  / \
+        # B   C     Path at 'bar' in B
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     Path at 'bar'
+        #     |
+        #     F     Path at 'foo'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', ('foo', 'bar'))]) # merge the rename
+        builder.build_snapshot('F-id', ['E-id'],
+            [('rename', ('bar', 'foo'))]) # Rename back to BASE
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('foo', wt.id2path('foo-id'))
+
+    def test_other_reverted_content_to_base(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'base content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('foo-id', 'B content\n'))]) # merge the content
+        builder.build_snapshot('F-id', ['E-id'],
+            [('modify', ('foo-id', 'base content\n'))]) # Revert back to BASE
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        # TODO: We need to use the per-file graph to properly select a BASE
+        #       before this will work. Or at least use the LCA trees to find
+        #       the appropriate content base. (which is B, not A).
+        self.expectFailure("Merge3Merger doesn't recognize reverted content",
+            self.assertEqual, 'base content\n', wt.get_file_text('foo-id'))
+
+    def test_other_modified_content(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'base content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('foo-id', 'B content\n'))]) # merge the content
+        builder.build_snapshot('F-id', ['E-id'],
+            [('modify', ('foo-id', 'F content\n'))]) # Override B content
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.expectFailure("Merge3Merger only uses BASE for content",
+            self.assertEqual, 'F content\n', wt.get_file_text('foo-id'))
+        self.assertEqual(0, conflicts)
+        self.assertEqual('F content\n', wt.get_file_text('foo-id'))
+
+    def test_all_wt(self):
+        """Check behavior if all trees are Working Trees."""
+        # The big issue is that entry.revision is None for WorkingTrees. (as is
+        # entry.text_sha1, etc. So we need to make sure we handle that case
+        # correctly.
+        #   A   Content of 'foo', path of 'a'
+        #   |\
+        #   B C B modifies content, C renames 'a' => 'b'
+        #   |X|
+        #   D E E updates content, renames 'b' => 'c'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'base content\n')),
+             ('add', (u'foo', 'foo-id', 'file', 'base content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('rename', ('a', 'b'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', ('b', 'c')),
+             ('modify', ('foo-id', 'E content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('rename', ('a', 'b'))]) # merged change
+        wt_this = self.get_wt_from_builder(builder)
+        wt_base = wt_this.bzrdir.sprout('base', 'A-id').open_workingtree()
+        wt_base.lock_read()
+        self.addCleanup(wt_base.unlock)
+        wt_lca1 = wt_this.bzrdir.sprout('b-tree', 'B-id').open_workingtree()
+        wt_lca1.lock_read()
+        self.addCleanup(wt_lca1.unlock)
+        wt_lca2 = wt_this.bzrdir.sprout('c-tree', 'C-id').open_workingtree()
+        wt_lca2.lock_read()
+        self.addCleanup(wt_lca2.unlock)
+        wt_other = wt_this.bzrdir.sprout('other', 'E-id').open_workingtree()
+        wt_other.lock_read()
+        self.addCleanup(wt_other.unlock)
+        merge_obj = _mod_merge.Merge3Merger(wt_this, wt_this, wt_base,
+            wt_other, lca_trees=[wt_lca1, wt_lca2], do_merge=False)
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'b']), u'c', u'b'),
+                           ((False, [False, False]), False, False)),
+                          ('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_nested_tree_unmodified(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub-tree',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub-tree/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        self.assertEqual([], entries)
+
+    def test_nested_tree_subtree_modified(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        self.build_tree_contents([('tree/sub/file', 'text2')])
+        sub_tree.commit('modify contents', rev_id='sub-B-id')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        # Nothing interesting about this sub-tree, because content changes are
+        # computed at a higher level
+        self.assertEqual([], entries)
+
+    def test_nested_tree_subtree_renamed(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        wt.rename_one('sub', 'alt_sub')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_last_revision('B-id')
+        wt.revert()
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('sub-tree-root', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'sub', [u'sub', u'sub']), u'alt_sub', u'sub'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_nested_tree_subtree_renamed_and_modified(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        self.build_tree_contents([('tree/sub/file', 'text2')])
+        sub_tree.commit('modify contents', rev_id='sub-B-id')
+        wt.rename_one('sub', 'alt_sub')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_last_revision('B-id')
+        wt.revert()
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('sub-tree-root', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'sub', [u'sub', u'sub']), u'alt_sub', u'sub'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+
+class TestLCAMultiWay(tests.TestCase):
+
+    def assertLCAMultiWay(self, expected, base, lcas, other, this,
+                          allow_overriding_lca=True):
+        self.assertEqual(expected, _mod_merge.Merge3Merger._lca_multi_way(
+                                (base, lcas), other, this,
+                                allow_overriding_lca=allow_overriding_lca))
+
+    def test_other_equal_equal_lcas(self):
+        """Test when OTHER=LCA and all LCAs are identical."""
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'bval'], 'bval', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval'], 'lcaval', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval', 'lcaval'], 'lcaval', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval', 'lcaval'], 'lcaval', 'tval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval', 'lcaval'], 'lcaval', None)
+
+    def test_other_equal_this(self):
+        """Test when other and this are identical."""
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'bval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            'bval', ['cval', 'dval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            'bval', [None, 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, [None, 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, ['lcaval', 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, ['cval', 'dval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, ['cval', 'dval'], None, None)
+        self.assertLCAMultiWay('this',
+            None, ['cval', 'dval', 'eval', 'fval'], 'oval', 'oval')
+
+    def test_no_lcas(self):
+        self.assertLCAMultiWay('this',
+            'bval', [], 'bval', 'tval')
+        self.assertLCAMultiWay('other',
+            'bval', [], 'oval', 'bval')
+        self.assertLCAMultiWay('conflict',
+            'bval', [], 'oval', 'tval')
+        self.assertLCAMultiWay('this',
+            'bval', [], 'oval', 'oval')
+
+    def test_lca_supersedes_other_lca(self):
+        """If one lca == base, the other lca takes precedence"""
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'lcaval'], 'lcaval', 'tval')
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'lcaval'], 'lcaval', 'bval')
+        # This is actually considered a 'revert' because the 'lcaval' in LCAS
+        # supersedes the BASE val (in the other LCA) but then OTHER reverts it
+        # back to bval.
+        self.assertLCAMultiWay('other',
+            'bval', ['bval', 'lcaval'], 'bval', 'lcaval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['bval', 'lcaval'], 'bval', 'tval')
+
+    def test_other_and_this_pick_different_lca(self):
+        # OTHER and THIS resolve the lca conflict in different ways
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'lca1val', 'lca2val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'lca2val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'bval'], 'lca1val', 'lca2val')
+
+    def test_other_in_lca(self):
+        # OTHER takes a value of one of the LCAs, THIS takes a new value, which
+        # theoretically supersedes both LCA values and 'wins'
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val'], 'lca1val', 'newval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'newval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'lca1val', 'newval',
+            allow_overriding_lca=False)
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'newval',
+            allow_overriding_lca=False)
+        # THIS reverted back to BASE, but that is an explicit supersede of all
+        # LCAs
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val', 'bval'], 'lca1val', 'bval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'bval',
+            allow_overriding_lca=False)
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'bval'], 'lca1val', 'bval',
+            allow_overriding_lca=False)
+
+    def test_this_in_lca(self):
+        # THIS takes a value of one of the LCAs, OTHER takes a new value, which
+        # theoretically supersedes both LCA values and 'wins'
+        self.assertLCAMultiWay('other',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca1val')
+        self.assertLCAMultiWay('other',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca2val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca1val',
+            allow_overriding_lca=False)
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca2val',
+            allow_overriding_lca=False)
+        # OTHER reverted back to BASE, but that is an explicit supersede of all
+        # LCAs
+        self.assertLCAMultiWay('other',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'bval', 'lca3val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'bval', 'lca3val',
+            allow_overriding_lca=False)
+
+    def test_all_differ(self):
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'tval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca2val'], 'oval', 'tval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'oval', 'tval')

=== modified file 'bzrlib/tests/test_pack_repository.py'
--- bzrlib/tests/test_pack_repository.py	2008-09-02 03:17:08 +0000
+++ bzrlib/tests/test_pack_repository.py	2008-09-25 22:25:09 +0000
@@ -22,7 +22,8 @@
 from cStringIO import StringIO
 from stat import S_ISDIR
 
-from bzrlib.index import GraphIndex, InMemoryGraphIndex
+from bzrlib.btree_index import BTreeGraphIndex
+from bzrlib.index import GraphIndex
 from bzrlib import (
     bzrdir,
     errors,
@@ -111,7 +112,7 @@
         self.assertFalse(t.has('knits'))
         # revision-indexes file-container directory
         self.assertEqual([],
-            list(GraphIndex(t, 'pack-names', None).iter_all_entries()))
+            list(self.index_class(t, 'pack-names', None).iter_all_entries()))
         self.assertTrue(S_ISDIR(t.stat('packs').st_mode))
         self.assertTrue(S_ISDIR(t.stat('upload').st_mode))
         self.assertTrue(S_ISDIR(t.stat('indices').st_mode))
@@ -152,9 +153,9 @@
         tree = self.make_branch_and_tree('.', format=format)
         trans = tree.branch.repository.bzrdir.get_repository_transport(None)
         self.assertEqual([],
-            list(GraphIndex(trans, 'pack-names', None).iter_all_entries()))
+            list(self.index_class(trans, 'pack-names', None).iter_all_entries()))
         tree.commit('foobarbaz')
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         index_nodes = list(index.iter_all_entries())
         self.assertEqual(1, len(index_nodes))
         node = index_nodes[0]
@@ -173,7 +174,7 @@
         tree1.branch.repository.fetch(tree2.branch.repository)
         trans = tree1.branch.repository.bzrdir.get_repository_transport(None)
         self.assertEqual([],
-            list(GraphIndex(trans, 'pack-names', None).iter_all_entries()))
+            list(self.index_class(trans, 'pack-names', None).iter_all_entries()))
 
     def test_commit_across_pack_shape_boundary_autopacks(self):
         format = self.get_format()
@@ -187,14 +188,14 @@
         for x in range(9):
             tree.commit('commit %s' % x)
         # there should be 9 packs:
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(9, len(list(index.iter_all_entries())))
         # insert some files in obsolete_packs which should be removed by pack.
         trans.put_bytes('obsolete_packs/foo', '123')
         trans.put_bytes('obsolete_packs/bar', '321')
         # committing one more should coalesce to 1 of 10.
         tree.commit('commit triggering pack')
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(1, len(list(index.iter_all_entries())))
         # packing should not damage data
         tree = tree.bzrdir.open_workingtree()
@@ -210,7 +211,7 @@
         large_pack_name = list(index.iter_all_entries())[0][1][0]
         # finally, committing again should not touch the large pack.
         tree.commit('commit not triggering pack')
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(2, len(list(index.iter_all_entries())))
         pack_names = [node[1][0] for node in index.iter_all_entries()]
         self.assertTrue(large_pack_name in pack_names)
@@ -239,7 +240,7 @@
         tree.commit('more work')
         tree.branch.repository.pack()
         # there should be 1 pack:
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(1, len(list(index.iter_all_entries())))
         self.assertEqual(2, len(tree.branch.repository.all_revision_ids()))
 
@@ -581,11 +582,11 @@
         for x in range(9):
             tree.commit('commit %s' % x)
         # there should be 9 packs:
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(9, len(list(index.iter_all_entries())))
         # committing one more should coalesce to 1 of 10.
         tree.commit('commit triggering pack')
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(1, len(list(index.iter_all_entries())))
         # packing should not damage data
         tree = tree.bzrdir.open_workingtree()
@@ -601,7 +602,7 @@
         large_pack_name = list(index.iter_all_entries())[0][1][0]
         # finally, committing again should not touch the large pack.
         tree.commit('commit not triggering pack')
-        index = GraphIndex(trans, 'pack-names', None)
+        index = self.index_class(trans, 'pack-names', None)
         self.assertEqual(2, len(list(index.iter_all_entries())))
         pack_names = [node[1][0] for node in index.iter_all_entries()]
         self.assertTrue(large_pack_name in pack_names)
@@ -613,26 +614,32 @@
     scenarios_params = [
          dict(format_name='pack-0.92',
               format_string="Bazaar pack repository format 1 (needs bzr 0.92)\n",
-              format_supports_external_lookups=False),
+              format_supports_external_lookups=False,
+              index_class=GraphIndex),
          dict(format_name='pack-0.92-subtree',
               format_string="Bazaar pack repository format 1 "
               "with subtree support (needs bzr 0.92)\n",
-              format_supports_external_lookups=False),
+              format_supports_external_lookups=False,
+              index_class=GraphIndex),
          dict(format_name='1.6',
               format_string="Bazaar RepositoryFormatKnitPack5 (bzr 1.6)\n",
-              format_supports_external_lookups=True),
+              format_supports_external_lookups=True,
+              index_class=GraphIndex),
          dict(format_name='1.6.1-rich-root',
               format_string="Bazaar RepositoryFormatKnitPack5RichRoot "
                   "(bzr 1.6.1)\n",
-              format_supports_external_lookups=True),
-         dict(format_name='development',
-              format_string="Bazaar development format 1 "
-                  "(needs bzr.dev from before 1.6)\n",
-              format_supports_external_lookups=True),
-         dict(format_name='development-subtree',
-              format_string="Bazaar development format 1 "
-                  "with subtree support (needs bzr.dev from before 1.6)\n",
-              format_supports_external_lookups=True),
+              format_supports_external_lookups=True,
+              index_class=GraphIndex),
+         dict(format_name='development2',
+              format_string="Bazaar development format 2 "
+                  "(needs bzr.dev from before 1.8)\n",
+              format_supports_external_lookups=True,
+              index_class=BTreeGraphIndex),
+         dict(format_name='development2-subtree',
+              format_string="Bazaar development format 2 "
+                  "with subtree support (needs bzr.dev from before 1.8)\n",
+              format_supports_external_lookups=True,
+              index_class=BTreeGraphIndex),
          ]
     adapter = tests.TestScenarioApplier()
     # name of the scenario is the format name

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2008-09-26 07:09:50 +0000
+++ bzrlib/tests/test_repository.py	2008-09-29 07:03:55 +0000
@@ -32,6 +32,7 @@
                            UnsupportedFormatError,
                            )
 from bzrlib import graph
+from bzrlib.btree_index import BTreeBuilder, BTreeGraphIndex
 from bzrlib.index import GraphIndex, InMemoryGraphIndex
 from bzrlib.repository import RepositoryFormat
 from bzrlib.smart import server
@@ -744,13 +745,16 @@
     def get_format(self):
         return bzrdir.format_registry.make_bzrdir('pack-0.92')
 
+    def get_packs(self):
+        format = self.get_format()
+        repo = self.make_repository('.', format=format)
+        return repo._pack_collection
+
     def test__max_pack_count(self):
         """The maximum pack count is a function of the number of revisions."""
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
         # no revisions - one pack, so that we can have a revision free repo
         # without it blowing up
+        packs = self.get_packs()
         self.assertEqual(1, packs._max_pack_count(0))
         # after that the sum of the digits, - check the first 1-9
         self.assertEqual(1, packs._max_pack_count(1))
@@ -772,21 +776,16 @@
         self.assertEqual(25, packs._max_pack_count(112894))
 
     def test_pack_distribution_zero(self):
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
+        packs = self.get_packs()
         self.assertEqual([0], packs.pack_distribution(0))
 
     def test_ensure_loaded_unlocked(self):
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
+        packs = self.get_packs()
         self.assertRaises(errors.ObjectNotLocked,
-                          repo._pack_collection.ensure_loaded)
+                          packs.ensure_loaded)
 
     def test_pack_distribution_one_to_nine(self):
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
+        packs = self.get_packs()
         self.assertEqual([1],
             packs.pack_distribution(1))
         self.assertEqual([1, 1],
@@ -808,9 +807,7 @@
 
     def test_pack_distribution_stable_at_boundaries(self):
         """When there are multi-rev packs the counts are stable."""
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
+        packs = self.get_packs()
         # in 10s:
         self.assertEqual([10], packs.pack_distribution(10))
         self.assertEqual([10, 1], packs.pack_distribution(11))
@@ -825,9 +822,7 @@
         self.assertEqual([100, 100, 10, 1], packs.pack_distribution(211))
 
     def test_plan_pack_operations_2009_revisions_skip_all_packs(self):
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
+        packs = self.get_packs()
         existing_packs = [(2000, "big"), (9, "medium")]
         # rev count - 2009 -> 2x1000 + 9x1
         pack_operations = packs.plan_autopack_combinations(
@@ -835,9 +830,7 @@
         self.assertEqual([], pack_operations)
 
     def test_plan_pack_operations_2010_revisions_skip_all_packs(self):
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
+        packs = self.get_packs()
         existing_packs = [(2000, "big"), (9, "medium"), (1, "single")]
         # rev count - 2010 -> 2x1000 + 1x10
         pack_operations = packs.plan_autopack_combinations(
@@ -845,15 +838,28 @@
         self.assertEqual([], pack_operations)
 
     def test_plan_pack_operations_2010_combines_smallest_two(self):
-        format = self.get_format()
-        repo = self.make_repository('.', format=format)
-        packs = repo._pack_collection
+        packs = self.get_packs()
         existing_packs = [(1999, "big"), (9, "medium"), (1, "single2"),
             (1, "single1")]
         # rev count - 2010 -> 2x1000 + 1x10 (3)
         pack_operations = packs.plan_autopack_combinations(
             existing_packs, [1000, 1000, 10])
-        self.assertEqual([[2, ["single2", "single1"]], [0, []]], pack_operations)
+        self.assertEqual([[2, ["single2", "single1"]]], pack_operations)
+
+    def test_plan_pack_operations_creates_a_single_op(self):
+        packs = self.get_packs()
+        existing_packs = [(50, 'a'), (40, 'b'), (30, 'c'), (10, 'd'),
+                          (10, 'e'), (6, 'f'), (4, 'g')]
+        # rev count 150 -> 1x100 and 5x10
+        # The two size 10 packs do not need to be touched. The 50, 40, 30 would
+        # be combined into a single 120 size pack, and the 6 & 4 would
+        # becombined into a size 10 pack. However, if we have to rewrite them,
+        # we save a pack file with no increased I/O by putting them into the
+        # same file.
+        distribution = packs.pack_distribution(150)
+        pack_operations = packs.plan_autopack_combinations(existing_packs,
+                                                           distribution)
+        self.assertEqual([[130, ['a', 'b', 'c', 'f', 'g']]], pack_operations)
 
     def test_all_packs_none(self):
         format = self.get_format()
@@ -971,9 +977,10 @@
         index_transport = self.get_transport('index')
         upload_transport.mkdir('.')
         pack = pack_repo.NewPack(upload_transport, index_transport,
-            pack_transport)
-        self.assertIsInstance(pack.revision_index, InMemoryGraphIndex)
-        self.assertIsInstance(pack.inventory_index, InMemoryGraphIndex)
+            pack_transport, index_builder_class=BTreeBuilder,
+            index_class=BTreeGraphIndex)
+        self.assertIsInstance(pack.revision_index, BTreeBuilder)
+        self.assertIsInstance(pack.inventory_index, BTreeBuilder)
         self.assertIsInstance(pack._hash, type(osutils.md5()))
         self.assertTrue(pack.upload_transport is upload_transport)
         self.assertTrue(pack.index_transport is index_transport)

=== modified file 'bzrlib/tests/test_selftest.py'
--- bzrlib/tests/test_selftest.py	2008-09-04 21:05:18 +0000
+++ bzrlib/tests/test_selftest.py	2008-09-26 05:47:03 +0000
@@ -472,23 +472,23 @@
         server2 = "b"
         format1 = WorkingTreeFormat2()
         format2 = WorkingTreeFormat3()
-        formats = [(str, format1, format2, "converter1"),
-            (int, format2, format1, "converter2")]
+        formats = [("1", str, format1, format2, "converter1"),
+            ("2", int, format2, format1, "converter2")]
         adapter = InterTreeTestProviderAdapter(server1, server2, formats)
         suite = adapter.adapt(input_test)
         tests = list(iter(suite))
         self.assertEqual(2, len(tests))
-        self.assertEqual(tests[0].intertree_class, formats[0][0])
-        self.assertEqual(tests[0].workingtree_format, formats[0][1])
-        self.assertEqual(tests[0].workingtree_format_to, formats[0][2])
-        self.assertEqual(tests[0].mutable_trees_to_test_trees, formats[0][3])
+        self.assertEqual(tests[0].intertree_class, formats[0][1])
+        self.assertEqual(tests[0].workingtree_format, formats[0][2])
+        self.assertEqual(tests[0].workingtree_format_to, formats[0][3])
+        self.assertEqual(tests[0].mutable_trees_to_test_trees, formats[0][4])
         self.assertEqual(tests[0]._workingtree_to_test_tree, return_parameter)
         self.assertEqual(tests[0].transport_server, server1)
         self.assertEqual(tests[0].transport_readonly_server, server2)
-        self.assertEqual(tests[1].intertree_class, formats[1][0])
-        self.assertEqual(tests[1].workingtree_format, formats[1][1])
-        self.assertEqual(tests[1].workingtree_format_to, formats[1][2])
-        self.assertEqual(tests[1].mutable_trees_to_test_trees, formats[1][3])
+        self.assertEqual(tests[1].intertree_class, formats[1][1])
+        self.assertEqual(tests[1].workingtree_format, formats[1][2])
+        self.assertEqual(tests[1].workingtree_format_to, formats[1][3])
+        self.assertEqual(tests[1].mutable_trees_to_test_trees, formats[1][4])
         self.assertEqual(tests[1]._workingtree_to_test_tree, return_parameter)
         self.assertEqual(tests[1].transport_server, server1)
         self.assertEqual(tests[1].transport_readonly_server, server2)
@@ -502,6 +502,19 @@
         self.assertIsSameRealPath(self.test_dir, cwd)
         self.assertIsSameRealPath(self.test_home_dir, os.environ['HOME'])
 
+    def test_assertEqualStat_equal(self):
+        from bzrlib.tests.test_dirstate import _FakeStat
+        self.build_tree(["foo"])
+        real = os.lstat("foo")
+        fake = _FakeStat(real.st_size, real.st_mtime, real.st_ctime,
+            real.st_dev, real.st_ino, real.st_mode)
+        self.assertEqualStat(real, fake)
+
+    def test_assertEqualStat_notequal(self):
+        self.build_tree(["foo", "bar"])
+        self.assertRaises(AssertionError, self.assertEqualStat,
+            os.lstat("foo"), os.lstat("bar"))
+
 
 class TestTestCaseWithMemoryTransport(TestCaseWithMemoryTransport):
 

=== modified file 'bzrlib/tests/test_sftp_transport.py'
--- bzrlib/tests/test_sftp_transport.py	2008-09-05 21:23:34 +0000
+++ bzrlib/tests/test_sftp_transport.py	2008-09-26 15:34:03 +0000
@@ -465,6 +465,8 @@
 class Test_SFTPReadvHelper(tests.TestCase):
 
     def checkGetRequests(self, expected_requests, offsets):
+        if not paramiko_loaded:
+            raise TestSkipped('you must have paramiko to run this test')
         helper = _mod_sftp._SFTPReadvHelper(offsets, 'artificial_test')
         self.assertEqual(expected_requests, helper._get_requests())
 

=== modified file 'bzrlib/tests/test_workingtree_4.py'
--- bzrlib/tests/test_workingtree_4.py	2008-07-08 14:55:19 +0000
+++ bzrlib/tests/test_workingtree_4.py	2008-09-22 05:15:20 +0000
@@ -18,6 +18,7 @@
 """Tests for WorkingTreeFormat4"""
 
 import os
+import time
 
 from bzrlib import (
     bzrdir,
@@ -577,6 +578,71 @@
         self.assertEqual([], changes)
         self.assertEqual(['', 'versioned', 'versioned2'], returned)
 
+    def get_tree_with_cachable_file_foo(self):
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo'])
+        tree.add(['foo'], ['foo-id'])
+        # a 4 second old timestamp is always hashable - sucks to delay 
+        # the test suite, but not testing this is worse.
+        time.sleep(4)
+        return tree
+
+    def test_commit_updates_hash_cache(self):
+        tree = self.get_tree_with_cachable_file_foo()
+        revid = tree.commit('a commit')
+        # tree's dirstate should now have a valid stat entry for foo.
+        tree.lock_read()
+        entry = tree._get_entry(path='foo')
+        expected_sha1 = osutils.sha_file_by_name('foo')
+        self.assertEqual(expected_sha1, entry[1][0][1])
+
+    def test_observed_sha1_cachable(self):
+        tree = self.get_tree_with_cachable_file_foo()
+        expected_sha1 = osutils.sha_file_by_name('foo')
+        statvalue = os.lstat("foo")
+        tree.lock_write()
+        try:
+            tree._observed_sha1("foo-id", "foo", (expected_sha1, statvalue))
+            self.assertEqual(expected_sha1,
+                tree._get_entry(path="foo")[1][0][1])
+        finally:
+            tree.unlock()
+        tree = tree.bzrdir.open_workingtree()
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        self.assertEqual(expected_sha1, tree._get_entry(path="foo")[1][0][1])
+
+    def test_observed_sha1_new_file(self):
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo'])
+        tree.add(['foo'], ['foo-id'])
+        tree.lock_read()
+        try:
+            current_sha1 = tree._get_entry(path="foo")[1][0][1]
+        finally:
+            tree.unlock()
+        tree.lock_write()
+        try:
+            tree._observed_sha1("foo-id", "foo",
+                (osutils.sha_file_by_name('foo'), os.lstat("foo")))
+            # Must not have changed
+            self.assertEqual(current_sha1,
+                tree._get_entry(path="foo")[1][0][1])
+        finally:
+            tree.unlock()
+
+    def test_get_file_with_stat_id_only(self):
+        # Explicit test to ensure we get a lstat value from WT4 trees.
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo'])
+        tree.add(['foo'], ['foo-id'])
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        file_obj, statvalue = tree.get_file_with_stat('foo-id')
+        expected = os.lstat('foo')
+        self.assertEqualStat(expected, statvalue)
+        self.assertEqual(["contents of foo\n"], file_obj.readlines())
+
 
 class TestCorruptDirstate(TestCaseWithTransport):
     """Tests for how we handle when the dirstate has been corrupted."""

=== modified file 'bzrlib/tests/tree_implementations/test_tree.py'
--- bzrlib/tests/tree_implementations/test_tree.py	2008-09-08 20:37:16 +0000
+++ bzrlib/tests/tree_implementations/test_tree.py	2008-09-08 20:43:46 +0000
@@ -17,6 +17,7 @@
 from bzrlib import (
     errors,
     conflicts,
+    osutils,
     revisiontree,
     tests,
     workingtree_4,
@@ -238,3 +239,38 @@
         tree.lock_read()
         self.addCleanup(tree.unlock)
         self.assertEqual(expected, list(tree.extras()))
+
+class TestHasId(TestCaseWithTree):
+
+    def test_has_id(self):
+        work_tree = self.make_branch_and_tree('tree')
+        self.build_tree(['tree/file'])
+        work_tree.add('file', 'file-id')
+        tree = self._convert_tree(work_tree)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        self.assertTrue(tree.has_id('file-id'))
+        self.assertFalse(tree.has_id('dir-id'))
+
+    def test___contains__(self):
+        work_tree = self.make_branch_and_tree('tree')
+        self.build_tree(['tree/file'])
+        work_tree.add('file', 'file-id')
+        tree = self._convert_tree(work_tree)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        self.assertTrue('file-id' in tree)
+        self.assertFalse('dir-id' in tree)
+
+
+class TestGetFileSha1(TestCaseWithTree):
+
+    def test_get_file_sha1(self):
+        work_tree = self.make_branch_and_tree('tree')
+        self.build_tree_contents([('tree/file', 'file content')])
+        work_tree.add('file', 'file-id')
+        tree = self._convert_tree(work_tree)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        expected = osutils.sha_strings('file content')
+        self.assertEqual(expected, tree.get_file_sha1('file-id'))

=== modified file 'bzrlib/tests/workingtree_implementations/__init__.py'
--- bzrlib/tests/workingtree_implementations/__init__.py	2008-07-17 00:42:40 +0000
+++ bzrlib/tests/workingtree_implementations/__init__.py	2008-09-22 05:15:20 +0000
@@ -101,6 +101,7 @@
         'bzrlib.tests.workingtree_implementations.test_commit',
         'bzrlib.tests.workingtree_implementations.test_executable',
         'bzrlib.tests.workingtree_implementations.test_flush',
+        'bzrlib.tests.workingtree_implementations.test_get_file_with_stat',
         'bzrlib.tests.workingtree_implementations.test_get_file_mtime',
         'bzrlib.tests.workingtree_implementations.test_get_parent_ids',
         'bzrlib.tests.workingtree_implementations.test_inv',

=== added file 'bzrlib/tests/workingtree_implementations/test_get_file_with_stat.py'
--- bzrlib/tests/workingtree_implementations/test_get_file_with_stat.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_get_file_with_stat.py	2008-09-22 05:15:20 +0000
@@ -0,0 +1,49 @@
+# Copyright (C) 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Test that all WorkingTree's implement get_file_with_stat."""
+
+import os
+
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+
+class TestGetFileWithStat(TestCaseWithWorkingTree):
+
+    def test_get_file_with_stat_id_only(self):
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo'])
+        tree.add(['foo'], ['foo-id'])
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        file_obj, statvalue = tree.get_file_with_stat('foo-id')
+        if statvalue is not None:
+            expected = os.lstat('foo')
+            self.assertEqualStat(expected, statvalue)
+        self.assertEqual(["contents of foo\n"], file_obj.readlines())
+
+    def test_get_file_with_stat_id_and_path(self):
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo'])
+        tree.add(['foo'], ['foo-id'])
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        file_obj, statvalue = tree.get_file_with_stat('foo-id', 'foo')
+        expected = os.lstat('foo')
+        if statvalue is not None:
+            expected = os.lstat('foo')
+            self.assertEqualStat(expected, statvalue)
+        self.assertEqual(["contents of foo\n"], file_obj.readlines())

=== modified file 'bzrlib/trace.py'
--- bzrlib/trace.py	2008-06-18 07:21:15 +0000
+++ bzrlib/trace.py	2008-09-26 22:14:42 +0000
@@ -388,6 +388,30 @@
     pass
 
 
+_short_fields = ('VmPeak', 'VmSize', 'VmRSS')
+
+def debug_memory(message='', short=True):
+    """Write out a memory dump."""
+    try:
+        status_file = file('/proc/%s/status' % os.getpid(), 'rb')
+    except IOError:
+        return
+    try:
+        status = status_file.read()
+    finally:
+        status_file.close()
+    if message:
+        note(message)
+    for line in status.splitlines():
+        if not short:
+            note(line)
+        else:
+            for field in _short_fields:
+                if line.startswith(field):
+                    note(line)
+                    break
+
+
 def report_exception(exc_info, err_file):
     """Report an exception to err_file (typically stderr) and to .bzr.log.
 

=== modified file 'bzrlib/transform.py'
--- bzrlib/transform.py	2008-09-08 20:38:46 +0000
+++ bzrlib/transform.py	2008-09-08 20:43:46 +0000
@@ -41,6 +41,7 @@
     has_symlinks,
     lexists,
     pathjoin,
+    sha_file,
     splitpath,
     supports_executable,
 )
@@ -1666,7 +1667,16 @@
             return None
 
     def get_file_sha1(self, file_id, path=None, stat_value=None):
-        return self._transform._tree.get_file_sha1(file_id)
+        trans_id = self._transform.trans_id_file_id(file_id)
+        kind = self._transform._new_contents.get(trans_id)
+        if kind is None:
+            return self._transform._tree.get_file_sha1(file_id)
+        if kind == 'file':
+            fileobj = self.get_file(file_id)
+            try:
+                return sha_file(fileobj)
+            finally:
+                fileobj.close()
 
     def is_executable(self, file_id, path=None):
         if file_id is None:

=== modified file 'bzrlib/transport/http/wsgi.py'
--- bzrlib/transport/http/wsgi.py	2008-05-16 05:55:23 +0000
+++ bzrlib/transport/http/wsgi.py	2008-09-14 21:48:40 +0000
@@ -27,7 +27,8 @@
 from bzrlib.urlutils import local_path_to_url
     
 
-def make_app(root, prefix, path_var='REQUEST_URI', readonly=True):
+def make_app(root, prefix, path_var='REQUEST_URI', readonly=True,
+    load_plugins=True, enable_logging=True):
     """Convenience function to construct a WSGI bzr smart server.
     
     :param root: a local path that requests will be relative to.
@@ -39,6 +40,12 @@
         base_transport = get_transport('readonly+' + local_url)
     else:
         base_transport = get_transport(local_url)
+    if load_plugins:
+        from bzrlib.plugin import load_plugins
+        load_plugins()
+    if enable_logging:
+        import bzrlib.trace
+        bzrlib.trace.enable_default_logging()
     app = SmartWSGIApp(base_transport, prefix)
     app = RelpathSetter(app, '', path_var)
     return app

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2008-09-20 17:20:39 +0000
+++ bzrlib/tree.py	2008-09-26 23:55:32 +0000
@@ -132,7 +132,8 @@
     def has_id(self, file_id):
         return self.inventory.has_id(file_id)
 
-    __contains__ = has_id
+    def __contains__(self, file_id):
+        return self.has_id(file_id)
 
     def has_or_had_id(self, file_id):
         if file_id == self.inventory.root.file_id:

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2008-09-22 09:43:40 +0000
+++ bzrlib/workingtree.py	2008-09-25 01:11:30 +0000
@@ -424,9 +424,14 @@
         return osutils.lexists(self.abspath(filename))
 
     def get_file(self, file_id, path=None):
+        return self.get_file_with_stat(file_id, path)[0]
+
+    def get_file_with_stat(self, file_id, path=None, _fstat=os.fstat):
+        """See MutableTree.get_file_with_stat."""
         if path is None:
             path = self.id2path(file_id)
-        return self.get_file_byname(path)
+        file_obj = self.get_file_byname(path)
+        return (file_obj, _fstat(file_obj.fileno()))
 
     def get_file_text(self, file_id):
         return self.get_file(file_id).read()

=== modified file 'bzrlib/workingtree_4.py'
--- bzrlib/workingtree_4.py	2008-09-08 20:55:25 +0000
+++ bzrlib/workingtree_4.py	2008-09-25 01:54:42 +0000
@@ -98,13 +98,6 @@
 from bzrlib.workingtree import WorkingTree, WorkingTree3, WorkingTreeFormat3
 
 
-# This is the Windows equivalent of ENOTDIR
-# It is defined in pywin32.winerror, but we don't want a strong dependency for
-# just an error code.
-ERROR_PATH_NOT_FOUND = 3
-ERROR_DIRECTORY = 267
-
-
 class WorkingTree4(WorkingTree3):
     """This is the Format 4 working tree.
 
@@ -149,6 +142,8 @@
         self._setup_directory_is_tree_reference()
         self._detect_case_handling()
         self._rules_searcher = None
+        #--- allow tests to select the dirstate iter_changes implementation
+        self._iter_changes = dirstate._process_entry
 
     @needs_tree_write_lock
     def _add(self, files, ids, kinds):
@@ -409,10 +404,19 @@
                     return None
                 else:
                     raise
-        link_or_sha1 = state.update_entry(entry, file_abspath,
-                                          stat_value=stat_value)
+        link_or_sha1 = dirstate.update_entry(state, entry, file_abspath,
+            stat_value=stat_value)
         if entry[1][0][0] == 'f':
-            return link_or_sha1
+            if link_or_sha1 is None:
+                file_obj, statvalue = self.get_file_with_stat(file_id, path)
+                try:
+                    sha1 = osutils.sha_file(file_obj)
+                finally:
+                    file_obj.close()
+                self._observed_sha1(file_id, path, (sha1, statvalue))
+                return sha1
+            else:
+                return link_or_sha1
         return None
 
     def _get_inventory(self):
@@ -548,6 +552,12 @@
                 # path is missing on disk.
                 continue
 
+    def _observed_sha1(self, file_id, path, (sha1, statvalue)):
+        """See MutableTree._observed_sha1."""
+        state = self.current_dirstate()
+        entry = self._get_entry(file_id=file_id, path=path)
+        state._observed_sha1(entry, sha1, statvalue)
+
     def kind(self, file_id):
         """Return the kind of a file.
 
@@ -1810,12 +1820,32 @@
         target.set_parent_ids([revid])
         return target.basis_tree(), target
 
+    @classmethod
+    def make_source_parent_tree_python_dirstate(klass, test_case, source, target):
+        result = klass.make_source_parent_tree(source, target)
+        result[1]._iter_changes = dirstate.ProcessEntryPython
+        return result
+
+    @classmethod
+    def make_source_parent_tree_compiled_dirstate(klass, test_case, source, target):
+        from bzrlib.tests.test__dirstate_helpers import \
+            CompiledDirstateHelpersFeature
+        if not CompiledDirstateHelpersFeature.available():
+            from bzrlib.tests import UnavailableFeature
+            raise UnavailableFeature(CompiledDirstateHelpersFeature)
+        from bzrlib._dirstate_helpers_c import ProcessEntryC
+        result = klass.make_source_parent_tree(source, target)
+        result[1]._iter_changes = ProcessEntryC
+        return result
+
     _matching_from_tree_format = WorkingTreeFormat4()
     _matching_to_tree_format = WorkingTreeFormat4()
 
     @classmethod
     def _test_mutable_trees_to_test_trees(klass, test_case, source, target):
-        return klass.make_source_parent_tree(source, target)
+        # This method shouldn't be called, because we have python and C
+        # specific flavours.
+        raise NotImplementedError
 
     def iter_changes(self, include_unchanged=False,
                       specific_files=None, pb=None, extra_trees=[],
@@ -1839,19 +1869,14 @@
             output. An unversioned file is defined as one with (False, False)
             for the versioned pair.
         """
-        utf8_decode = cache_utf8._utf8_decode
-        _minikind_to_kind = dirstate.DirState._minikind_to_kind
-        cmp_by_dirs = dirstate.cmp_by_dirs
         # NB: show_status depends on being able to pass in non-versioned files
         # and report them as unknown
         # TODO: handle extra trees in the dirstate.
         if (extra_trees or specific_files == []):
             # we can't fast-path these cases (yet)
-            for f in super(InterDirStateTree, self).iter_changes(
+            return super(InterDirStateTree, self).iter_changes(
                 include_unchanged, specific_files, pb, extra_trees,
-                require_versioned, want_unversioned=want_unversioned):
-                yield f
-            return
+                require_versioned, want_unversioned=want_unversioned)
         parent_ids = self.target.get_parent_ids()
         if not (self.source._revision_id in parent_ids
                 or self.source._revision_id == NULL_REVISION):
@@ -1874,37 +1899,22 @@
         if specific_files:
             specific_files_utf8 = set()
             for path in specific_files:
+                # Note, if there are many specific files, using cache_utf8
+                # would be good here.
                 specific_files_utf8.add(path.encode('utf8'))
             specific_files = specific_files_utf8
         else:
             specific_files = set([''])
         # -- specific_files is now a utf8 path set --
+        search_specific_files = set()
         # -- get the state object and prepare it.
         state = self.target.current_dirstate()
         state._read_dirblocks_if_needed()
-        def _entries_for_path(path):
-            """Return a list with all the entries that match path for all ids.
-            """
-            dirname, basename = os.path.split(path)
-            key = (dirname, basename, '')
-            block_index, present = state._find_block_index_from_key(key)
-            if not present:
-                # the block which should contain path is absent.
-                return []
-            result = []
-            block = state._dirblocks[block_index][1]
-            entry_index, _ = state._find_entry_index(key, block)
-            # we may need to look at multiple entries at this path: walk while the specific_files match.
-            while (entry_index < len(block) and
-                block[entry_index][0][0:2] == key[0:2]):
-                result.append(block[entry_index])
-                entry_index += 1
-            return result
         if require_versioned:
             # -- check all supplied paths are versioned in a search tree. --
             all_versioned = True
             for path in specific_files:
-                path_entries = _entries_for_path(path)
+                path_entries = state._entries_for_path(path)
                 if not path_entries:
                     # this specified path is not present at all: error
                     all_versioned = False
@@ -1926,610 +1936,17 @@
             if not all_versioned:
                 raise errors.PathsNotVersionedError(specific_files)
         # -- remove redundancy in supplied specific_files to prevent over-scanning --
-        search_specific_files = set()
         for path in specific_files:
             other_specific_files = specific_files.difference(set([path]))
             if not osutils.is_inside_any(other_specific_files, path):
                 # this is a top level path, we must check it.
                 search_specific_files.add(path)
-        # sketch: 
-        # compare source_index and target_index at or under each element of search_specific_files.
-        # follow the following comparison table. Note that we only want to do diff operations when
-        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
-        # for the target.
-        # cases:
-        # 
-        # Source | Target | disk | action
-        #   r    | fdlt   |      | add source to search, add id path move and perform
-        #        |        |      | diff check on source-target
-        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
-        #        |        |      | ???
-        #   r    |  a     |      | add source to search
-        #   r    |  a     |  a   | 
-        #   r    |  r     |      | this path is present in a non-examined tree, skip.
-        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
-        #   a    | fdlt   |      | add new id
-        #   a    | fdlt   |  a   | dangling locally added file, skip
-        #   a    |  a     |      | not present in either tree, skip
-        #   a    |  a     |  a   | not present in any tree, skip
-        #   a    |  r     |      | not present in either tree at this path, skip as it
-        #        |        |      | may not be selected by the users list of paths.
-        #   a    |  r     |  a   | not present in either tree at this path, skip as it
-        #        |        |      | may not be selected by the users list of paths.
-        #  fdlt  | fdlt   |      | content in both: diff them
-        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
-        #  fdlt  |  a     |      | unversioned: output deleted id for now
-        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
-        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
-        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
-        #        |        |      | this id at the other path.
-        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
-        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
-        #        |        |      | this id at the other path.
-
-        # for all search_indexs in each path at or under each element of
-        # search_specific_files, if the detail is relocated: add the id, and add the
-        # relocated path as one to search if its not searched already. If the
-        # detail is not relocated, add the id.
-        searched_specific_files = set()
-        NULL_PARENT_DETAILS = dirstate.DirState.NULL_PARENT_DETAILS
-        # Using a list so that we can access the values and change them in
-        # nested scope. Each one is [path, file_id, entry]
-        last_source_parent = [None, None]
-        last_target_parent = [None, None]
 
         use_filesystem_for_exec = (sys.platform != 'win32')
-
-        # Just a sentry, so that _process_entry can say that this
-        # record is handled, but isn't interesting to process (unchanged)
-        uninteresting = object()
-
-        old_dirname_to_file_id = {}
-        new_dirname_to_file_id = {}
-        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
-        #       keeping a cache of directories that we have seen.
-
-        def _process_entry(entry, path_info):
-            """Compare an entry and real disk to generate delta information.
-
-            :param path_info: top_relpath, basename, kind, lstat, abspath for
-                the path of entry. If None, then the path is considered absent.
-                (Perhaps we should pass in a concrete entry for this ?)
-                Basename is returned as a utf8 string because we expect this
-                tuple will be ignored, and don't want to take the time to
-                decode.
-            :return: None if these don't match
-                     A tuple of information about the change, or
-                     the object 'uninteresting' if these match, but are
-                     basically identical.
-            """
-            if source_index is None:
-                source_details = NULL_PARENT_DETAILS
-            else:
-                source_details = entry[1][source_index]
-            target_details = entry[1][target_index]
-            target_minikind = target_details[0]
-            if path_info is not None and target_minikind in 'fdlt':
-                if not (target_index == 0):
-                    raise AssertionError()
-                link_or_sha1 = state.update_entry(entry, abspath=path_info[4],
-                                                  stat_value=path_info[3])
-                # The entry may have been modified by update_entry
-                target_details = entry[1][target_index]
-                target_minikind = target_details[0]
-            else:
-                link_or_sha1 = None
-            file_id = entry[0][2]
-            source_minikind = source_details[0]
-            if source_minikind in 'fdltr' and target_minikind in 'fdlt':
-                # claimed content in both: diff
-                #   r    | fdlt   |      | add source to search, add id path move and perform
-                #        |        |      | diff check on source-target
-                #   r    | fdlt   |  a   | dangling file that was present in the basis.
-                #        |        |      | ???
-                if source_minikind in 'r':
-                    # add the source to the search path to find any children it
-                    # has.  TODO ? : only add if it is a container ?
-                    if not osutils.is_inside_any(searched_specific_files,
-                                                 source_details[1]):
-                        search_specific_files.add(source_details[1])
-                    # generate the old path; this is needed for stating later
-                    # as well.
-                    old_path = source_details[1]
-                    old_dirname, old_basename = os.path.split(old_path)
-                    path = pathjoin(entry[0][0], entry[0][1])
-                    old_entry = state._get_entry(source_index,
-                                                 path_utf8=old_path)
-                    # update the source details variable to be the real
-                    # location.
-                    if old_entry == (None, None):
-                        raise errors.CorruptDirstate(state._filename,
-                            "entry '%s/%s' is considered renamed from %r"
-                            " but source does not exist\n"
-                            "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
-                    source_details = old_entry[1][source_index]
-                    source_minikind = source_details[0]
-                else:
-                    old_dirname = entry[0][0]
-                    old_basename = entry[0][1]
-                    old_path = path = None
-                if path_info is None:
-                    # the file is missing on disk, show as removed.
-                    content_change = True
-                    target_kind = None
-                    target_exec = False
-                else:
-                    # source and target are both versioned and disk file is present.
-                    target_kind = path_info[2]
-                    if target_kind == 'directory':
-                        if path is None:
-                            old_path = path = pathjoin(old_dirname, old_basename)
-                        new_dirname_to_file_id[path] = file_id
-                        if source_minikind != 'd':
-                            content_change = True
-                        else:
-                            # directories have no fingerprint
-                            content_change = False
-                        target_exec = False
-                    elif target_kind == 'file':
-                        if source_minikind != 'f':
-                            content_change = True
-                        else:
-                            # We could check the size, but we already have the
-                            # sha1 hash.
-                            content_change = (link_or_sha1 != source_details[1])
-                        # Target details is updated at update_entry time
-                        if use_filesystem_for_exec:
-                            # We don't need S_ISREG here, because we are sure
-                            # we are dealing with a file.
-                            target_exec = bool(stat.S_IEXEC & path_info[3].st_mode)
-                        else:
-                            target_exec = target_details[3]
-                    elif target_kind == 'symlink':
-                        if source_minikind != 'l':
-                            content_change = True
-                        else:
-                            content_change = (link_or_sha1 != source_details[1])
-                        target_exec = False
-                    elif target_kind == 'tree-reference':
-                        if source_minikind != 't':
-                            content_change = True
-                        else:
-                            content_change = False
-                        target_exec = False
-                    else:
-                        raise Exception, "unknown kind %s" % path_info[2]
-                if source_minikind == 'd':
-                    if path is None:
-                        old_path = path = pathjoin(old_dirname, old_basename)
-                    old_dirname_to_file_id[old_path] = file_id
-                # parent id is the entry for the path in the target tree
-                if old_dirname == last_source_parent[0]:
-                    source_parent_id = last_source_parent[1]
-                else:
-                    try:
-                        source_parent_id = old_dirname_to_file_id[old_dirname]
-                    except KeyError:
-                        source_parent_entry = state._get_entry(source_index,
-                                                               path_utf8=old_dirname)
-                        source_parent_id = source_parent_entry[0][2]
-                    if source_parent_id == entry[0][2]:
-                        # This is the root, so the parent is None
-                        source_parent_id = None
-                    else:
-                        last_source_parent[0] = old_dirname
-                        last_source_parent[1] = source_parent_id
-                new_dirname = entry[0][0]
-                if new_dirname == last_target_parent[0]:
-                    target_parent_id = last_target_parent[1]
-                else:
-                    try:
-                        target_parent_id = new_dirname_to_file_id[new_dirname]
-                    except KeyError:
-                        # TODO: We don't always need to do the lookup, because the
-                        #       parent entry will be the same as the source entry.
-                        target_parent_entry = state._get_entry(target_index,
-                                                               path_utf8=new_dirname)
-                        if target_parent_entry == (None, None):
-                            raise AssertionError(
-                                "Could not find target parent in wt: %s\nparent of: %s"
-                                % (new_dirname, entry))
-                        target_parent_id = target_parent_entry[0][2]
-                    if target_parent_id == entry[0][2]:
-                        # This is the root, so the parent is None
-                        target_parent_id = None
-                    else:
-                        last_target_parent[0] = new_dirname
-                        last_target_parent[1] = target_parent_id
-
-                source_exec = source_details[3]
-                if (include_unchanged
-                    or content_change
-                    or source_parent_id != target_parent_id
-                    or old_basename != entry[0][1]
-                    or source_exec != target_exec
-                    ):
-                    if old_path is None:
-                        old_path = path = pathjoin(old_dirname, old_basename)
-                        old_path_u = utf8_decode(old_path)[0]
-                        path_u = old_path_u
-                    else:
-                        old_path_u = utf8_decode(old_path)[0]
-                        if old_path == path:
-                            path_u = old_path_u
-                        else:
-                            path_u = utf8_decode(path)[0]
-                    source_kind = _minikind_to_kind[source_minikind]
-                    return (entry[0][2],
-                           (old_path_u, path_u),
-                           content_change,
-                           (True, True),
-                           (source_parent_id, target_parent_id),
-                           (utf8_decode(old_basename)[0], utf8_decode(entry[0][1])[0]),
-                           (source_kind, target_kind),
-                           (source_exec, target_exec))
-                else:
-                    return uninteresting
-            elif source_minikind in 'a' and target_minikind in 'fdlt':
-                # looks like a new file
-                path = pathjoin(entry[0][0], entry[0][1])
-                # parent id is the entry for the path in the target tree
-                # TODO: these are the same for an entire directory: cache em.
-                parent_id = state._get_entry(target_index,
-                                             path_utf8=entry[0][0])[0][2]
-                if parent_id == entry[0][2]:
-                    parent_id = None
-                if path_info is not None:
-                    # Present on disk:
-                    if use_filesystem_for_exec:
-                        # We need S_ISREG here, because we aren't sure if this
-                        # is a file or not.
-                        target_exec = bool(
-                            stat.S_ISREG(path_info[3].st_mode)
-                            and stat.S_IEXEC & path_info[3].st_mode)
-                    else:
-                        target_exec = target_details[3]
-                    return (entry[0][2],
-                           (None, utf8_decode(path)[0]),
-                           True,
-                           (False, True),
-                           (None, parent_id),
-                           (None, utf8_decode(entry[0][1])[0]),
-                           (None, path_info[2]),
-                           (None, target_exec))
-                else:
-                    # Its a missing file, report it as such.
-                    return (entry[0][2],
-                           (None, utf8_decode(path)[0]),
-                           False,
-                           (False, True),
-                           (None, parent_id),
-                           (None, utf8_decode(entry[0][1])[0]),
-                           (None, None),
-                           (None, False))
-            elif source_minikind in 'fdlt' and target_minikind in 'a':
-                # unversioned, possibly, or possibly not deleted: we dont care.
-                # if its still on disk, *and* theres no other entry at this
-                # path [we dont know this in this routine at the moment -
-                # perhaps we should change this - then it would be an unknown.
-                old_path = pathjoin(entry[0][0], entry[0][1])
-                # parent id is the entry for the path in the target tree
-                parent_id = state._get_entry(source_index, path_utf8=entry[0][0])[0][2]
-                if parent_id == entry[0][2]:
-                    parent_id = None
-                return (entry[0][2],
-                       (utf8_decode(old_path)[0], None),
-                       True,
-                       (True, False),
-                       (parent_id, None),
-                       (utf8_decode(entry[0][1])[0], None),
-                       (_minikind_to_kind[source_minikind], None),
-                       (source_details[3], None))
-            elif source_minikind in 'fdlt' and target_minikind in 'r':
-                # a rename; could be a true rename, or a rename inherited from
-                # a renamed parent. TODO: handle this efficiently. Its not
-                # common case to rename dirs though, so a correct but slow
-                # implementation will do.
-                if not osutils.is_inside_any(searched_specific_files, target_details[1]):
-                    search_specific_files.add(target_details[1])
-            elif source_minikind in 'ra' and target_minikind in 'ra':
-                # neither of the selected trees contain this file,
-                # so skip over it. This is not currently directly tested, but
-                # is indirectly via test_too_much.TestCommands.test_conflicts.
-                pass
-            else:
-                raise AssertionError("don't know how to compare "
-                    "source_minikind=%r, target_minikind=%r"
-                    % (source_minikind, target_minikind))
-                ## import pdb;pdb.set_trace()
-            return None
-
-        while search_specific_files:
-            # TODO: the pending list should be lexically sorted?  the
-            # interface doesn't require it.
-            current_root = search_specific_files.pop()
-            current_root_unicode = current_root.decode('utf8')
-            searched_specific_files.add(current_root)
-            # process the entries for this containing directory: the rest will be
-            # found by their parents recursively.
-            root_entries = _entries_for_path(current_root)
-            root_abspath = self.target.abspath(current_root_unicode)
-            try:
-                root_stat = os.lstat(root_abspath)
-            except OSError, e:
-                if e.errno == errno.ENOENT:
-                    # the path does not exist: let _process_entry know that.
-                    root_dir_info = None
-                else:
-                    # some other random error: hand it up.
-                    raise
-            else:
-                root_dir_info = ('', current_root,
-                    osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
-                    root_abspath)
-                if root_dir_info[2] == 'directory':
-                    if self.target._directory_is_tree_reference(
-                        current_root.decode('utf8')):
-                        root_dir_info = root_dir_info[:2] + \
-                            ('tree-reference',) + root_dir_info[3:]
-
-            if not root_entries and not root_dir_info:
-                # this specified path is not present at all, skip it.
-                continue
-            path_handled = False
-            for entry in root_entries:
-                result = _process_entry(entry, root_dir_info)
-                if result is not None:
-                    path_handled = True
-                    if result is not uninteresting:
-                        yield result
-            if want_unversioned and not path_handled and root_dir_info:
-                new_executable = bool(
-                    stat.S_ISREG(root_dir_info[3].st_mode)
-                    and stat.S_IEXEC & root_dir_info[3].st_mode)
-                yield (None,
-                       (None, current_root_unicode),
-                       True,
-                       (False, False),
-                       (None, None),
-                       (None, splitpath(current_root_unicode)[-1]),
-                       (None, root_dir_info[2]),
-                       (None, new_executable)
-                      )
-            initial_key = (current_root, '', '')
-            block_index, _ = state._find_block_index_from_key(initial_key)
-            if block_index == 0:
-                # we have processed the total root already, but because the
-                # initial key matched it we should skip it here.
-                block_index +=1
-            if root_dir_info and root_dir_info[2] == 'tree-reference':
-                current_dir_info = None
-            else:
-                dir_iterator = osutils._walkdirs_utf8(root_abspath, prefix=current_root)
-                try:
-                    current_dir_info = dir_iterator.next()
-                except OSError, e:
-                    # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
-                    # python 2.5 has e.errno == EINVAL,
-                    #            and e.winerror == ERROR_DIRECTORY
-                    e_winerror = getattr(e, 'winerror', None)
-                    win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
-                    # there may be directories in the inventory even though
-                    # this path is not a file on disk: so mark it as end of
-                    # iterator
-                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
-                        current_dir_info = None
-                    elif (sys.platform == 'win32'
-                          and (e.errno in win_errors
-                               or e_winerror in win_errors)):
-                        current_dir_info = None
-                    else:
-                        raise
-                else:
-                    if current_dir_info[0][0] == '':
-                        # remove .bzr from iteration
-                        bzr_index = bisect_left(current_dir_info[1], ('.bzr',))
-                        if current_dir_info[1][bzr_index][0] != '.bzr':
-                            raise AssertionError()
-                        del current_dir_info[1][bzr_index]
-            # walk until both the directory listing and the versioned metadata
-            # are exhausted. 
-            if (block_index < len(state._dirblocks) and
-                osutils.is_inside(current_root, state._dirblocks[block_index][0])):
-                current_block = state._dirblocks[block_index]
-            else:
-                current_block = None
-            while (current_dir_info is not None or
-                   current_block is not None):
-                if (current_dir_info and current_block
-                    and current_dir_info[0][0] != current_block[0]):
-                    if cmp_by_dirs(current_dir_info[0][0], current_block[0]) < 0:
-                        # filesystem data refers to paths not covered by the dirblock.
-                        # this has two possibilities:
-                        # A) it is versioned but empty, so there is no block for it
-                        # B) it is not versioned.
-
-                        # if (A) then we need to recurse into it to check for
-                        # new unknown files or directories.
-                        # if (B) then we should ignore it, because we don't
-                        # recurse into unknown directories.
-                        path_index = 0
-                        while path_index < len(current_dir_info[1]):
-                                current_path_info = current_dir_info[1][path_index]
-                                if want_unversioned:
-                                    if current_path_info[2] == 'directory':
-                                        if self.target._directory_is_tree_reference(
-                                            current_path_info[0].decode('utf8')):
-                                            current_path_info = current_path_info[:2] + \
-                                                ('tree-reference',) + current_path_info[3:]
-                                    new_executable = bool(
-                                        stat.S_ISREG(current_path_info[3].st_mode)
-                                        and stat.S_IEXEC & current_path_info[3].st_mode)
-                                    yield (None,
-                                        (None, utf8_decode(current_path_info[0])[0]),
-                                        True,
-                                        (False, False),
-                                        (None, None),
-                                        (None, utf8_decode(current_path_info[1])[0]),
-                                        (None, current_path_info[2]),
-                                        (None, new_executable))
-                                # dont descend into this unversioned path if it is
-                                # a dir
-                                if current_path_info[2] in ('directory',
-                                                            'tree-reference'):
-                                    del current_dir_info[1][path_index]
-                                    path_index -= 1
-                                path_index += 1
-
-                        # This dir info has been handled, go to the next
-                        try:
-                            current_dir_info = dir_iterator.next()
-                        except StopIteration:
-                            current_dir_info = None
-                    else:
-                        # We have a dirblock entry for this location, but there
-                        # is no filesystem path for this. This is most likely
-                        # because a directory was removed from the disk.
-                        # We don't have to report the missing directory,
-                        # because that should have already been handled, but we
-                        # need to handle all of the files that are contained
-                        # within.
-                        for current_entry in current_block[1]:
-                            # entry referring to file not present on disk.
-                            # advance the entry only, after processing.
-                            result = _process_entry(current_entry, None)
-                            if result is not None:
-                                if result is not uninteresting:
-                                    yield result
-                        block_index +=1
-                        if (block_index < len(state._dirblocks) and
-                            osutils.is_inside(current_root,
-                                              state._dirblocks[block_index][0])):
-                            current_block = state._dirblocks[block_index]
-                        else:
-                            current_block = None
-                    continue
-                entry_index = 0
-                if current_block and entry_index < len(current_block[1]):
-                    current_entry = current_block[1][entry_index]
-                else:
-                    current_entry = None
-                advance_entry = True
-                path_index = 0
-                if current_dir_info and path_index < len(current_dir_info[1]):
-                    current_path_info = current_dir_info[1][path_index]
-                    if current_path_info[2] == 'directory':
-                        if self.target._directory_is_tree_reference(
-                            current_path_info[0].decode('utf8')):
-                            current_path_info = current_path_info[:2] + \
-                                ('tree-reference',) + current_path_info[3:]
-                else:
-                    current_path_info = None
-                advance_path = True
-                path_handled = False
-                while (current_entry is not None or
-                    current_path_info is not None):
-                    if current_entry is None:
-                        # the check for path_handled when the path is adnvaced
-                        # will yield this path if needed.
-                        pass
-                    elif current_path_info is None:
-                        # no path is fine: the per entry code will handle it.
-                        result = _process_entry(current_entry, current_path_info)
-                        if result is not None:
-                            if result is not uninteresting:
-                                yield result
-                    elif (current_entry[0][1] != current_path_info[1]
-                          or current_entry[1][target_index][0] in 'ar'):
-                        # The current path on disk doesn't match the dirblock
-                        # record. Either the dirblock is marked as absent, or
-                        # the file on disk is not present at all in the
-                        # dirblock. Either way, report about the dirblock
-                        # entry, and let other code handle the filesystem one.
-
-                        # Compare the basename for these files to determine
-                        # which comes first
-                        if current_path_info[1] < current_entry[0][1]:
-                            # extra file on disk: pass for now, but only
-                            # increment the path, not the entry
-                            advance_entry = False
-                        else:
-                            # entry referring to file not present on disk.
-                            # advance the entry only, after processing.
-                            result = _process_entry(current_entry, None)
-                            if result is not None:
-                                if result is not uninteresting:
-                                    yield result
-                            advance_path = False
-                    else:
-                        result = _process_entry(current_entry, current_path_info)
-                        if result is not None:
-                            path_handled = True
-                            if result is not uninteresting:
-                                yield result
-                    if advance_entry and current_entry is not None:
-                        entry_index += 1
-                        if entry_index < len(current_block[1]):
-                            current_entry = current_block[1][entry_index]
-                        else:
-                            current_entry = None
-                    else:
-                        advance_entry = True # reset the advance flaga
-                    if advance_path and current_path_info is not None:
-                        if not path_handled:
-                            # unversioned in all regards
-                            if want_unversioned:
-                                new_executable = bool(
-                                    stat.S_ISREG(current_path_info[3].st_mode)
-                                    and stat.S_IEXEC & current_path_info[3].st_mode)
-                                try:
-                                    relpath_unicode = utf8_decode(current_path_info[0])[0]
-                                except UnicodeDecodeError:
-                                    raise errors.BadFilenameEncoding(
-                                        current_path_info[0], osutils._fs_enc)
-                                yield (None,
-                                    (None, relpath_unicode),
-                                    True,
-                                    (False, False),
-                                    (None, None),
-                                    (None, utf8_decode(current_path_info[1])[0]),
-                                    (None, current_path_info[2]),
-                                    (None, new_executable))
-                            # dont descend into this unversioned path if it is
-                            # a dir
-                            if current_path_info[2] in ('directory'):
-                                del current_dir_info[1][path_index]
-                                path_index -= 1
-                        # dont descend the disk iterator into any tree 
-                        # paths.
-                        if current_path_info[2] == 'tree-reference':
-                            del current_dir_info[1][path_index]
-                            path_index -= 1
-                        path_index += 1
-                        if path_index < len(current_dir_info[1]):
-                            current_path_info = current_dir_info[1][path_index]
-                            if current_path_info[2] == 'directory':
-                                if self.target._directory_is_tree_reference(
-                                    current_path_info[0].decode('utf8')):
-                                    current_path_info = current_path_info[:2] + \
-                                        ('tree-reference',) + current_path_info[3:]
-                        else:
-                            current_path_info = None
-                        path_handled = False
-                    else:
-                        advance_path = True # reset the advance flagg.
-                if current_block is not None:
-                    block_index += 1
-                    if (block_index < len(state._dirblocks) and
-                        osutils.is_inside(current_root, state._dirblocks[block_index][0])):
-                        current_block = state._dirblocks[block_index]
-                    else:
-                        current_block = None
-                if current_dir_info is not None:
-                    try:
-                        current_dir_info = dir_iterator.next()
-                    except StopIteration:
-                        current_dir_info = None
+        iter_changes = self.target._iter_changes(include_unchanged,
+            use_filesystem_for_exec, search_specific_files, state,
+            source_index, target_index, want_unversioned, self.target)
+        return iter_changes.iter_changes()
 
     @staticmethod
     def is_compatible(source, target):

=== modified file 'doc/developers/development-repo.txt'
--- doc/developers/development-repo.txt	2008-08-28 06:26:43 +0000
+++ doc/developers/development-repo.txt	2008-09-25 22:25:09 +0000
@@ -226,17 +226,18 @@
 development
 -----------
 
-Currently an alias for Development1
+Currently an alias for Development2
 
 development-subtree
 -------------------
 
-Currently an alias for Development1Subtree
+Currently an alias for Development2Subtree
 
-Development1[Subtree]
+Development2[Subtree]
 ---------------------
 
-These formats were introduced during the development of branch stacking.
+These formats use B+Tree indices which are considerably faster than
+the earlier indices in bzr.
 
 
 ..

=== added file 'doc/developers/lca_tree_merging.txt'
--- doc/developers/lca_tree_merging.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/lca_tree_merging.txt	2008-09-05 02:29:34 +0000
@@ -0,0 +1,126 @@
+================
+LCA Tree Merging
+================
+
+There are 2 ways that you get LCA merge resolution in bzr. First, if you use
+``bzr merge --lca``, the *content* of files will be resolved using a Least Common
+Ancestors algorithm. That is described in <lca-merge.html> not here.
+
+This document describes how we handle merging tree-shape when there is not
+a single unique ancestor (criss-cross merge). With a single LCA, we use
+simple 3-way-merge logic.
+
+When there are multiple possible LCAs, we use a different algorithm for
+handling tree-shape merging. Described here.
+
+As a simple example, here is a revision graph which we will refer to often::
+
+  .    BASE
+  .  /      \
+  . LCA1   LCA2
+  . |   \ /   |
+  . |    X    |
+  . |   / \   |
+  . THIS  OTHER
+
+In this graph, ``THIS`` and ``OTHER`` both have ``LCA1`` and ``LCA2`` in their
+ancestry but neither is an ancestor of the other, so we have 2 least common
+ancestors. The unique common ancestor is ``BASE``. (It should be noted that in
+this text we will talk directly about ``LCA1`` and ``LCA2``, but the algorithms
+are designed to cope with more than 2 LCAs.)
+
+
+Scalars
+=======
+
+Definition
+----------
+
+I'm defining scalar values as ones that cannot be 'merged' on their own. For
+example, the name of a file is "scalar". If one person changes "foo.txt" to
+"foo.c" and someone else changes "foo.txt" to "bar.txt" we don't merge the
+changes to be "bar.c", we simply conflict and expect the user to sort it out.
+
+We use a slightly different algorithm for scalars.
+
+
+Resolution Algorithm
+--------------------
+
+(This can be seen as ``bzrlib.merge.Merge3Merger._lca_multi_way```
+
+1. If ``THIS`` and ``OTHER`` have the same value, use it. There is no need to
+   inspect any other values in this case. Either nothing was changed (all
+   interesting nodes would have the same value), or we have "accidental
+   convergence" (both sides made the same change.).
+
+2. Find the values from ``LCA1`` and ``LCA2`` which are not the same as
+   ``BASE``. The idea here is to provide a rudimentary "heads" comparison.
+   Often, the whole tree graph will have a criss-cross, but the per-file
+   (per-scalar) graph would be linear, and the value in one LCA strictly
+   dominates the other. It is possible to construct a scenario where one side
+   dominates the other, but the dominated value is not ``BASE``, but a second
+   intermediate value. Most scalars are rarely changed, so this is unlikely to
+   be an issue. The trade-off is having to generate and inspect the
+   per-scalar graph.
+
+   If there are no LCA values that are different from ``BASE``, we use a simple
+   3-way merge with ``BASE`` as the base value.
+
+3. Find the unique set of LCA values that do not include the ``BASE`` value.
+   If there is only one unique LCA value, we again use three-way merge logic
+   using that unique value as the base.
+
+4. At this point, we have determined that we have at least 2 unique values in
+   our LCAs which means that ``THIS`` and ``OTHER`` would both have to resolve
+   the conflict. If they resolved it in the same way, we would have caught that
+   in step 1. So they either both picked a different LCA value, or one (or
+   both) chose a new value to use.
+
+   If ``OTHER`` and ``THIS`` both picked a different LCA value, we conflict.
+
+   If ``OTHER`` and ``THIS`` both have values that are not LCA values, we also
+   conflict. (Same as 3-way, both sides modified a value in different ways.)
+
+5. (optional) The only tricky part is this: if ``OTHER`` has a LCA value, but
+   ``THIS`` does not, then we go with ``THIS``, and conversely if ``THIS`` has
+   an LCA value, but ``OTHER`` does not, then we go with ``OTHER``. The idea is
+   that ``THIS`` and ``OTHER`` may have resolved things in the same way, and
+   then later changed the value to something newer. (They could have also
+   resolved it differently, and then one side updated again.)
+
+
+``InventoryEntry.revision``
+---------------------------
+
+The last-modified revision for an entry gets treated differently. This is
+because how it is generated allows us to infer more information. Specifically,
+any time there is a change to an entry (rename, or content change) the last
+modified revision is updated. Further, if we are merging, and both sides
+updated the entry, then we update the last-modified revision at the merge
+point.
+
+For a picture example::
+
+    .   A
+    .  / \
+    . B   C
+    .  \ /
+    .   D
+
+
+For a single entry, the last modified revision in ``D`` is:
+
+1) ``A`` if neither ``B`` or ``C`` modified it
+2) ``B`` if ``B`` modified and ``C`` did not
+3) ``C`` if ``C`` modified and ``B`` did not
+4) ``D`` if ``B`` and ``C`` modified it
+
+This means that if the last modified revision is the same, there have been no
+changes in the intermediate time. If ``OTHER`` also has the same last modified
+revision as *any* LCA, then we know that all other LCAs' last-modified
+revisions are in the ancestry of that value. (Otherwise, when ``OTHER`` would
+need to create a new last modified revision as part of the merge.)
+
+..
+   vim: ft=rst tw=74 ai

=== modified file 'doc/en/user-guide/http_smart_server.txt'
--- doc/en/user-guide/http_smart_server.txt	2008-09-11 06:09:44 +0000
+++ doc/en/user-guide/http_smart_server.txt	2008-09-14 21:51:09 +0000
@@ -135,7 +135,9 @@
         root='/srv/example.com/www/code',
         prefix='/code/',
         path_var='REQUEST_URI',
-        readonly=True)
+        readonly=True,
+        load_plugins=True, 
+        enable_logging=True)
 
     fcgi.WSGIServer(smart_server_app).run()
         
@@ -159,7 +161,9 @@
         root='/srv/example.com/www/code',
         prefix='/code/',
         path_var='REQUEST_URI',
-        readonly=True)
+        readonly=True,
+        load_plugins=True, 
+        enable_logging=True)
 
     def handler(request):
         """Handle a single request."""

=== modified file 'setup.py'
--- setup.py	2008-09-24 00:09:49 +0000
+++ setup.py	2008-09-26 05:14:51 +0000
@@ -226,15 +226,18 @@
 
 
 add_pyrex_extension('bzrlib._btree_serializer_c')
-add_pyrex_extension('bzrlib._dirstate_helpers_c')
 add_pyrex_extension('bzrlib._knit_load_data_c')
 if sys.platform == 'win32':
+    add_pyrex_extension('bzrlib._dirstate_helpers_c',
+                         libraries=['Ws2_32']
+                       )
     # pyrex uses the macro WIN32 to detect the platform, even though it should
     # be using something like _WIN32 or MS_WINDOWS, oh well, we can give it the
     # right value.
     add_pyrex_extension('bzrlib._walkdirs_win32',
                         define_macros=[('WIN32', None)])
 else:
+    add_pyrex_extension('bzrlib._dirstate_helpers_c')
     add_pyrex_extension('bzrlib._readdir_pyx')
 ext_modules.append(Extension('bzrlib._patiencediff_c', ['bzrlib/_patiencediff_c.c']))
 

